/*
 *
 * filtprog_c.c -- support for filters using the C compiler
 *
 */

/*
 *      Copyrights:
 *
 *    	      Copyright (c) 1996 Smithsonian Astrophysical Observatory
 *
 *        Permission to use, copy, modify, distribute, and  sell  this
 *        software  and  its  documentation  for any purpose is hereby
 *        granted without  fee,  provided  that  the  above  copyright
 *        notice  appear  in  all  copies and that both that copyright
 *        notice and this permission notice appear in supporting docu-
 *        mentation,  and  that  the  name  of the  Smithsonian Astro-
 *	  physical Observatory not be used in advertising or publicity
 *  	  pertaining to distribution of the software without specific,
 *	  written  prior  permission.   The Smithsonian  Astrophysical 
 *        Observatory makes no representations about  the  suitability
 *	  of  this  software for any purpose.  It is provided  "as is"
 *	  without express or implied warranty.
 *        THE  SMITHSONIAN  ASTROPHYSICAL  OBSERVATORY  DISCLAIMS  ALL
 *	  WARRANTIES  WITH  REGARD  TO  THIS  SOFTWARE,  INCLUDING ALL
 *	  IMPLIED  WARRANTIES  OF  MERCHANTABILITY AND FITNESS, IN  NO
 *	  EVENT SHALL THE  SMITHSONIAN  ASTROPHYSICAL  OBSERVATORY  BE 
 *	  LIABLE FOR  ANY SPECIAL, INDIRECT  OR  CONSEQUENTIAL DAMAGES
 *	  OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS  OF USE,  DATA
 *	  OR  PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 *	  OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
 *	  THE  USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <file.h>
#include <find.h>
#include <macro.h>
#include <word.h>
#include <filter.h>
#include <regions_h.h>
#include <events_c.h>
#include <image_c.h>
#include <evregions_c.h>
#include <imregions_c.h>

#ifndef REGIONS_H
#define REGIONS_H ""
#endif

#ifndef EVENTS_C
#define EVENTS_C ""
#endif

#ifndef IMAGE_C
#define IMAGE_C ""
#endif

#ifndef EVREGIONS_C
#define EVREGIONS_C ""
#endif

#ifndef IMREGIONS_C
#define IMREGIONS_C ""
#endif

/*
 *
 * Private Routines
 *
 */

#ifdef ANSI_FUNC
static char *
GetType(int t)
#else
static char *GetType(t)
     int t;
#endif
{
  switch(t){
  case 'B':
    return("unsigned char");
    break;
  case 'I':
    return("short");
    break;
  case 'U':
    return("unsigned short");
    break;
  case 'J':
    return("int");
    break;
  case 'V':
    return("unsigned int");
    break;
  case 'E':
    return("float");
    break;
  case 'D':
    return("double");
    break;
  case 'X':
    return("char");
    break;
  default:
    return("void *");
    break;
  }
}

#ifdef ANSI_FUNC
static double
GetTloff(int t)
#else
static double GetTloff(t)
     int t;
#endif
{
  switch(t){
  case 'B':
  case 'I':
  case 'U':
  case 'J':
  case 'V':
  case 'X':
    return 1.0;
  case 'E':
  case 'D':
    return 0.5;
  default:
    return 1.0;
  }
}

#ifdef ANSI_FUNC
static int
GetN(int t, int n)
#else
static int GetN(t, n)
     int t;
     int n;
#endif
{
  switch(t){
  case 'X':
    return((n+7)/8);
    break;
  default:
    return(n);
    break;
  }
}

/*
 *
 * FilterProgOpen_C -- return filter program as a file for writing
 *
 */
#ifdef ANSI_FUNC
static int
FilterProgOpen_C (Filter filter)
#else
static int FilterProgOpen_C(filter)
     Filter filter;
#endif
{
  char tbuf[SZ_LINE];
  char *contents=NULL;
  char *tmpdir=NULL;

  /* make sure we have something to work with */
  if( !filter )
    return(0);

  /* get a unique executable name */
  if( !(tmpdir = (char *)getenv("FILTER_TMPDIR")) )
    tmpdir = DEFAULT_FILTER_TMPDIR;
  sprintf(tbuf, "%s/filt_XXXXXX", tmpdir);
  mktemp(tbuf);
  filter->prog = NewString(tbuf);

  /* use the same name for the routine name */
  sprintf(tbuf, "Filter%d%d", getpid(), FilterNum());
  filter->pname = NewString(tbuf);

  strcpy(tbuf, filter->prog);
  if( filter->debug ){
    filter->fp = stdout;
    return(1);
  }
  else{
    strcat(tbuf, ".c");
    if( (filter->fp = fopen(tbuf, "w")) == NULL ){
      fprintf(stderr, "ERROR: could not open filter program: %s\n", tbuf);
      return(0);
    }
  }

  /* prepend the filter header */
  contents = NewString(REGIONS_H);
  if( (contents != NULL) && (*contents != '\0') ){
    /* use printf so that we handle \n correctly in TEMPLATE */
    fprintf(filter->fp, "%s\n", contents);
    if( contents )
      free(contents);
  }
  return(1);
}

/*
 *
 * FilterProgClose_C -- close the filter program file
 *
 */
#ifdef ANSI_FUNC
static int 
FilterProgClose_C (Filter filter)
#else
static int FilterProgClose_C(filter)
     Filter filter;
#endif
{
  /* make sure we have something to work with */
  if( !filter )
    return(0);

  /* close file if we are not in debug mode */
  if( !filter->debug && filter->fp ){
    fclose(filter->fp);
    filter->fp = NULL;
  }
  return(1);
}

/*
 *
 * FilterProgAppend_C -- append the filter program body
 *
 */
#ifdef ANSI_FUNC
static int
FilterProgAppend_C (Filter filter)
#else
static int FilterProgAppend_C(filter)
     Filter filter;
#endif
{
  char *contents=NULL;

  /* make sure we have something to work with */
  if( !filter )
    return(0);

  /* make sure we are not in debug mode */
  if( filter->debug )
    return(1);

  /* for self-contained processes, write region code to output file */
  if(filter->ptype == PTYPE_CONTAINED){
    fprintf(filter->fp, "#define SELF_CONTAINED 1\n");
    switch( filter->type ){
    case TYPE_EVENTS:
      /* normally, we filter events analytically using evregions.o */
      if( !filter->evsect )
	contents = NewString(EVREGIONS_C);
      /* if evsect=xxx is specified, we filter by image pixels */
      else
	contents = NewString(IMREGIONS_C);
      break;
    case TYPE_IMAGE:
      /* image are filtered by image pixels */
      contents = NewString(IMREGIONS_C);
      break;
    }
    if( (contents != NULL) && (*contents != '\0') ){
      /* use fprintf so that we handle \n correctly in TEMPLATE */
      fprintf(filter->fp, "%s\n", contents);
      if( contents )
	free(contents);
    }
    else{
      fprintf(stderr, "ERROR: could not write body of filter program\n");
      if( contents )
	free(contents);
      return(0);
    }
  }

  /* get body of filter program */
  switch( filter->type ){
  case TYPE_EVENTS:
    contents = NewString(EVENTS_C);
    break;
  case TYPE_IMAGE:
    contents = NewString(IMAGE_C);
    break;
  }
  if( (contents != NULL) && (*contents != '\0') ){
    /* use fprintf so that we handle \n correctly in TEMPLATE */
    fprintf(filter->fp, "%s\n", contents);
    if( contents )
      free(contents);
    return(1);
  }
  else{
    fprintf(stderr, "ERROR: could not write body of filter program\n");
    if( contents )
      free(contents);
    return(0);
  }
}

/*
 *
 * FilterProgCompile_C -- compile the filter program
 *
 */
#ifdef ANSI_FUNC
static int
FilterProgCompile_C (Filter filter)
#else
static int FilterProgCompile_C(filter)
     Filter filter;
#endif
{
  char *s;
  char *math;
  char tbuf[SZ_LINE];
  char pmode[SZ_LINE];
  int len;
  int got;

  /* make sure we have something to work with */
  if( !filter )
    return(0);

  /* make sure we are not in debug mode */
  if( !filter->cc || filter->debug )
    return(1);

  /* we add the math library so users have access to math routines */
  math = "-lm";
      
  /* delete old version */
  unlink(filter->prog);

  switch(filter->ptype){
  case PTYPE_PROCESS:
    /* make up the compile command */
    sprintf(tbuf, "%s -o %s %s.c %s %s %s 1>/dev/null 2>%s.log", 
	    filter->cc, filter->prog, filter->prog, 
	    filter->objs ? filter->objs : " ", 
	    filter->extra, math, filter->prog);
    strcpy(pmode, "x");
    break;
  case PTYPE_CONTAINED:
    /* make up the compile command */
    sprintf(tbuf, "%s -o %s %s.c %s %s 1>/dev/null 2>%s.log", 
	    filter->cc, filter->prog, filter->prog, 
	    filter->extra, math, filter->prog);
    strcpy(pmode, "x");
    break;
#ifdef USE_DL
  case PTYPE_DYNAMIC:
    sprintf(tbuf,
	    "%s %s -o %s %s.c %s 1>/dev/null 2>%s.log", 
	    filter->cc, filter->shflags,
	    filter->prog, filter->prog, filter->extra, filter->prog);
    strcpy(pmode, "r");
    break;
#endif
  default:
    return(0);
  }
  
  /* issue the shell command to compile the program */
  got = system(tbuf);

  /* delete the filter program body in any case */
  sprintf(tbuf, "%s.c", filter->prog);
  if( !(s=getenv("FILTER_KEEP")) || !istrue(s) )
     unlink(tbuf);

  /* now we can see if we succeeded in issuing the command */
  if( got < 0 ){
    fprintf(stderr, "ERROR: could not run filter compilation\n");
    return(0);
  }

  /* if we have an executable program, we succeeded */
  sprintf(tbuf, "%s.log", filter->prog);
  if( (s=Find(filter->prog, pmode, NULL, FilterPath())) != NULL ){
    unlink(tbuf);
    if( s ) free(s);
  }
  else{
    fprintf(stderr, "ERROR: filter compilation failed\n");
    s = FileContents(tbuf, 0, &len);
    unlink(tbuf);
    if( *s && len ){
      fprintf(stderr, "Compilation error message:\n%s\n", s);
      if( s ) free(s);
    }
    return(0);
  }
  
  /* good news */
  return(1);
}


#ifdef ANSI_FUNC
static char *
_FilterInitString(char *filtstr)
#else
static char *_FilterInitString(filtstr)
     char *filtstr;
#endif
{
  char *ibuf, *iptr, *fptr;
  int paren=0;

  ibuf = calloc(strlen(filtstr)*10, sizeof(char));
  /* make sure we have at least one shape */
  if( !FilterShapeCount() )
    return ibuf;
  iptr = ibuf;
  fptr = filtstr;
  while( *fptr ){
    /* skip operators, parens, up to region name */
    while( *fptr && !(isalpha(*fptr) || (*fptr=='_') || (*fptr=='$')) ){
      fptr++;
    }
    /* might reach eof here */
    if( !*fptr )
      break;
    /* otherwise get next name */
    while( *fptr && (*fptr != '(') )
      *iptr++ = *fptr++;
    /* append init suffix */
    *iptr++ = 'i';
    /* copy the paren */
    if( *fptr == '(' ){
      *iptr++ = *fptr++;
      paren++;
    }
    /* copy til end of region, i.e., when paren goes to 0 */
    while( *fptr && paren ){
      if( *fptr == '(' ) 
	paren++;
      if( *fptr == ')' ) 
	paren--;
      *iptr++ = *fptr++;
    }
    /* copy ';' */
    *iptr++ = ';';
  }
  return(ibuf);
}

/*
 *
 * FilterProgWrite_C -- write the symbols for filtering
 *
 */
#ifdef ANSI_FUNC
static int
FilterProgWrite_C(Filter filter)
#else
static int FilterProgWrite_C(filter)
     Filter filter;
#endif
{
  int i;
  int offset;
  int pad;
  int dsize;
  int evsize;
  int tltyps[2];
  double tlmins[2];
  char vbuf[1024];
  char *v;
  char *ibuf;
  char *filtstr;
  FILE *fd;
  FilterSymbols sp;
  FITSHead fhd;

  /* make sure we have something to work with */
  if( !filter )
    return(0);

  /* make sure we are init'ed */
  if( filter->fhd == NULL ){
    fprintf(stderr, "ERROR: Symbol table not initialized\n");
    return(0);
  }

  /* make sure we are init'ed */
  if( filter->fp == NULL ){
    fprintf(stderr, "ERROR: No output file for parser\n");
    return(0);
  }

  /* get the filter string */
  if( (filtstr = (char *)_FilterString()) == NULL ){
    return(0);
  }

  /* init temps */
  fhd = filter->fhd;
  fd= filter->fp;
  offset = 0;
  evsize = 0;

  /* include math.h */
  fprintf(fd, "#include <math.h>\n");

  /* ptype-specific processing */
  switch(filter->ptype){
  case PTYPE_PROCESS:
  case PTYPE_CONTAINED:
    fprintf(fd, "#define FILTER_MAIN 1\n");
    break;
  }

  /* output counts of shapes */
  fprintf(fd, "#define NSHAPE %d\n",  FilterShapeCount());
  fprintf(fd, "#define NREGION %d\n", FilterRegionCount(TOK_IREG|TOK_EREG));
  /* output the filter itself */
  fprintf(fd, "#define FILTER %s\n", filtstr);
  if( !filter->debug ){
    /* string version of the filter -- used to check for FIELD optimization */
    fprintf(fd, "#define FILTSTR \"%s\"\n", filtstr);
    /* output the initialization string */
    ibuf = _FilterInitString(filtstr);
    fprintf(fd, "#define FINIT %s\n", ibuf);
    if( ibuf ) free(ibuf);
  }

  /* for type image, we generate an initialization string */
  switch(filter->type){
  case TYPE_IMAGE:
    fprintf(fd, "#define IMFILTRTN %s\n", filter->pname);
    break;
  case TYPE_EVENTS:
    fprintf(fd, "#define EVFILTRTN %s\n", filter->pname);
    /* output the event section and tlmin values, if necessary */
    if( filter->evsect && FilterTlmin(tlmins) && FilterTltyp(tltyps) ){
      fprintf(fd, "#define EVSECT %s\n", filter->evsect);
      fprintf(fd, "#define TLMINX %f\n", tlmins[0]);
      fprintf(fd, "#define TLMINY %f\n", tlmins[1]);
      fprintf(fd, "#define TLOFF %f\n",  GetTloff(tltyps[0]));
      fprintf(fd, "#define TLOFF %f\n",  GetTloff(tltyps[1]));
    }
    /* loop through the symbol table and process #defines */
    for(i=0; i<filter->nsyms; i++){
      sp = &(filter->symtab[i]);
      /* skip accidentally empty ones */
      if( (sp->name == NULL) || (*sp->name == '\0') )
	continue;
      /* process this type of symbol */
      switch(sp->type){
      case SYM_COL:
	/* make sure we are aligned */
	dsize = ft_sizeof(filter->fhd->table->col[sp->idx].type);
	pad =  dsize - (offset % dsize);
	if( pad == dsize ) pad = 0;
	offset += pad;
	evsize += pad;
	sp->offset = offset;
	if( fhd->table->col[sp->idx].n == 1 ){
	  fprintf(fd, "#define %s *((%s *)(eptr+%d))\n",
		  sp->name,
		  GetType((int)fhd->table->col[sp->idx].type),
		  offset);
	}
	else{
	  fprintf(fd, "#define %s ((%s *)(eptr+%d))\n",
		  sp->name,
		  GetType((int)fhd->table->col[sp->idx].type),
		  offset);
	}
	offset +=
	  fhd->table->col[sp->idx].size*fhd->table->col[sp->idx].n;
	evsize +=
	  fhd->table->col[sp->idx].size*fhd->table->col[sp->idx].n;
	break;
      case SYM_PAR:
	/* we have to distinguish between numbers and strings here */
	/* strip off enclosing white space */
	(void)nowhite(sp->value, vbuf);
	/* check for all white space */
	if( *vbuf == '\0' ){
	  fprintf(fd, "#define %s \"\"\n", sp->name);
	}
	else{
	  /* see if its a pure number */
	  (void)strtod(vbuf, &v);
	  if( (v == NULL) || (*v == '\0') )
	    fprintf(fd, "#define %s %s\n", sp->name, vbuf);
	  else
	    fprintf(fd, "#define %s \"%s\"\n", sp->name, vbuf);
	}
	break;
      }
    }
    /* make sure each record is aligned */
    pad =  8 - (evsize % 8);
    if( pad == 8 ) pad = 0;
    evsize += pad;
    /* make sure we have at least one byte to send to co-process */
    if( evsize <=0 ) evsize = 1;
    /* output the size of the filter record */
    fprintf(fd, "#define EVSIZE %d\n", evsize);
    /* save for later use */
    filter->evsize = evsize;
    break;
  default:
    break;
  }

  /* write it now */
  fflush(fd);
  return(1);
}

/*
 *
 * FilterProgEnd_C -- end the filtering process
 *
 */
#ifdef ANSI_FUNC
static int
FilterProgEnd_C (Filter filter)
#else
int FilterProgEnd_C(filter)
     Filter filter;
#endif
{
  char *s;
  char tbuf[SZ_LINE];
  int status=0;

  /* make sure we have something to play with */
  if( filter == NULL)
    return(0);

  /* delete the filter program */
  unlink(filter->prog);
  /* delete the filter program body, if necessary */
  sprintf(tbuf, "%s.c", filter->prog);
  if( !(s=getenv("FILTER_KEEP")) || !istrue(s) )
     unlink(tbuf);
  return(status);
}

/*
 *
 * Public Routines
 *
 */

/*
 *
 * FilterProgLoad_C -- load the routines needed to support C filtering
 *
 */
#ifdef ANSI_FUNC
int
FilterProgLoad_C (Filter filter)
#else
int FilterProgLoad_C(filter)
     Filter filter;
#endif
{
  /* make sure we have something to work with */
  if( !filter )
    return(0);

  filter->filt_open = FilterProgOpen_C;
  filter->filt_close = FilterProgClose_C;
  filter->filt_write = FilterProgWrite_C;
  filter->filt_append = FilterProgAppend_C;
  filter->filt_compile = FilterProgCompile_C;
  filter->filt_end = FilterProgEnd_C;
  return(1);
}
