/*
 *
 * GIO -- generalized I/O to disk files, streams, and memory
 *
 */

/*
 *      Copyrights:
 *
 *    	      Copyright (c) 1999 Smithsonian Astrophysical Observatory
 *
 *        Permission to use, copy, modify, distribute, and  sell  this
 *        software  and  its  documentation  for any purpose is hereby
 *        granted without  fee,  provided  that  the  above  copyright
 *        notice  appear  in  all  copies and that both that copyright
 *        notice and this permission notice appear in supporting docu-
 *        mentation,  and  that  the  name  of the  Smithsonian Astro-
 *	  physical Observatory not be used in advertising or publicity
 *  	  pertaining to distribution of the software without specific,
 *	  written  prior  permission.   The Smithsonian  Astrophysical 
 *        Observatory makes no representations about  the  suitability
 *	  of  this  software for any purpose.  It is provided  "as is"
 *	  without express or implied warranty.
 *        THE  SMITHSONIAN  ASTROPHYSICAL  OBSERVATORY  DISCLAIMS  ALL
 *	  WARRANTIES  WITH  REGARD  TO  THIS  SOFTWARE,  INCLUDING ALL
 *	  IMPLIED  WARRANTIES  OF  MERCHANTABILITY AND FITNESS, IN  NO
 *	  EVENT SHALL THE  SMITHSONIAN  ASTROPHYSICAL  OBSERVATORY  BE 
 *	  LIABLE FOR  ANY SPECIAL, INDIRECT  OR  CONSEQUENTIAL DAMAGES
 *	  OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS  OF USE,  DATA
 *	  OR  PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 *	  OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
 *	  THE  USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <gio.h>

/*
 *
 * Private Routines
 *
 */

#define EBUFSIZE 8192
static char _gerrors[EBUFSIZE];
static int _gerror = -1;

static char _gwarnings[EBUFSIZE];
static int _gwarning = -1;

/*
 *
 * gnew -- create a new GIO structure
 *
 */
#ifdef ANSI_FUNC
static GIO
gnew (void)
#else
static GIO gnew()
#endif
{
  GIO gio;
  gio = (GIO)calloc(1, sizeof(GFile));
  return(gio);
}

/*
 *
 * gfree -- destroy a GIO structure
 *
 */
#ifdef ANSI_FUNC
static void
gfree (GIO gio)
#else
static void gfree(gio)
     GIO gio;
#endif
{
  if( gio == NULL )
    return;
  if( gio->name )
    free(gio->name);
  if( gio->mode )
    free(gio->mode);
  free((void *)gio);
}


/*
 *
 * gmemseek -- memory "seek" of data
 *
 */
#ifdef ANSI_FUNC
static int
gmemseek (GIO gio, long offset, int whence)
#else
static int gmemseek(gio, offset, whence)
     GIO gio;
     long offset;
     int whence;
#endif
{
  int offset2;

  switch(whence){
  case 0:
    offset2 = offset;
    break;
  case 1:
    offset2 = gio->cur + offset;
    break;
  case 2:
    offset2 = gio->len + offset;
    break;
  default:
    return(-1);
  }
  if( offset2 >= gio->len ){
    if( gio->extend ){
      gio->len = offset2 + 1;
      gio->buf = (char *)realloc(gio->buf, gio->len);
      gio->cur = offset2;
      return(0);
    }
    else{
      return(-1);
    }
  } else {
    gio->cur = offset2;
    return(0);
  }
}

/*
 *
 * gmemread -- memory "read" of data
 *
 */
#ifdef ANSI_FUNC
static void *
gmemread (GIO gio, void *buf, int size, int n, int *got)
#else
static void *gmemread(gio, buf, size, n, got)
     GIO gio;
     void *buf;
     int size;
     int n;
     int *got;
#endif
{
  char *ptr;

  /* sanity checks */
  if( (gio==NULL) || (gio->buf == NULL) )
    return((void *)NULL);
  if( !strchr(gio->mode, 'r') ){
    gerror(stderr,
	   "Illegal read operation on write-only data (%s)\n", gio->name);
    return((void *)NULL);
  }
  /* see how much we can grab to "EOF" */
  *got = MIN(size*n, gio->len - gio->cur);
  /* copy bytes, if a buffer was supplied */
  if( buf ){
    ptr = buf;
    memmove(ptr, &(gio->buf[gio->cur]), *got);
  }
  else{
    ptr = &(gio->buf[gio->cur]);
  }
  /* bump position pointer */
  gio->cur += *got;
  /* convert got to elements */
  *got /= size;
  /* return buffer */
  return(ptr);
}

/*
 *
 * gmemwrite -- memory "write" of data
 *
 */
#ifdef ANSI_FUNC
static int
gmemwrite (GIO gio, char *buf, int size, int n)
#else
static int gmemwrite(gio, buf, size, n)
     GIO gio;
     char *buf;
     int size;
     int n;
#endif
{
  int n2;
  int len;

  /* sanity checks */
  if( (gio==NULL) || (gio->buf == NULL) )
    return(0);
  if( !strchr(gio->mode, 'r') ){
    gerror(stderr,
	   "Illegal write operation on read-only data (%s)\n", gio->name);
    return(0);
  }
  len = size*n;
  /* make sure we have enough space */
  if( gio->cur + len >= gio->len ){
    if( gio->extend ){
      gio->len += MAX(len, GIO_BUFINC);
      if( (gio->buf = (char *)realloc(gio->buf, gio->len)) != NULL )
	n2 = len;
      /* could not grab space -- try just a bit to recover */
      else{
	gerror(stderr, "Could not realloc space for memory write\n");
	n2 = 0;
      }
    }
    else{
      n2 = gio->len - gio->cur;
    }
  }
  else
    n2 = len;
  /* copy bytes */
  memmove(&(gio->buf[gio->cur]), buf, n2);
  /* bump pointer */
  gio->cur += n2;
  return(n/size);
}

/* 
 *
 * gexec -- open a connected subprocess with bidirectional
 * IPC channels, implemented with pipes for this version
 *
 */
#ifdef ANSI_FUNC
static int
gexec(GIO gio, char *cmd, char *mode)
#else
static int gexec(gio, cmd, mode)
     GIO gio;
     char *cmd;
     char *mode;
#endif
{
  int pin[2], pout[2];
  int maxforks = 3;
  char *argv[4096];

  /* start out pessimisticaly */
  gio->ifd = -1;
  gio->ofd = -1;

  /* open pipes */
  pipe(pin);
  if( pipe(pout) != 0){
    return(-1);
  }

  /* Create child process. The child inherits the open stdio files. 
   * The fork can fail if swap space is full or if we have too many processes.
   */
  while ((gio->pid = fork()) == -1) {
    if (--maxforks == 0) {
      close (pin[0]);  close (pin[1]);
      close (pout[0]); close (pout[1]);
      return(-1);
    }
    sleep (2);
  }
  if(gio->pid == 0) {
    /* New child process.  Make child think the pipe is its stdin/out.
     */
    close (pin[0]);
    close (pout[1]);
    close (0);  dup (pout[0]);  close (pout[0]);
    close (1);  dup (pin[1]);   close (pin[1]);

    /* run through shell so that we expand shell meta characters */
    argv[0] = "sh";
    argv[1] = "-c";
    argv[2] = cmd;
    argv[3] = NULL;

    /* Exec the new process.  Will not return if successful. */
    execvp(argv[0], argv);
    
    /* If we get here the new process could not be executed for some
     * reason.  Shutdown, calling _exit to avoid flushing parent's
     * io buffers.
     */
    _exit(1);
  } else {
    /* Existing, parent process. */
    close (pin[1]);
    close (pout[0]);
    if( strchr(mode, 'r') )
      gio->ifd  = pin[0];
    else{
      close(pin[0]);
    }
    if( strchr(mode, 'w') )
      gio->ofd = pout[1];
    else{
      close(pout[1]);
    }
  }
  /* clean up and return */
  return(0);
}

/*
 *
 * gopenpipe -- open a command pipe
 *
 */
#ifdef ANSI_FUNC
static GIO
gopenpipe (char *name, char *mode)
#else
static GIO gopenpipe(name, mode)
     char *name;
     char *mode;
#endif
{
  GIO gio;
  char *s;

  /* allocate the new struct */
  if( (gio = gnew()) == NULL )
    return(NULL);

  /* skip past prefix */
  s = name;
  if( !strncasecmp(s, "pipe:", 5) ) s += 5;
  while( isspace((int)*s) ) s++;

  /* start the piped process */
  if( !gexec(gio, s, mode) ){
    gclose(gio);
    return(NULL);
  }

  /* fill in the record */
  gio->name = NewString(s);
  gio->mode = NewString(mode);
  gio->type = GIO_PIPE;
  gio->extend = 0;
  return(gio);
}

/*
 *
 * gopenstdio -- open stdin or stdout
 *
 */
#ifdef ANSI_FUNC
static GIO
gopenstdio (char *name, char *mode)
#else
static GIO gopenstdio(name, mode)
     char *name;
     char *mode;
#endif
{
  GIO gio;

  /* allocate a new struct */
  if( (gio = gnew()) == NULL )
    return(NULL);

  if( !strncasecmp(name, "stdout", 6) ){
    if( strchr(mode, 'r') ){
      gerror(stderr, "can't open STDOUT for reading\n");
      gfree(gio);
      return(NULL);
    }
    else{
      /* fill in the record */
      gio->name = NewString("stdout");
      gio->mode = NewString(mode);
      gio->type = GIO_STREAM;
      gio->extend = 0;
      gio->fp = stdout;
    }
  }
  /* look for stdin */
  else if( !strncasecmp(name, "stdin", 5) ){
    if( strchr(mode, 'w') ){
      gerror(stderr, "Can't open STDIN for writing\n");
      gfree(gio);
      return(NULL);
    }
    else{
      /* fill in the record */
      gio->name = NewString("stdin");
      gio->mode = NewString(mode);
      gio->type = GIO_STREAM;
      gio->extend = 0;
      gio->fp = stdin;
    }
  }
  else{
    gerror(stderr, "Invalid stdout specification (%s)\n", name);
    gfree(gio);
    return(NULL);
  }
  return(gio);
}

#ifdef HAVE_SYS_SHM_H
/*
 *
 * gopenshm -- open shared memory for reading or writing
 *
 */
#ifdef ANSI_FUNC
static GIO
gopenshm (char *name, char *mode)
#else
static GIO gopenshm(name, mode)
     char *name;
     char *mode;
#endif
{
  GIO gio;
  int shmid=-1;
  int shmkey=-1;
  int shmoff=0;
  int shmflg=0;
  int shmlen=0;
  char *shmaddr=NULL;
  char *shmbuf;
  char *s, *t;

  /* allocate a new struct */
  if( (gio = gnew()) == NULL )
    return(NULL);

  /* skip past prefix */
  s = name;
  if( !strncasecmp(s, "shm:", 4) ) s += 4;
  while( isspace((int)*s) ) s++;

  /* fill in the record */
  gio->name = NewString(s);
  gio->mode = NewString(mode);
  gio->type = GIO_SHM;
  gio->extend = 0;

  /* parse the shared memory specification */
  if( !s || !*s ){
    gerror(stderr, "missing shared memory specification\n");
    goto error;
  }
  if( *s == '@' ){
    s++;
    shmkey = strtol(s, &t, 0);
    s = t;
  }
  else{
    shmid = strtol(s, &t, 0);
    s = t;
  }
  if( (*s != ':') && !isspace((int)*s) ){
    gerror(stderr, "missing memory mapping length (%s)\n", name);
    goto error;
  }
  else{
    s++;
    shmlen = strtol(s, &t, 0);
    s = t;
  }

  /* map shared memory containing a data */
  if( shmid == -1 ){
    if( shmkey != -1 )
      shmid = shmget(shmkey, shmlen, 0400);
    else{
      gerror(stderr, "invalid shared memory key (%s)\n", name);
      goto error;
    }
    if( shmid == -1 ){
	gerror(stderr, "invalid shared memory id (%s)\n", name);
	goto error;
    }
  }
  shmbuf = (char *)shmat(shmid, shmaddr, shmflg);
  if( shmbuf == (char *)-1){
    gerror(stderr, "can't map shared memory image (%s)\n", name);
    goto error;
  }
  /* set the shared memory id */
  gio->buf = shmbuf;
  gio->len = shmlen;
  gio->shmid = shmid;
  gio->shmoff = shmoff;
  return(gio);

error:
  gfree(gio);
  return(NULL);
}
#endif

#ifdef HAVE_SYS_MMAN_H
/*
 *
 * gopenmmap -- open memory mapped file for reading or writing
 *
 */
#ifdef ANSI_FUNC
static GIO
gopenmmap (char *name, char *mode)
#else
static GIO gopenmmap(name, mode)
     char *name;
     char *mode;
#endif
{
  GIO gio;
  char xmode[SZ_LINE];
  int prot;
  int flags;
  int len;
  char *buf;
  char *s;
  struct stat statbuf;

  /* allocate a new struct */
  if( (gio = gnew()) == NULL )
    return(NULL);

  /* skip past prefix */
  s = name;
  if( !strncasecmp(s, "mmap:", 5) ) s += 5;
  while( isspace((int)*s) ) s++;

  /* fill in the record */
  gio->name = NewString(s);
  gio->mode = NewString(mode);
  gio->type = GIO_MMAP;

  *xmode = '\0';
  if( strchr(mode, 'r') ) strcat(xmode, "r");
  if( strchr(mode, 'w') ) strcat(xmode, "w");
  if( strchr(mode, 'a') ) strcat(xmode, "a");
  gio->fp = fopen(s, gio->mode);
  if( gio->fp == NULL ){
    gerror(stderr, "can't open data file (%s)\n", s);
    goto error;
  }
  fstat(fileno(gio->fp), &statbuf);
  len = statbuf.st_size;
  prot = 0;
  if( strchr(mode, 'r') ) prot |= PROT_READ;
  if( strchr(mode, 'w') ) prot |= PROT_WRITE;
  flags = MAP_PRIVATE;
  if( (buf = mmap(0, len, prot, flags, fileno(gio->fp), 0)) == NULL ){
    gerror(stderr, "can't mmap data file (%s)\n", s);
    goto error;
  }
  gio->buf = (void *)buf;
  gio->len = len;
  gio->cur = 0;
  if( strchr(mode, 'w') ) 
    gio->extend = 1;
  else
    gio->extend = 0;
  return(gio);

error:
  gfree(gio);
  return(NULL);
}
#endif

/*
 *
 * gopenfile -- open ordinary file for reading or writing
 *
 */
#ifdef ANSI_FUNC
static GIO
gopenfile (char *name, char *mode)
#else
static GIO gopenfile(name, mode)
     char *name;
     char *mode;
#endif
{
  GIO gio;
  char xmode[SZ_LINE];
  char *s;

  /* allocate a new struct */
  if( (gio = gnew()) == NULL )
    return(NULL);

  /* skip past prefix */
  s = name;
  if( !strncasecmp(s, "file:", 5) ) s += 5;
  if( !strncasecmp(s, "unfile:", 7) ){
    gio->unbuffered = 1;
    s += 7;
  }
  while( isspace((int)*s) ) s++;

  /* fill in the record */
  gio->name = NewString(s);
  gio->mode = NewString(mode);
  gio->type = GIO_DISK;
  gio->extend = 1;
  if( strchr(mode, 'u') ) gio->unbuffered = 1;

  *xmode = '\0';
  if( strchr(mode, 'r') ) strcat(xmode, "r");
  if( strchr(mode, 'w') ) strcat(xmode, "w");
  if( strchr(mode, 'a') ) strcat(xmode, "a");
  gio->fp = fopen(s, gio->mode);
  if( gio->fp == NULL ){
    gerror(stderr, "can't open data file (%s)\n", s);
    goto error;
  }
  else{
    return(gio);
  }

error:
  gfree(gio);
  return(NULL);
}

/*
 *
 * Public Routines
 *
 */

/*
 *
 * gopen -- open a file or a stream
 *
 */
#ifdef ANSI_FUNC
GIO
gopen (char *name, char *mode)
#else
GIO gopen(name, mode)
     char *name;
     char *mode;
#endif
{
  GIO gio;

  /* make sure we have a name */
  if( (name==NULL) || (*name=='\0') )
    return(NULL);

  /* look for command pipe */
  if( !strncasecmp(name, "pipe:", 5) || strchr(mode, 'p') )
    gio = gopenpipe(name, mode);
#ifdef HAVE_SYS_MMAN_H
  /* look for a mapped file */
  else if( !strncasecmp(name, "mmap:", 5) || strchr(mode, 'm') )
    gio = gopenmmap(name, mode);
#endif
#ifdef HAVE_SYS_SHM_H
  /* look for shared memory segment */
  else if( !strncasecmp(name, "shm:", 4) || strchr(mode, 's') )
    gio = gopenshm(name, mode);
#endif
  /* look for an explicit file */
  else if( !strncasecmp(name, "file:", 5)   ||
           !strncasecmp(name, "unfile:", 7) || strchr(mode, 'u') )
    gio = gopenfile(name, mode);
  /* look for stdio */
  else if( !strncasecmp(name, "stdin", 5) || !strncasecmp(name, "stdout", 6) )
    gio = gopenstdio(name, mode);
  /* if all else fails, its an ordinary file */
  else
    gio = gopenfile(name, mode);
  return(gio);
}

/*
 *
 * gmemopen -- open memory for reading or writing
 *
 */
#ifdef ANSI_FUNC
GIO
gmemopen (char *buf, int len, char *mode)
#else
GIO gmemopen(buf, len, mode)
     char *buf;
     int len;
     char *mode;
#endif
{
  GIO gio;

  /* allocate a new struct */
  if( (gio = gnew()) == NULL )
    return(NULL);

  /* fill in the record */
  gio->name = NewString("MEMORY");
  gio->mode = NewString(mode);
  gio->type = GIO_MEMORY;

  /* process open type */
  if( strchr(mode, 'w') ){
    if( (buf!=NULL) && (len>0) ){
      gio->buf = (void *)buf;
      gio->len = len;
    }
    else{
      gio->buf = (char *)malloc(GIO_BUFINC);
      gio->len = GIO_BUFINC;
    }
    gio->cur = 0;
    gio->extend = 1;
  } else {
    gio->buf = (void *)buf;
    gio->len = len;
    gio->cur = 0;
    gio->extend = 0;
  }

  /* return the news */
  return(gio);
}

/*
 *
 * _gread -- read bytes from a data I/O structure
 *
 */
#ifdef ANSI_FUNC
void *
_gread (GIO gio, char *buf, int size, int n, int *got)
#else
void *_gread(gio, buf, size, n, got)
     GIO gio;
     char *buf;
     int size;
     int n;
     int *got;
#endif
{
  char *ptr;

  /* sanity checks */
  if( gio==NULL )
    return(NULL);
  if( !strchr(gio->mode, 'r') ){
    gerror(stderr, 
	   "Illegal read operation on write-only data (%s)\n", gio->name);
    return(NULL);
  }

  switch(gio->type){
  case GIO_DISK:
    if( buf )
      ptr = buf;
    else
      ptr = (char *)malloc(size*n);
    if( gio->unbuffered ){
      *got = read(fileno(gio->fp), ptr, size*n);
      if( *got >=0 )*got /= size;
    }
    else{
      *got = fread(ptr, size, n, gio->fp);
    }
    if( *got <= 0 ){
      if( !buf )
	free(ptr);
      ptr = NULL;
    }
    return(ptr);
    break;
  case GIO_STREAM:
    if( buf )
      ptr = buf;
    else
      ptr = (char *)malloc(size*n);
    *got = fread(ptr, size, n, gio->fp);
    if( *got <= 0 ){
      if( !buf )
	free(ptr);
      ptr = NULL;
    }
    return(ptr);
    break;
  case GIO_MEMORY:
    return(gmemread(gio, buf, size, n, got));
    break;
#ifdef HAVE_SYS_MMAN_H
  case GIO_MMAP:
    return(gmemread(gio, buf, size, n, got));
    break;
#endif
#ifdef HAVE_SYS_SHM_H
  case GIO_SHM:
    return(gmemread(gio, buf, size, n, got));
    break;
#endif
  case GIO_PIPE:
    if( buf )
      ptr = buf;
    else
      ptr = (char *)malloc(size*n);
    if( gio->ifd >=0 )
      *got = read(gio->ifd, ptr, size*n);
    if( *got <= 0 ){
      if( !buf )
	free(ptr);
      ptr = NULL;
    }
    else{
      gio->cur += *got;
      *got /= size;
    }
    return(ptr);
    break;
  default:
    return(NULL);
  }
}

/*
 *
 * gread -- read bytes from a data I/O structure
 *
 */
#ifdef ANSI_FUNC
int
gread (GIO gio, char *buf, int size, int n)
#else
int gread(gio, buf, size, n)
     GIO gio;
     char *buf;
     int size;
     int n;
#endif
{
  int got;
  if( buf == NULL )
    return(0);
  _gread(gio, buf, size, n, &got);
  return (got);
}

/*
 *
 * gwrite -- write bytes to a structure
 *
 */
#ifdef ANSI_FUNC
int
gwrite (GIO gio, char *buf, int size, int n)
#else
int gwrite(gio, buf, size, n)
     GIO gio;
     char *buf;
     int size;
     int n;
#endif
{
  int got;

  /* sanity checks */
  if( gio==NULL )
    return(0);
  if( !strchr(gio->mode, 'w') && !strchr(gio->mode, 'a') ){
    gerror(stderr, 
	   "Illegal write operation on read-only data (%s)\n", gio->name);
    return(0);
  }

  switch(gio->type){
  case GIO_DISK:
    if( gio->unbuffered ){
      got = write(fileno(gio->fp), buf, size*n);
      if( got >=0 ) got /= size;
      return(got);
    }
    else{
      return(fwrite(buf, size, n, gio->fp));
    }
    break;
  case GIO_STREAM:
    return(fwrite(buf, size, n, gio->fp));
    break;
  case GIO_MEMORY:
    return(gmemwrite(gio, buf, size, n));
    break;
#ifdef HAVE_SYS_MMAN_H
  case GIO_MMAP:
    return(gmemwrite(gio, buf, size, n));
    break;
#endif
#ifdef HAVE_SYS_SHM_H
  case GIO_SHM:
    return(gmemwrite(gio, buf, size, n));
    break;
#endif
  case GIO_PIPE:
    if( gio->ofd >= 0 ){
      got = write(gio->ofd, buf, size * n);
      if( got <= 0 )
	return(got);
      else{
	gio->cur += got;
	return(got/size);
      }
    }
    else
      return(0);
    break;
  default:
    return(0);
  }
}

/*
 *
 * gflush -- flush I/O
 *
 */
#ifdef ANSI_FUNC
int
gflush (GIO gio)
#else
int gflush(gio)
     GIO gio;
#endif
{
  long ftell();

  /* sanity checks */
  if( gio==NULL )
    return(-1);

  switch(gio->type){
  case GIO_DISK:
    if( gio->unbuffered )
      return(0);
    else
      return(fflush(gio->fp));
    break;
  case GIO_STREAM:
    return(fflush(gio->fp));
    break;
  case GIO_MEMORY:
    return(0);
    break;
#ifdef HAVE_SYS_MMAN_H
  case GIO_MMAP:
    return(0);
    break;
#endif
#ifdef HAVE_SYS_SHM_H
  case GIO_SHM:
    return(0);
    break;
#endif
  case GIO_PIPE:
    return(0);
    break;
  default:
    return(-1);
  }
}

/*
 *
 * gseek -- set current pointer position in a "file"
 *
 */
#ifdef ANSI_FUNC
int
gseek (GIO gio, long offset, int whence)
#else
int gseek(gio, offset, whence)
     GIO gio;
     long offset;
     int whence;
#endif
{
  /* sanity checks */
  if( gio==NULL )
    return(-1);

  switch(gio->type){
  case GIO_DISK:
    if( gio->unbuffered )
      return(lseek(fileno(gio->fp), offset, whence));
    else
      return(fseek(gio->fp, offset, whence));
    break;
  case GIO_STREAM:
    return(fseek(gio->fp, offset, whence));
    break;
  case GIO_MEMORY:
    return(gmemseek(gio, offset, whence));
    break;
#ifdef HAVE_SYS_MMAN_H
  case GIO_MMAP:
    return(gmemseek(gio, offset, whence));
    break;
#endif
#ifdef HAVE_SYS_SHM_H
  case GIO_SHM:
    return(gmemseek(gio, offset, whence));
    break;
#endif
  case GIO_PIPE:
    return(-1);
    break;
  default:
    return(-1);
  }
}

/*
 *
 * gtell -- return current pointer position in a "file"
 *
 */
#ifdef ANSI_FUNC
long
gtell (GIO gio)
#else
long gtell(gio)
     GIO gio;
#endif
{
  long lval;
  /* sanity checks */
  if( gio==NULL )
    return((long)-1);

  switch(gio->type){
  case GIO_DISK:
    if( gio->unbuffered )
      return(lseek(fileno(gio->fp), 0, SEEK_CUR));
    else
      return(ftell(gio->fp));
    break;
  case GIO_STREAM:
    lval = ftell(gio->fp);
    if( lval >=0 )
      return lval;
    else
      return gio->cur;
    break;
  case GIO_MEMORY:
    return(gio->cur);
    break;
#ifdef HAVE_SYS_MMAN_H
  case GIO_MMAP:
    return(gio->cur);
    break;
#endif
#ifdef HAVE_SYS_SHM_H
  case GIO_SHM:
    return(gio->cur);
    break;
#endif
  case GIO_PIPE:
    return(gio->cur);
    break;
  default:
    return((long)-1);
  }
}

/*
 *
 * gskip - skip reading bytes to a file position
 *
 */
#ifdef ANSI_FUNC
int
gskip (GIO gio, int n)
#else
int gskip(gio, n)
     GIO gio;
     int n;
#endif
{
  char buf[1];

  /* sanity checks */
  if( !gio )
    return(-1);

  switch(gio->type){
  case GIO_DISK:
    if( gio->unbuffered )
      return(lseek(fileno(gio->fp), (long)n, SEEK_CUR));
    else
      return(fseek(gio->fp, (long)n, SEEK_CUR));
    break;
  case GIO_STREAM:
    /* for stdin, read the bytes */
    if( gio->fp == stdin ){
      if( n < 0 )
	return(-1);
      while( n-- ){
	if( fread(buf, sizeof(char), 1, gio->fp) != 1 )
	  return(0);
      }
      return(0);
    }
    /* who knows what will happen here */
    else{
      return(ftell(gio->fp));
    }
    break;
  case GIO_MEMORY:
    if( (gio->cur + n >= gio->len) || (gio->cur + n <0) ){
      return(-1);
    }
    else{
      gio->cur += n;
      return(0);
    }
    break;
#ifdef HAVE_SYS_MMAN_H
  case GIO_MMAP:
    if( (gio->cur + n >= gio->len) || (gio->cur + n <0) ){
      return(-1);
    }
    else{
      gio->cur += n;
      return(0);
    }
    break;
#endif
#ifdef HAVE_SYS_SHM_H
  case GIO_SHM:
    if( (gio->cur + n >= gio->len) || (gio->cur + n <0) ){
      return(-1);
    }
    else{
      gio->cur += n;
      return(0);
    }
    break;
#endif
  case GIO_PIPE:
    if( n < 0 )
      return(-1);
    if( gio->ifd >= 0 ){
      while( n-- ){
	if( read(gio->ifd, buf, 1) != 1 )
	  return(0);
	gio->cur++;
      }
      return(0);
    }
    else{
      return(-1);
    }
    break;
  default:
    return(-1);
  }
}

/*
 *
 * ginfo -- return information about a "file"
 *
 */
#ifdef ANSI_FUNC
int
ginfo (GIO gio, char **name, int *type, int *cur, char **buf, int *len)
#else
int ginfo(gio, name, type, cur, buf, len)
     GIO gio;
     char **name;
     int *type;
     int *cur;
     char **buf;
     int *len;
#endif
{
  if( gio==NULL )
    return(-1);
  if( name );
    *name = gio->name;
  if( type )
    *type = gio->type;
  if( cur )
    *cur = gtell(gio);
  if( buf )
    *buf = gio->buf;
  if( len )
    *len = gio->len;
  return(0);
}

/*
 *
 * gfreebuf -- free allocated data buffers
 *
 */
#ifdef ANSI_FUNC
void
gfreebuf (GIO gio, void *buf)
#else
void gfreebuf(gio, buf)
     GIO gio;
     void *buf;
#endif
{
  if( (gio==NULL) )
    return;

  switch(gio->type){
  case GIO_DISK:
    if( buf )
      free(buf);
    break;
  case GIO_STREAM:
    if( buf )
      free(buf);
    break;
  case GIO_MEMORY:
    break;
#ifdef HAVE_SYS_MMAN_H
  case GIO_MMAP:
    break;
#endif
#ifdef HAVE_SYS_SHM_H
  case GIO_SHM:
    break;
#endif
  case GIO_PIPE:
    if( buf )
      free(buf);
    break;
  }
}

/*
 *
 * gclose -- close an I/O structure
 *
 */
#ifdef ANSI_FUNC
void
gclose (GIO gio)
#else
void gclose(gio)
     GIO gio;
#endif
{
  int status;

  /* sanity checks */
  if( gio==NULL )
    return;

  switch(gio->type){
  case GIO_DISK:
    if( gio->unbuffered ){
      fclose(gio->fp);
    }
    else{
      fflush(gio->fp);
      fclose(gio->fp);
    }
    break;
  case GIO_STREAM:
    fflush(gio->fp);
    break;
  case GIO_MMAP:
    munmap(gio->buf, gio->len);
    break;
#ifdef HAVE_SYS_MMAN_H
  case GIO_MEMORY:
    break;
#endif
#ifdef HAVE_SYS_SHM_H
  case GIO_SHM:
    break;
#endif
  case GIO_PIPE:
    if( gio->ifd >=0 )
      close (gio->ifd);
    if( gio->ofd >=0 )
      close (gio->ofd);
    if( gio->pid )
      waitpid(gio->pid, &status, 0);
    break;
  }
  gfree(gio);
}

/*
 *
 * gerrorstring -- return last exception string
 *
 */
#ifdef ANSI_FUNC
char *
gerrorstring(void)
#else
char *gerrorstring()
#endif
{
  return(_gerrors);
}

/*
 *
 * setgerror -- set the error flag
 *
 */
#ifdef ANSI_FUNC
int
setgerror(int flag)
#else
int setgerror(flag)
     int flag;
#endif
{
  int oflag;
  oflag = _gerror;
  _gerror = flag;
  return oflag;
}

/*
 *
 * gerror - error message handler
 * actions based on value of error flag:
 *
 *  flag    action
 *  ----    ------
 *   0      store error message in _gerrors string
 *   1      store, print error message and continue
 *   2      print error message and exit
 *
 */
#ifdef __STDC__
void gerror(FILE *fd, char *format, ...) 
{
    char tbuf[SZ_LINE];
    va_list args;
    va_start(args, format);
#else
void gerror(va_alist) va_dcl
{
    FILE *fd;
    char *format;
    char tbuf[SZ_LINE];
    va_list args;

    va_start(args);
    fd  = va_arg(args, FILE *);
    format = va_arg(args, char *);
#endif
    /* initialize level, if not already done */
    if( _gerror == -1 ){
      char *s;
      if( (s=getenv("GERROR")) )
	_gerror = atoi(s);
      else
	_gerror = 1;
    }
    sprintf(tbuf, "ERROR: %s", format);
    vsprintf(_gerrors, tbuf, args);
    /* if the error flag is positive, we output immediately */
    if( (fd != NULL) && _gerror ){
      fprintf(fd, _gerrors);
      fflush(fd);
    }
    /* if the error flag is set high, we exit */
    if( _gerror >= 2 )
      exit(_gerror);
}

/*
 *
 * gwarningstring -- return last exception string
 *
 */
#ifdef ANSI_FUNC
char *
gwarningstring(void)
#else
char *gwarningstring()
#endif
{
  return(_gwarnings);
}

/*
 *
 * setgwarning -- set the warning flag
 *
 */
#ifdef ANSI_FUNC
int
setgwarning(int flag)
#else
int setgwarning(flag)
     int flag;
#endif
{
  int oflag;
  oflag = _gwarning;
  _gwarning = flag;
  return oflag;
}

#ifdef __STDC__
void gwarning(FILE *fd, char *format, ...) 
{
    char tbuf[SZ_LINE];
    va_list args;
    va_start(args, format);
#else
void gwarning(va_alist) va_dcl
{
    FILE *fd;
    char *format;
    char tbuf[SZ_LINE];
    va_list args;

    va_start(args);
    fd  = va_arg(args, FILE *);
    format = va_arg(args, char *);
#endif
    /* initialize level, if not already done */
    if( _gwarning == -1 ){
      char *s;
      if( (s=getenv("GWARNING")) )
	_gwarning = atoi(s);
      else
	_gwarning = 1;
    }
    sprintf(tbuf, "WARNING: %s", format);
    vsprintf(_gwarnings, tbuf, args);
    /* if the warning flag is positive, we output immediately */
    if( (fd != NULL) && _gwarning ){
      fprintf(fd, _gwarnings);
      fflush(fd);
    }
}

