#include "boxannulus.h"
#include "frame.h"
#include "util.h"

BoxAnnulus::BoxAnnulus(Frame* p, const Vector& ctr, const Vector& s, double a,
		     const char* clr, const char* f, const char* t, 
		     unsigned short prop)
  : BaseBox(p, ctr, a, clr, f, t, prop)
{
  annuli = 1;
  size = new Vector[1];
  size[0] = s;

  strcpy(type,"box annulus");
  handle = new Vector[4];
  numHandle = 4;

  updateBBox();
}

BoxAnnulus::BoxAnnulus(Frame* p, const Vector& ctr,
		       const Vector& inner, const Vector& outer, int num,
		       double a,
		       const char* clr, const char* f, const char* t, 
		       unsigned short prop)
  : BaseBox(p, ctr, a, clr, f, t, prop)
{
  annuli = num+1;
  size = new Vector[annuli];

  for (int i=0; i<annuli; i++)
    size[i] = ((outer-inner)/num)*i+inner;

  strcpy(type,"box annulus");
  numHandle = 4 + annuli;
  handle = new Vector[numHandle];

  updateBBox();
}

BoxAnnulus::BoxAnnulus(Frame* p, const Vector& ctr, double a, int an,
		       Vector* s,
		       const char* clr,  const char* f, const char* t, 
		       unsigned short prop)
  : BaseBox(p, ctr, a, clr, f, t, prop)
{
  annuli = an;
  size = new Vector[annuli];

  for (int i=0; i<annuli; i++)
    size[i] = s[i];
  sortSize();

  strcpy(type, "box annulus");
  numHandle = 4 + annuli;
  handle = new Vector[numHandle];

  updateBBox();
  doCallBacks(&editCB);
}

void BoxAnnulus::editBegin(int h)
{
  if (h<5) {
    switch (h) {
    case 1:
      return;
    case 2:
      size[annuli-1] = Vector(-size[annuli-1][0],size[annuli-1][1]);
      return;
    case 3:
      size[annuli-1] = -size[annuli-1];
      return;
    case 4:
      size[annuli-1] = Vector(size[annuli-1][0],-size[annuli-1][1]);
      return;
    }
  }
}

void BoxAnnulus::edit(const Vector& v, int h)
{
  // This sizes about the center
  //  size = ((v-center)*Rotate(angle)*2).abs();

  if (h<5) {
    // This sizes about the opposite node

    Vector old = size[annuli-1];
    Vector op = center+size[annuli-1]/2;
    size[annuli-1] = op - 
      v * Translate(-center) * flip * Rotate(-angle) * Translate(center);
    center = v+((size[annuli-1]/2) * (Rotate(angle) * flip));

    double r0 = fabs(size[annuli-1][0]/old[0]);
    double r1 = fabs(size[annuli-1][1]/old[1]);
    for (int i=0; i<annuli-1; i++) {
      size[i][0] *= r0;
      size[i][1] *= r1;
    }
  }
  else {
    double x = (v * Translate(-center) * flip * Rotate(-angle) * 2).abs()[0];
    double y = x*size[h-5][1]/size[h-5][0];
    size[h-5] = Vector(x,y);
  }

  updateBBox();
  doCallBacks(&editCB);
}

void BoxAnnulus::editEnd()
{
  for (int i=1; i<annuli; i++)
    size[i] = size[i].abs();
  sortSize();

  updateBBox();
  doCallBacks(&editCB);
}

void BoxAnnulus::setSize(const Vector& inner, const Vector& outer,
			  int num)
{
  annuli = num+1;
  if (size)
    delete [] size;
  size = new Vector[annuli];

  for (int i=0; i<annuli; i++)
    size[i] = ((outer-inner)/num)*i+inner;
  sortSize();

  numHandle = 4 + annuli;
  delete [] handle;
  handle = new Vector[numHandle];

  updateBBox();
  doCallBacks(&editCB);
}

void BoxAnnulus::setSize(const Vector* s, int num)
{
  annuli = num;
  if (size)
    delete [] size;
  size = new Vector[annuli];

  for (int i=0; i<annuli; i++)
    size[i] = s[i];
  sortSize();

  numHandle = 4 + annuli;
  delete [] handle;
  handle = new Vector[numHandle];

  updateBBox();
  doCallBacks(&editCB);
}

// private

void BoxAnnulus::updateHandles()
{
  // handles are in canvas coords
  // we can't garantee that the radii have been sorted yet

  Vector max;
  for(int i=0; i<annuli; i++)
    if (max[0]<size[i][0])
      max = size[i];

  Vector s = max;
  if (properties & FIXED)
    s /= parent->getZoom();

  Matrix m = Rotate(angle) * flip * Translate(center) * parent->refToCanvas;

  handle[0] = Vector(-s[0],-s[1])/2 * m;
  handle[1] = Vector( s[0],-s[1])/2 * m;
  handle[2] = Vector( s[0], s[1])/2 * m;
  handle[3] = Vector(-s[0], s[1])/2 * m;

  // the rest are radii

  {
    for (int i=0; i<annuli; i++) {
      Vector ss = Vector((size[i][0])/2.,0) * m;
      if (properties & FIXED)
	ss /= parent->getZoom();
      handle[i+4] = ss;
    }
  }
}

void BoxAnnulus::sortSize()
{
  for (int i=0; i<annuli; i++)
    for (int j=i+1; j<annuli; j++)
      if (size[i][0]>size[j][0]) {
	Vector d = size[i];
	size[i] = size[j];
	size[j] = d;
      }
}

int BoxAnnulus::addSize(const Vector& v)
{
  // new size array
  Vector* old = size;
  size = new Vector[annuli+1];

  // copy old values
  for (int i=0; i<annuli; i++)
    size[i] = old[i];

  // delete old
  if (old)
    delete [] old;

  // new size on end
  size[annuli][0] = ((v*Translate(-center)*flip*Rotate(-angle)).abs())[0];
  size[annuli][1] = size[annuli][0]*size[0][1]/size[0][0];
  annuli++;

  numHandle = 4 + annuli;
  delete [] handle;
  handle = new Vector[numHandle];

  updateBBox();
  //  doCallBacks(&editCB);

  // return number of size
  return annuli;
}

void BoxAnnulus::deleteSize(int h)
{
  if (annuli>1 && h<annuli) {
    // new size array
    Vector* old = size;
    size = new Vector[annuli-1];

    // copy up to size in question
    {
      for (int i=0; i<h; i++)
	size[i] = old[i];
    }

    // copy remainder
    {
      for (int i=h; i<annuli-1; i++)
	size[i] = old[i+1];
    }

    if (old)
      delete [] old;
    annuli--;

    numHandle = 4 + annuli;
    delete [] handle;
    handle = new Vector[numHandle];

    updateBBox();
    doCallBacks(&editCB);
  }
}

// list

void BoxAnnulus::list(ostream& str, int sys, int format, const char* delim)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      for (int i=0; i<annuli; i++) {
	listCoordSystem(str, sys);
	str << ';';

	Vector r = parent->mapLenFromRef(size[i],(Frame::CoordSystem)sys);
	str << "box(" << setprecision(8) << v[0] << ',' << v[1] << ',' 
	    << r[0] << ',' << r[1] << ',' << radToDeg(angle) << ')';

	if (i!=0) {
	  Vector r1 = parent->mapLenFromRef(size[i-1],(Frame::CoordSystem)sys);
	  str << " & !box(" << setprecision(8) << v[0] << ',' << v[1] 
	      << ',' << r1[0] << ',' << r1[1] << ',' << radToDeg(angle) << ')';
	}
	else
	  listProperties(str);

	str << delim;
      }
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	for (int i=0; i<annuli; i++) {
	  listCoordSystem(str, sys);
	  str << ';';

	  Vector r = parent->mapLenFromRef(size[i],
					   (Frame::CoordSystem)sys,
					   Frame::ARCSEC);
	  str << "box(" << setprecision(8) << v[0] << ',' << v[1] 
	      << ',' << r[0] << "\"" << ',' << r[1] << "\"" 
	      << ',' << radToDeg(angle) << ')';

	  if (i!=0) {
	    Vector r1 = parent->mapLenFromRef(size[i-1],
					      (Frame::CoordSystem)sys,
					      Frame::ARCSEC);
	    str << " & !box(" << setprecision(8) << v[0] << ',' << v[1] 
		<< ',' << r1[0] << "\"" << ',' << r1[1] << "\"" 
		<< ',' << radToDeg(angle) <<')';
	  }
	  else
	    listProperties(str);

	  str << delim;
	}
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;

	for (int i=0; i<annuli; i++) {
	  listCoordSystem(str, sys);
	  str << ';';

	  Vector r = parent->mapLenFromRef(size[i],
					   (Frame::CoordSystem)sys,
					   Frame::ARCSEC);
	  str << "box(" << ra << ',' << dec << ',' 
	      << r[0] << "\""<< ',' << r[1] << "\""<< ',' 
	      << radToDeg(angle) << ')';

	  if (i!=0) {
	    Vector r1 = parent->mapLenFromRef(size[i-1],
					      (Frame::CoordSystem)sys,
					      Frame::ARCSEC);
	    str << "& !box(" << ra << ',' << dec << ',' 
		<< r1[0] << "\""<< ',' << r1[1] << "\""<< ',' 
		<< radToDeg(angle) << ')';
	  }
	  else
	    listProperties(str);

	  str << delim;
	}
      }
      break;
    }
    break;
  }
}

void BoxAnnulus::listSAOimage(ostream& str, const char* delim)
{
  // all coords are in image coords

  //  if (!(properties&INCLUDE))
  //    str << '-';

  for (int i=0; i<annuli; i++) {
    Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
    str << "box(" << setprecision(8) << v[0] << ',' << v[1] << ',' 
	<< size[i][0] << ',' << size[i][1] << ',' << radToDeg(angle) << ')';

    if (i!=0)
      str << " & !box(" << setprecision(8) << v[0] << ',' << v[1] 
	  << ',' << size[i-1][0] << ',' << size[i-1][1] << ',' 
	  << radToDeg(angle) << ')';

    str << delim;
  }
}
