#include <strstream.h>

#include "circle.h"
#include "frame.h"
#include "util.h"

Circle::Circle(Frame* p, const Vector& ctr, double r, 
	       const char* clr, const char* f, const char* t, 
	       unsigned short prop)
  : BaseCircle(p, ctr, 0, clr, f, t, prop)
{
  annuli = 1;
  radii = new double[1];
  radii[0] = r;

  strcpy(type, "circle");
  handle = new Vector[4];
  numHandle = 4;

  updateBBox();
}

void Circle::updateBBox()
{
  // bound marker

  Vector c = center * parent->refToCanvas;
  Vector r = Vector(radii[0],radii[0]);
  if (!(properties & FIXED))
    r *= parent->getZoom();
  bbox = BBox(c-r, c+r);

  // generate handles

  handle[0] = bbox.ll;
  handle[1] = bbox.lr();
  handle[2] = bbox.ur;
  handle[3] = bbox.ul();

  // bound handles

  for (int i=0; i<numHandle; i++)
    bbox.bound(handle[i]);

  // make room for handles

  bbox.expand(3);

  // calculate overall bbox

  calcAllBBox();
}

void Circle::edit(const Vector& v, int h)
{
  Vector r = (v-center).abs();
  radii[0] = r[0] > r[1] ? r[0] : r[1];

  updateBBox();
  doCallBacks(&editCB);
}

void Circle::setRadius(double r)
{
  radii[0] = r;

  updateBBox();
  doCallBacks(&editCB);
}

// list

void Circle::list(ostream& str, int sys, int format, const char* delim)
{
  listCoordSystem(str, sys);
  str << ';';

  if (!(properties&INCLUDE))
    str << '-';

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << parent->mapLenFromRef(radii[0], (Frame::CoordSystem)sys) << ')';
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ','
	    << parent->mapLenFromRef(radii[0], 
				     (Frame::CoordSystem)sys,
				     Frame::ARCSEC) << '"' << ')';
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << '(' << ra << ',' << dec << ',' 
	    << parent->mapLenFromRef(radii[0], 
				     (Frame::CoordSystem)sys,
				     Frame::ARCSEC) << '"' << ')';
      }
      break;
    }
    break;
  }

  listProperties(str);
  str << delim;
}

void Circle::listCiao(ostream& str, int sys, int format, const char* delim)
{
  if (!(properties&INCLUDE))
    str << '-';

  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << parent->mapLenFromRef(radii[0], (Frame::CoordSystem)sys) << ')';
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    {
      char buf[64];
      parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			  (Frame::CoordFormat)format, buf, 64);
      char ra[16];
      char dec[16];
      istrstream wcs(buf,64);
      wcs >> ra >> dec;
      str << type << '(' << ra << ',' << dec << ',' 
	  << parent->mapLenFromRef(radii[0],
				   (Frame::CoordSystem)sys,
				   Frame::ARCMIN) << "')";
    }
    break;
  }

  str << delim;
}

void Circle::listFtools(ostream& str, int sys, int format, const char* delim)
{
  if (!(properties&INCLUDE))
    str << '-';

  Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
  str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
      << parent->mapLenFromRef(radii[0], (Frame::CoordSystem)sys) << ')';

  str << delim;
}

void Circle::listPros(ostream& str, int sys, int format, const char* delim)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
      str << type << ' ' << setprecision(8) << v << radii[0];
    }
    break;

  case Frame::PHYSICAL:
    {
      str << "physical; ";
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      str << type << ' ' << setprecision(8) << v 
	  << parent->mapLenFromRef(radii[0], (Frame::CoordSystem)sys);
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    listProsCoordSystem(str,sys);
    str << "; ";

    // display wcs radius in ARCSEC

    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	str << type << ' ' << setprecision(8) << v[0] << "d " << v[1] << "d " 
	    << parent->mapLenFromRef(radii[0],
				     (Frame::CoordSystem)sys,
				     Frame::ARCSEC) << '"';
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;
	str << type << ' ' << ra << ' ';
	if (dec[0]=='+')
	  str << dec+1 << ' ';
	else
	  str << dec << ' ';

	str << parent->mapLenFromRef(radii[0],
				     (Frame::CoordSystem)sys,
				     Frame::ARCSEC) << '"';
      }
      break;
    }
    break;
  }

  str << delim;
}

void Circle::listSAOimage(ostream& str, const char* delim)
{
  // all coords are in image coords

  if (!(properties&INCLUDE))
    str << '-';

  Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
  str << "circle(" << setprecision(8) << v[0] << ',' << v[1] 
      << ',' << radii[0] << ')';

  str << delim;
}
