#include "ellipseannulus.h"
#include "frame.h"
#include "util.h"

EllipseAnnulus::EllipseAnnulus(Frame* p, const Vector& ctr, const Vector& r, 
			       double a, 
			       const char* clr,  const char* f, const char* t, 
			       unsigned short prop)
  : BaseEllipse(p, ctr, a, clr, f, t, prop)
{
  annuli = 1;
  radii = new Vector[1];

  radii[0] = r;

  strcpy(type,"ellipse annulus");
  handle = new Vector[4];
  numHandle = 4;

  updateBBox();
}

EllipseAnnulus::EllipseAnnulus(Frame* p, const Vector& ctr, 
			       const Vector& inner, const Vector& outer, 
			       int num,
			       double a, 
			       const char* clr,  const char* f, const char* t, 
			       unsigned short prop)
  : BaseEllipse(p, ctr, a, clr, f, t, prop)
{
  annuli = num+1;
  radii = new Vector[annuli];

  for (int i=0; i<annuli; i++)
    radii[i] = ((outer-inner)/num)*i+inner;

  strcpy(type, "ellipse annulus");
  numHandle = 4 + annuli;
  handle = new Vector[numHandle];

  updateBBox();
  doCallBacks(&editCB);
}

EllipseAnnulus::EllipseAnnulus(Frame* p, const Vector& ctr, double a, int an,
			       Vector* r,
			       const char* clr,  const char* f, const char* t, 
			       unsigned short prop)
  : BaseEllipse(p, ctr, a, clr, f, t, prop)
{
  annuli = an;
  radii = new Vector[annuli];

  for (int i=0; i<annuli; i++)
    radii[i] = r[i];
  sortRadii();

  strcpy(type, "ellipse annulus");
  numHandle = 4 + annuli;
  handle = new Vector[numHandle];

  updateBBox();
  doCallBacks(&editCB);
}

void EllipseAnnulus::edit(const Vector& v, int h)
{
  if (h<5) {
    for (int i=0; i<annuli; i++)
      radii[i] = (v * Translate(-center) * flip * Rotate(-angle)).abs()
	/radii[annuli-1][0]*radii[i][0];
  }
  else {
    double x = ((v * Translate(-center) * flip * Rotate(-angle)).abs())[0];
    double y = x*radii[h-5][1]/radii[h-5][0];
    radii[h-5] = Vector(x,y);
  }

  updateBBox();
  doCallBacks(&editCB);
}

void EllipseAnnulus::editEnd()
{
  sortRadii();

  updateBBox();
  doCallBacks(&editCB);
}

void EllipseAnnulus::setRadii(const Vector& inner, const Vector& outer,
			      int num)
{
  annuli = num+1;
  if (radii)
    delete [] radii;
  radii = new Vector[annuli];

  for (int i=0; i<annuli; i++)
    radii[i] = ((outer-inner)/num)*i+inner;
  sortRadii();

  numHandle = 4 + annuli;
  delete [] handle;
  handle = new Vector[numHandle];

  updateBBox();
  doCallBacks(&editCB);
}

void EllipseAnnulus::setRadii(const Vector* r, int num)
{
  annuli = num;
  if (radii)
    delete [] radii;
  radii = new Vector[annuli];

  for (int i=0; i<annuli; i++)
    radii[i] = r[i];
  sortRadii();

  numHandle = 4 + annuli;
  delete [] handle;
  handle = new Vector[numHandle];

  updateBBox();
  doCallBacks(&editCB);
}

// private

void EllipseAnnulus::updateHandles()
{
  // handles are in canvas coords
  // we can't garantee that the radii have been sorted yet

  Vector max;
  for(int i=0; i<annuli; i++)
    if (max[0]<radii[i][0])
      max = radii[i];

  Vector r = max;
  if (properties & FIXED)
    r /= parent->getZoom();

  Matrix m = Rotate(angle) * flip * Translate(center) * parent->refToCanvas;

  handle[0] = Vector(-r[0],-r[1]) * m;
  handle[1] = Vector( r[0],-r[1]) * m;
  handle[2] = Vector( r[0], r[1]) * m;
  handle[3] = Vector(-r[0], r[1]) * m;

  // the rest are radii
  {
    for (int i=0; i<annuli; i++) {
      Vector rr = Vector(radii[i][0],0) * m;
      if (properties & FIXED)
	rr /= parent->getZoom();
      handle[i+4] = rr;
    }
  }
}

void EllipseAnnulus::sortRadii()
{
  for (int i=0; i<annuli; i++)
    for (int j=i+1; j<annuli; j++)
      if (radii[i][0]>radii[j][0]) {
	Vector d = radii[i];
	radii[i] = radii[j];
	radii[j] = d;
      }
}

int EllipseAnnulus::addRadii(const Vector& v)
{
  // new radii array
  Vector* old = radii;
  radii = new Vector[annuli+1];

  // copy old values
  for (int i=0; i<annuli; i++)
    radii[i] = old[i];

  // delete old
  if (old)
    delete [] old;

  // new radii on end
  radii[annuli][0] = ((v*Translate(-center)*flip*Rotate(-angle)).abs())[0];
  radii[annuli][1] = radii[annuli][0]*radii[0][1]/radii[0][0];
  annuli++;

  numHandle = 4 + annuli;
  delete [] handle;
  handle = new Vector[numHandle];

  updateBBox();
  //  doCallBacks(&editCB);

  // return number of radii
  return annuli;
}

void EllipseAnnulus::deleteRadii(int h)
{
  if (annuli>1 && h<annuli) {
    // new radii array
    Vector* old = radii;
    radii = new Vector[annuli-1];

    // copy up to radii in question
    {
      for (int i=0; i<h; i++)
	radii[i] = old[i];
    }

    // copy remainder
    {
      for (int i=h; i<annuli-1; i++)
	radii[i] = old[i+1];
    }

    if (old)
      delete [] old;
    annuli--;

    numHandle = 4 + annuli;
    delete [] handle;
    handle = new Vector[numHandle];

    updateBBox();
    doCallBacks(&editCB);
  }
}

// list

void EllipseAnnulus::list(ostream& str, int sys, int format, const char* delim)
{
  switch ((Frame::CoordSystem)sys) {
  case Frame::IMAGE:
  case Frame::PHYSICAL:
  case Frame::DETECTOR:
  case Frame::AMPLIFIER:
    {
      Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
      for (int i=0; i<annuli; i++) {
	listCoordSystem(str, sys);
	str << ';';

	Vector r = parent->mapLenFromRef(radii[i],(Frame::CoordSystem)sys);
	str << "ellipse(" << setprecision(8) << v[0] << ',' << v[1] << ',' 
	    << r[0] << ',' << r[1] << ',' << radToDeg(angle) << ')';

	if (i!=0) {
	  Vector r1 = parent->mapLenFromRef(radii[i-1],(Frame::CoordSystem)sys);
	  str << " & !ellipse(" << setprecision(8) << v[0] << ',' << v[1] 
	      << ',' << r1[0] << ',' << r1[1] << ',' << radToDeg(angle) << ')';
	}
	else
	  listProperties(str);

	str << delim;
      }
    }
    break;

  case Frame::EQUATORIAL_FK4:
  case Frame::EQUATORIAL_FK5:
  case Frame::ECLIPTIC:
  case Frame::GALACTIC:
  case Frame::LINEAR:
  case Frame::ICRS:
    switch ((Frame::CoordFormat)format) {
    case Frame::DEGREES:
      {
	Vector v = parent->map(center,Frame::REF,(Frame::CoordSystem)sys);
	for (int i=0; i<annuli; i++) {
	  listCoordSystem(str, sys);
	  str << ';';

	  Vector r = parent->mapLenFromRef(radii[i],
					   (Frame::CoordSystem)sys,
					   Frame::ARCSEC);
	  str << "ellipse(" << setprecision(8) << v[0] << ',' << v[1] 
	      << ',' << r[0] << "\"" << ',' << r[1] << "\"" 
	      << ',' << radToDeg(angle) << ')';

	  if (i!=0) {
	    Vector r1 = parent->mapLenFromRef(radii[i-1],
					      (Frame::CoordSystem)sys,
					      Frame::ARCSEC);
	    str << " & !ellipse(" << setprecision(8) << v[0] << ',' << v[1] 
		<< ',' << r1[0] << "\"" << ',' << r1[1] << "\"" 
		<< ',' << radToDeg(angle) <<')';
	  }
	  else
	    listProperties(str);

	  str << delim;
	}
      }
      break;
    case Frame::SEXAGESIMAL:
      {
	char buf[64];
	parent->mapToWCSStr(center, Frame::REF, (Frame::CoordSystem)sys,
			    (Frame::CoordFormat)format, buf, 64);
	char ra[16];
	char dec[16];
	istrstream wcs(buf,64);
	wcs >> ra >> dec;

	for (int i=0; i<annuli; i++) {
	  listCoordSystem(str, sys);
	  str << ';';

	  Vector r = parent->mapLenFromRef(radii[i],
					   (Frame::CoordSystem)sys,
					   Frame::ARCSEC);
	  str << "ellipse(" << ra << ',' << dec 
	      << ',' << r[0] << "\"" << ',' << r[1] << "\"" 
	      << ',' << radToDeg(angle) << ')';

	  if (i!=0) {
	    Vector r1 = parent->mapLenFromRef(radii[i-1],
					      (Frame::CoordSystem)sys,
					      Frame::ARCSEC);
	    str << "& !ellipse(" << ra << ',' << dec 
		<< ',' << r1[0] << "\"" << ',' << r1[1] << "\""
		<< ',' << radToDeg(angle) << ')';
	  }
	  else
	    listProperties(str);

	  str << delim;
	}
      }
      break;
    }
    break;
  }
}

void EllipseAnnulus::listSAOimage(ostream& str, const char* delim)
{
  // all coords are in image coords

  //  if (!(properties&INCLUDE))
  //    str << '-';

  for (int i=0; i<annuli; i++) {
    Vector v = parent->map(center,Frame::REF,Frame::IMAGE);
    str << "ellipse(" << setprecision(8) << v[0] << ',' << v[1] << ',' 
	<< radii[i][0] << ',' << radii[i][1] << ',' << radToDeg(angle) << ')';

    if (i!=0)
      str << " & !ellipse(" << setprecision(8) << v[0] << ',' << v[1] 
	  << ',' << radii[i-1][0] << ',' << radii[i-1][1] << ',' 
	  << radToDeg(angle) << ')';

    str << delim;
  }
}
