#include "frame.h"
#include "fitsimage.h"

Vector Frame::map(const Vector& v, CoordSystem in, CoordSystem out, int* valid)
{
  if (valid) // assume we will succeed
    *valid = 1;

  if (in==out)
    return v;

  switch (in) {
  case NATIVEWCS:
  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    return mapWCS(v,in,out,valid);

  case IMAGE:
    return mapImage(v,out,valid);
  case PHYSICAL:
    return mapPhysical(v,out,valid);
  case AMPLIFIER:
    return mapAmplifier(v,out,valid);
  case DETECTOR:
    return mapDetector(v,out,valid);
  case REF:
    return mapRef(v,out,valid);
  case USER:
    return mapUser(v,out,valid);
  case CANVAS:
    return mapCanvas(v,out,valid);
  case PANNER:
    return mapPanner(v,out,valid);

  default:
    cerr << "Frame Internal Error: unable to map invalid coordinate system." 
	 << endl;
  }

  // if we get here, we have an error condition

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapWCS(const Vector& v, CoordSystem in, CoordSystem out, 
		     int* valid)
{
  if (currentFits) {
    if (in!=NATIVEWCS) {
      switch (out) {
      case IMAGE:
	return currentFits->wcs2pix(v, in);
      case REF:
	return currentFits->wcs2pix(v, in) * currentFits->getImageToRef();
      case USER:
	return currentFits->wcs2pix(v, in) * currentFits->getImageToUser();
      default:
	cerr << "Frame Internal Error: unable to map WCS." << endl;
      }
    }
    else {
      switch (out) {
      case IMAGE:
	return currentFits->wcs2pix(v);
      case REF:
	return currentFits->wcs2pix(v) * currentFits->getImageToRef();
      case USER:
	return currentFits->wcs2pix(v) * currentFits->getImageToUser();
      default:
	cerr << "Frame Internal Error: unable to map WCS." << endl;
      }
    }
  }

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapImage(const Vector& v, CoordSystem out, int* valid)
{
  if (currentFits) {
    switch (out) {
    case REF:
      return v * currentFits->getImageToRef();
    case USER:
      return v * currentFits->getImageToUser();

    default:
      cerr << "Frame Internal Error: unable to map IMAGE." << endl;
    }
  }

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapPhysical(const Vector& v, CoordSystem out, int* valid)
{
  if (currentFits) {
    switch (out) {
    case REF:
      return v * currentFits->getPhysicalToRef();
    case USER:
      return v * currentFits->getPhysicalToUser();

    default:
      cerr << "Frame Internal Error: unable to map Physical" << endl;
    }
  }

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapAmplifier(const Vector& v, CoordSystem out, int* valid)
{
  if (currentFits) {
    switch (out) {
    case REF:
      return v * currentFits->getAmplifierToRef();
    case USER:
      return v * currentFits->getAmplifierToUser();

    default:
      cerr << "Frame Internal Error: unable to map AMPLIFIER." << endl;
    }
  }

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapDetector(const Vector& v, CoordSystem out, int* valid)
{
  if (currentFits) {
    switch (out) {
    case REF:
      return v * currentFits->getDetectorToRef();
    case USER:
      return v * currentFits->getDetectorToUser();

    default:
      cerr << "Frame Internal Error: unable to map DETECTOR." << endl;
    }
  }

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapRef(const Vector& v, CoordSystem out, int* valid)
{
  switch (out) {
  case USER:
    return v * refToUser;
  case WIDGET:
    return v * refToWidget;
  case CANVAS:
    return v * refToCanvas;
  case WINDOW:
    return v *refToWindow;

  case IMAGE:
    if (currentFits)
      return v * currentFits->getRefToImage();
    break;
  case PHYSICAL:
    if (currentFits)
      return v * currentFits->getRefToPhysical();
    break;
  case AMPLIFIER:
    if (currentFits)
      return v * currentFits->getRefToAmplifier();
    break;
  case DETECTOR:
    if (currentFits)
      return v * currentFits->getRefToDetector();
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getRefToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  if (ptr->hasWCS())
	    return ptr->pix2wcs(v * ptr->getRefToImage(), out);

	ptr = ptr->next();
      }
    }
    break;

  default:
    cerr << "Frame Internal Error: unable to map REF." << endl;
  }

  if (valid)
    *valid = 0;
  return v;
}      

Vector Frame::mapUser(const Vector& v, CoordSystem out, int* valid)
{
  switch (out) {
  case REF:
    return v * userToRef;
  case WIDGET:
    return v * userToWidget;
  case CANVAS:
    return v * userToCanvas;

  case IMAGE:
    if (currentFits)
      return v * currentFits->getUserToImage();
    break;
  case PHYSICAL:
    if (currentFits)
      return v * currentFits->getUserToPhysical();
    break;
  case AMPLIFIER:
    if (currentFits)
      return v * currentFits->getUserToAmplifier();
    break;
  case DETECTOR:
    if (currentFits)
      return v * currentFits->getUserToDetector();
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getUserToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  if (ptr->hasWCS())
	    return ptr->pix2wcs(v * ptr->getUserToImage(), out);

	ptr = ptr->next();
      }
    }
    break;

  default:
    cerr << "Frame Internal Error: unable to map USER." << endl;
  }

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapCanvas(const Vector& v, CoordSystem out, int* valid)
{
  switch (out) {
  case REF:
    return v * canvasToRef;
  case USER:
    return v * canvasToUser;
  case WIDGET:
    return v * canvasToWidget;
  case WINDOW:
    return v * canvasToWindow;

  case IMAGE:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getCanvasToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  return v * ptr->getCanvasToImage();

	ptr = ptr->next();
      }
    }
    break;
  case PHYSICAL:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getCanvasToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  if (ptr->hasLTMV())
	    return v * ptr->getCanvasToPhysical();

	ptr = ptr->next();
      }
    }
    break;
  case AMPLIFIER:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getCanvasToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  if (ptr->hasATMV() && ptr->hasLTMV())
	    return v * ptr->getCanvasToAmplifier();

	ptr = ptr->next();
      }
    }
    break;
  case DETECTOR:
    if (currentFits && currentFits->hasDTMV() && currentFits->hasLTMV())
      return v * currentFits->getCanvasToDetector();
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getCanvasToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  if (ptr->hasWCS())
	    return ptr->pix2wcs(v * ptr->getCanvasToImage(), out);

	ptr = ptr->next();
      }
    }
    break;

  default:
    cerr << "Frame Internal Error: unable to map CANVAS." << endl;
  }

  if (valid)
    *valid = 0;
  return v;
}

Vector Frame::mapPanner(const Vector& v, CoordSystem out, int* valid)
{
  switch (out) {
  case REF:
    return v * pannerToRef;
  case USER:
    return v * pannerToUser;

  case IMAGE:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getPannerToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  return v * ptr->getPannerToImage();

	ptr = ptr->next();
      }
    }
    break;
  case PHYSICAL:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getPannerToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  if (ptr->hasLTMV())
	    return v * ptr->getPannerToPhysical();
	
	ptr = ptr->next();
      }
    }
    break;
  case AMPLIFIER:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getPannerToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height())
	  if (ptr->hasATMV() && ptr->hasLTMV())
	    return v * ptr->getPannerToAmplifier();

	ptr = ptr->next();
      }
    }
    break;
  case DETECTOR:
    if (currentFits && currentFits->hasDTMV() && currentFits->hasLTMV())
      return v * currentFits->getPannerToDetector();
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * ptr->getPannerToData();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]< ptr->height())
	  if (ptr->hasWCS())
	    return ptr->pix2wcs(v * ptr->getPannerToImage(), out);

	ptr = ptr->next();
      }
    }
    break;

  default:
    cerr << "Frame Internal Error: unable to map PANNER." << endl;
  }

  if (valid)
    *valid = 0;
  return v;
}

void Frame::mapToWCSStr(const Vector& v, CoordSystem in, CoordSystem out, 
			CoordFormat format, char* buf, int length)
{
  switch (in) {
  case REF:
    mapMatrixToWCSStr(v, out, format, buf, length,
		      &FitsImage::getRefToData, &FitsImage::getRefToImage);
    return;
  case USER:
    mapMatrixToWCSStr(v, out, format, buf, length,
		      &FitsImage::getUserToData, &FitsImage::getUserToImage);
    return;
  case CANVAS:
    mapMatrixToWCSStr(v, out, format, buf, length, 
		      &FitsImage::getCanvasToData, 
		      &FitsImage::getCanvasToImage);
    return;
  case PANNER:
    mapMatrixToWCSStr(v, out, format, buf, length,
		      &FitsImage::getPannerToData, 
		      &FitsImage::getPannerToImage);
    return;

  default:
    cerr << "Frame Internal Error: unable to mapToWCSStr invalid " 
	 << "coordinate system." << endl;
  }

  // if we get here, we have an error condition

  strcpy(buf,"");
}

void Frame::mapMatrixToWCSStr(const Vector& v, CoordSystem wcs, 
			      CoordFormat format, char* buf, int length,
			      Matrix& (FitsImage::*getMatrixData)(),
			      Matrix& (FitsImage::*getMatrixImage)())
{
  if (fits) {
    if (fitsCount==1) { // only one wcs to check
      fits->pix2wcs(v * (fits->*getMatrixImage)(), wcs, format, buf, length);
      return;
    }
    else { // check all available wcs
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = v * (ptr->*getMatrixData)();
	if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	    img[1]<ptr->height()) {
	  ptr->pix2wcs(v * (ptr->*getMatrixImage)(), wcs, format, buf, length);
	  return;
	}
	ptr = ptr->next();
      }

      // ok, its not on any images, use the first wcs
      fits->pix2wcs(v * (fits->*getMatrixImage)(), wcs, format, buf, length);
      return;
    }
  }

  // if we get here, we have an error condition

  strcpy(buf,"");
}

// mapLen

double Frame::mapLenToRef(double d, CoordSystem sys, CoordFormat format)
{
  Vector r = mapLenToRef(Vector(d,d), sys, format);
  return r[0];
}

Vector Frame::mapLenToRef(const Vector& v, CoordSystem sys, CoordFormat format)
{
  // what we are doing is mapping len to ref type coords (really image coords)

  if (currentFits) {
    switch (sys) {
    case CANVAS:
      return v; // special case
    case IMAGE:
      return v;
    case PHYSICAL:
      return v * (currentFits->getPhysicalToImage()).abs();
    case AMPLIFIER:
      return v * (currentFits->getAmplifierToPhysical()*
		  currentFits->getPhysicalToImage()).abs();
    case DETECTOR:
      return v * (currentFits->getDetectorToPhysical()*
		  currentFits->getPhysicalToImage()).abs();
    case NATIVEWCS:
    case EQUATORIAL_FK4:
    case EQUATORIAL_FK5:
    case GALACTIC:
    case ECLIPTIC:
    case LINEAR:
    case ICRS: 
      switch (format) {
      case DEGREES:
	return currentFits->wcs2pixDist(v);
      case ARCMIN:
	return currentFits->wcs2pixDist(v/60);
      case ARCSEC:
	return currentFits->wcs2pixDist(v/(60*60));
      }
      break;
    }
  }

  // error
  cerr << "Frame Internal Error: mapLenToRef." << endl;
  return v;
}

double Frame::mapLenFromRef(double d, CoordSystem sys, CoordFormat format)
{
  Vector r = mapLenFromRef(Vector(d,d),sys,format);
  return r[0];
}

Vector Frame::mapLenFromRef(const Vector& v,CoordSystem sys,CoordFormat format)
{
  // really from image coords

  if (currentFits) {
    switch (sys) {
    case IMAGE:
      return v;
    case PHYSICAL:
      return v * (currentFits->getImageToPhysical()).abs();
    case AMPLIFIER:
      return v * (currentFits->getImageToPhysical()*
		  currentFits->getPhysicalToAmplifier()).abs();
    case DETECTOR:
      return v * (currentFits->getImageToPhysical()*
		  currentFits->getPhysicalToDetector()).abs();

    case EQUATORIAL_FK4:
    case EQUATORIAL_FK5:
    case GALACTIC:
    case ECLIPTIC:
    case LINEAR:
    case ICRS:
      switch (format) {
      case DEGREES:
	return currentFits->pix2wcsDist(v);
      case ARCMIN:
	return currentFits->pix2wcsDist(v*60);
      case ARCSEC:
	return currentFits->pix2wcsDist(v*60*60);
      }
      break;
    }
  }

  // error
  cerr << "Frame Internal Error: mapLenFromRef" << endl;
  return v;
}

double Frame::mapLenToRef(double d, CoordFormat format)
{
  Vector r = mapLenToRef(Vector(d,d), format);
  return r[0];
}

Vector Frame::mapLenToRef(const Vector& v, CoordFormat format)
{
  // really to image coords

  if (currentFits) {
    switch (format) {
    case PIXELS:
      return Vector(v).abs();
    case DEGREES:
      return currentFits->wcs2pixDist(v);
    case ARCMIN:
      return currentFits->wcs2pixDist(v/60);
    case ARCSEC:
      return currentFits->wcs2pixDist(v/(60*60));
    }
  }

  // error
  cerr << "Frame Internal Error: mapLenFromRef" << endl;
  return v;
}

double Frame::mapLenFromRef(double d, CoordFormat format)
{
  Vector r = mapLenFromRef(Vector(d,d), format);
  return r[0];
}

Vector Frame::mapLenFromRef(const Vector& v, CoordFormat format)
{
  // really from image coords

  if (currentFits) {
    switch (format) {
    case PIXELS:
      return Vector(v).abs();
    case DEGREES:
      return currentFits->pix2wcsDist(v);
    case ARCMIN:
      return currentFits->pix2wcsDist(v*60);
    case ARCSEC:
      return currentFits->pix2wcsDist(v*60*60);
    }
  }

  // error
  cerr << "Frame Internal Error: mapLenFromRef" << endl;
  return v;
}

double Frame::mapLenFromRef(const Vector& vv1, const Vector& vv2, 
			    CoordFormat format)
{
  if (currentFits) {
    Vector v1 = vv1 * currentFits->getRefToImage();
    Vector v2 = vv2 * currentFits->getRefToImage();

    switch (format) {
    case PIXELS:
      return (v2-v1).length();
    case DEGREES:
      return wcsdist(v1,v2);
    case ARCMIN:
      return wcsdist(v1,v2)*60;
    case ARCSEC:
      return wcsdist(v1,v2)*60*60;
    }
  }

  // error
  cerr << "Frame Internal Error: mapLenFromRef" << endl;
  return 0;
}

double Frame::wcsdist(const Vector& v1, const Vector& v2)
{
  // v1 and v2 are in REF coords

  if (fits) {
    FitsImage* ptr = fits;
    while (ptr) {
      Vector img = v1 * ptr->getRefToData();
      if (img[0]>=0 && img[0]<ptr->width() && img[1]>=0 && 
	  img[1]<ptr->height()) {
	if (ptr->hasWCS()) {
	  Vector a = ptr->pix2wcs(v1 * ptr->getRefToImage());
	  Vector b = ptr->pix2wcs(v2 * ptr->getRefToImage());
	  return ptr->wcsdist(a,b);
	}
	else
	  return 0;
      }
      ptr = ptr->next();
    }
  }

  // if we get here, we have an error condition

  return 0;
}





