%{
#define YYPARSE_PARAM fr
#define FR ((Frame*)fr)
#define YYDEBUG 1

#define DISCARD_(x) {yyclearin; mkDiscard(x);}

#include <math.h>
#include <string.h>

#include <iostream.h>

#include "frame.h"
#include "util.h"
#include "vector.h"
#include "list.h"
#include "basemarker.h"

extern int mklex(void);
extern void mkerror(const char*);
extern void mkDiscard(int);

static Frame::CoordSystem globalSystem;

static unsigned short globalProps;
static unsigned short localProps;

static char globalColor[16];
static char localColor[16];

static char globalFont[32];
static char localFont[32];

static char globalText[80];
static char localText[80];

static int globalLine1;
static int localLine1;
static int globalLine2;
static int localLine2;

static int globalPoint;
static int localPoint;

static Frame::CoordFormat globalRuler;
static Frame::CoordFormat localRuler;

static List<Vertex>* polylist;

static double aAnnuli[MAXANNULI];
static Vector aVector[MAXANNULI];
static int aNum;
static int aStatus;
static Vector aCenter;
static double aAngle;
static unsigned short aProps;
static char aColor[16];
static char aFont[32];
static char aText[80];

static void setProps(unsigned short* props, unsigned short prop, int value);
static Frame::CoordSystem checkWCSSystem();

enum {CIRCLE,BOX,DIAMOND,CROSS,X,ARROW,BOXCIRCLE};
	
%}

%union {
  float real;
  int integer;
  char str[256];
  double vector[3];
}

%type <real> numeric
%type <integer> yesno

%type <real> angle
%type <real> optangle
%type <real> value
%type <real> sexagesimal
%type <vector> coord
%type <integer> property
%type <integer> coordFormat
%type <integer> coordSystem
%type <integer> pointProp
%type <integer> numberof

%token <integer> INT
%token <real> REAL
%token <str> STRING

%token <integer> HOUR
%token <integer> MINUTE
%token <real> SECOND
%token <real> DEGREE
%token <real> ARCMINUTE
%token <real> ARCSECOND
%token <real> RADIAN
%token <real> PHYCOORD
%token <real> IMGCOORD

%token <str> SEXSTR

%token AMPLIFIER_
%token ANNULUS_
%token ARCMIN_
%token ARCSEC_
%token ARROW_
%token B1950_
%token BOX_
%token BOXCIRCLE_
%token CIRCLE_
%token COLOR_
%token CROSS_
%token DEBUG_
%token DEGREES_
%token DELETE_
%token DETECTOR_
%token DIAMOND_
%token ECLIPTIC_
%token EDIT_
%token ELLIPSE_
%token FALSE_
%token FIELD_
%token FIXED_
%token FK4_
%token FK5_
%token FONT_
%token GALACTIC_
%token GLOBAL_
%token ICRS_
%token IMAGE_
%token INCLUDE_
%token J2000_
%token LINE_
%token LINEAR_
%token LOGICAL_
%token MOVE_
%token N_
%token NO_
%token OFF_
%token ON_
%token PANDA_
%token PHYSICAL_
%token PIE_
%token PIXELS_
%token POINT_
%token POLYGON_
%token ROTATE_
%token ROTBOX_
%token RULER_
%token SELECT_
%token TEXT_
%token TRUE_
%token VERSION_
%token X_
%token Y_
%token YES_

%%

start	: initGlobal commands
	;

commands: commands command terminator
	| command terminator
	;

command : /* empty */
	| DEBUG_ debug
	| VERSION_ {cerr << "DS9 Regions File 3.0" << endl;}
	| GLOBAL_ global comment
	| coordSystem {globalSystem = (Frame::CoordSystem)$1;} comment
	| initLocal include shape
	| generalComment
	;

generalComment: '#' {DISCARD_(0)} STRING
	;

comment : /* empty */
	| '#' {DISCARD_(1)} STRING
	;

shapeComment : /* empty */ processAnnulus
	| '#' {DISCARD_(1)} STRING processAnnulus
	| '#' local processAnnulus
	;

terminator: '\n'
	| ';'
	;

numeric	: REAL {$$=$1;}
	| INT {$$=$1;}
	;

debug	: ON_ {yydebug=1;}
	| OFF_ {yydebug=0;}
	;

yesno	: INT {$$=($1 ? 1 : 0);}

	| YES_ {$$=1;}
	| Y_ {$$=1;}
	| ON_ {$$=1;}
	| TRUE_ {$$=1;}

	| NO_ {$$=0;}
	| N_ {$$=0;}
	| OFF_ {$$=0;}
	| FALSE_ {$$=0;}
	;

sp	: /* empty */
	| ','
	;

bp	: /* empty */
	| '('
	;

ep	: /* emtpy */
	| ')'
	;

optangle: /* empty */ {$$ = 0;}
	| angle {$$ = $1;}
	;

angle	: numeric {$$ = degToRad($1);} /* assume degree */
	| DEGREE {$$ = degToRad($1);}
	| RADIAN {$$ = $1;}
	;

value	: numeric {$$ = FR->mapLenToRef($1, globalSystem);}
	| PHYCOORD
	  {$$ = FR->mapLenToRef($1, Frame::PHYSICAL);}
	| IMGCOORD
	  {$$ = FR->mapLenToRef($1, Frame::IMAGE);}
	| DEGREE
	  {$$ = FR->mapLenToRef($1, checkWCSSystem(), Frame::DEGREES);}
	| ARCMINUTE 
	  {$$ = FR->mapLenToRef($1, checkWCSSystem(), Frame::ARCMIN);}
	| ARCSECOND 
	  {$$ = FR->mapLenToRef($1, checkWCSSystem(), Frame::ARCSEC);}
	;

numberof: N_ '=' INT {$$ = $3;}
	;

sexagesimal: SEXSTR {$$ = parseDMS($1);}

coord	: sexagesimal sp sexagesimal
	{
	  Vector r;
	  if (globalSystem == Frame::GALACTIC ||
	      globalSystem == Frame::ECLIPTIC) 
	    r = FR->map(Vector($1,$3), checkWCSSystem(), Frame::REF);
	  else
	    r = FR->map(Vector($1*360./24.,$3), checkWCSSystem(), Frame::REF);

	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| HOUR MINUTE SECOND sp INT ARCMINUTE ARCSECOND
        {
	  Vector r = FR->map(
	    Vector(hmsToDegree($1,$2,$3), dmsToDegree($5,$6,$7)),
	    checkWCSSystem(), Frame::REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| INT ARCMINUTE ARCSECOND sp INT ARCMINUTE ARCSECOND
        {
	  Vector r = FR->map(
	    Vector(dmsToDegree($1,$2,$3), dmsToDegree($5,$6,$7)),
	    checkWCSSystem(), Frame::REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| numeric sp numeric 
	{
	  Vector r = FR->map(Vector($1,$3), globalSystem, Frame::REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| DEGREE sp DEGREE
	{
	  Vector r = FR->map(Vector($1,$3), checkWCSSystem(), Frame::REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| IMGCOORD sp IMGCOORD
	{
	  Vector r = FR->map(Vector($1,$3), Frame::IMAGE, Frame::REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| PHYCOORD sp PHYCOORD
	{
	  Vector r = FR->map(Vector($1,$3), Frame::PHYSICAL, Frame::REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	}
	| IMGCOORD sp PHYCOORD
	{
	  Vector rx = FR->map(Vector($1,$1), Frame::IMAGE, Frame::REF);
	  Vector ry = FR->map(Vector($3,$3), Frame::PHYSICAL, Frame::REF);
	  $$[0] = rx[0];
	  $$[1] = ry[0];
	  $$[2] = 1;
	}
	| PHYCOORD sp IMGCOORD
	{
	  Vector rx = FR->map(Vector($1,$1), Frame::PHYSICAL, Frame::REF);
	  Vector ry = FR->map(Vector($3,$3), Frame::IMAGE, Frame::REF);
	  $$[0] = rx[0];
	  $$[1] = ry[0];
	  $$[2] = 1;
	}
	;

coordSystem : IMAGE_ {$$ = Frame::IMAGE;}
	| LOGICAL_ {$$ = Frame::IMAGE;}
	| PHYSICAL_ {$$ = Frame::PHYSICAL;}
	| AMPLIFIER_ {$$ = Frame::AMPLIFIER;}
	| DETECTOR_ {$$ = Frame::DETECTOR;}
	| FK4_ {$$ = Frame::EQUATORIAL_FK4;}
	| B1950_  {$$ = Frame::EQUATORIAL_FK4;}
	| FK5_ {$$ = Frame::EQUATORIAL_FK5;}
	| J2000_  {$$ = Frame::EQUATORIAL_FK5;}	
	| ICRS_ {$$ = Frame::ICRS;}
	| GALACTIC_ {$$ = Frame::GALACTIC;}
	| ECLIPTIC_ {$$ = Frame::ECLIPTIC;}
	| LINEAR_ {$$ = Frame::LINEAR;}
	;

coordFormat : PIXELS_ {$$=Frame::PIXELS;}
	| DEGREES_ {$$=Frame::DEGREES;}
	| ARCMIN_ {$$=Frame::ARCMIN;}
	| ARCSEC_ {$$=Frame::ARCSEC;}
	;

property : SELECT_ {$$ = Marker::SELECT;}
	| EDIT_ {$$ = Marker::EDIT;}
	| MOVE_  {$$ = Marker::MOVE;}
	| ROTATE_ {$$ = Marker::ROTATE;}
	| DELETE_ {$$ = Marker::DELETE;}
	| INCLUDE_ {$$ = Marker::INCLUDE;}
	| FIXED_ {$$ = Marker::FIXED;}
	;

global	: global sp globalProperty
	| globalProperty
	;

globalProperty : property '=' yesno 
	{
	  setProps(&globalProps,$1,$3);
	  setProps(&localProps,$1,$3);
	}
	| COLOR_ '=' STRING 
	{
	  strncpy(globalColor,$3,16);
	  strncpy(localColor,$3,16);
	}
	| FONT_ '=' STRING 
	{
	  strncpy(globalFont,$3,32);
	  strncpy(localFont,$3,32);
	}
	| TEXT_ '=' STRING 
	{
	  strncpy(globalText,$3,80);
	  strncpy(localText,$3,80);
	}
	| POINT_ '=' pointProp 
	{
	  globalPoint = localPoint = $3;
	}
	| LINE_ '=' INT INT 
	{
	  globalLine1 = localLine1 = $3;
	  globalLine2 = localLine2 = $4;
	}
	| RULER_ '=' coordFormat 
	{
	  globalRuler = localRuler = (Frame::CoordFormat)$3;
	}
	;

initGlobal:{
	  // global properties
	  globalSystem = Frame::PHYSICAL;
	  globalProps =
	    Marker::SELECT | Marker::EDIT | Marker::MOVE |
	    Marker::ROTATE | Marker::DELETE | Marker::INCLUDE;
	  strcpy(globalColor,"green");
	  strcpy(globalFont,"helvetica 10 normal");
	  strcpy(globalText,"");

	  // unique properties
	  globalLine1 = 0;
	  globalLine2 = 0;
	  globalRuler = Frame::PIXELS;
	  globalPoint = BOXCIRCLE;
	} 
	;

local	: local sp localProperty
	| localProperty
	;

localProperty : property '=' yesno {setProps(&localProps,$1,$3);}
	| COLOR_ '=' STRING {strncpy(localColor,$3,16);}
	| FONT_ '=' STRING {strncpy(localFont,$3,32);}
	| TEXT_ '=' STRING {strncpy(localText,$3,80);}
	| POINT_ '=' pointProp {localPoint = $3;}
	| LINE_ '=' INT INT {localLine1=$3; localLine2=$4;}
	| RULER_ '=' coordFormat {localRuler = (Frame::CoordFormat)$3;}
	;

initLocal : {
	  // global properties
	  localProps = globalProps;
	  strcpy(localColor,globalColor);
	  strcpy(localFont,globalFont);
	  strcpy(localText,globalText);

	  // unique properties
	  localLine1 = globalLine1;
	  localLine2 = globalLine2;
	  localRuler = globalRuler;
	  localPoint = globalPoint;
	}
	;

pointProp : CIRCLE_ {$$ = CIRCLE;}
	| BOX_ {$$ = BOX;}
	| DIAMOND_ {$$ = DIAMOND;}
	| CROSS_ {$$ = CROSS;}
	| X_ {$$ = X;}
	| ARROW_ {$$ = ARROW;}
	| BOXCIRCLE_ {$$ = BOXCIRCLE;}
	;

include	: /* empty */ {setProps(&localProps, Marker::INCLUDE, 1);}
	| '+' {setProps(&localProps, Marker::INCLUDE, 1);}
	| '-' {setProps(&localProps, Marker::INCLUDE, 0);}
	;

shape	: CIRCLE_ bp coord sp value ep shapeComment
	  {FR->createCircleCmd(Vector($3),$5,
	    localColor,localFont,localText,localProps);}

	| ELLIPSE_ bp coord sp value sp value sp optangle ep shapeComment
	{
	  // for ellipse annulus
	  aStatus = 1;
	  aCenter = Vector($3);
	  aAngle = $9;
	  aVector[0] = Vector($5,$7);
	  aNum = 1;
	  strncpy(aColor,localColor,16);
	  strncpy(aFont,localFont,32);
	  strncpy(aText,localText,80);
	  aProps = localProps;

	  FR->createEllipseCmd(Vector($3),Vector($5,$7),$9,
	  localColor,localFont,localText,localProps);
	}

	| BOX_ bp coord sp value sp value sp optangle ep shapeComment
	{
	  // for box annulus
	  aStatus = 3;
	  aCenter = Vector($3);
	  aAngle = $9;
	  aVector[0] = Vector($5,$7);
	  aNum = 1;
	  strncpy(aColor,localColor,16);
	  strncpy(aFont,localFont,32);
	  strncpy(aText,localText,80);
	  aProps = localProps;

	  FR->createBoxCmd(Vector($3),Vector($5,$7),$9,
	    localColor,localFont,localText,localProps);
	}
	| ROTBOX_ bp coord sp value sp value sp optangle ep shapeComment
	  {FR->createBoxCmd(Vector($3),Vector($5,$7),$9,
	    localColor,localFont,localText,localProps);}

	| POLYGON_ {polylist=new List<Vertex>();} bp polyNodes ep shapeComment
	  {FR->createPolygonCmd(polylist, 
	    localColor,localFont,localText,localProps);} 

	| LINE_ bp coord sp coord ep shapeComment
	  {FR->createLineCmd(Vector($3),Vector($5),
	    localLine1,localLine2,localColor,localFont,localText,localProps);}

	| TEXT_ bp coord ep shapeComment
	  {FR->createTextCmd(Vector($3),
	    localColor,localFont,localText,localProps);}
	| TEXT_ bp coord sp STRING ep {strncpy(localText,$5,80);} shapeComment
	  {FR->createTextCmd(Vector($3),
	    localColor,localFont,localText,localProps);}

	| RULER_ bp coord sp coord ep shapeComment
	  {FR->createRulerCmd(Vector($3),Vector($5),
	    localRuler,localColor,localFont,localText,localProps);}

	| POINT_ bp coord ep shapeComment 
	{
	  switch (localPoint) {
	  case CIRCLE:
	    FR->createCirclePointCmd(Vector($3),
	      localColor,localFont,localText,localProps);
	    break;
	  case BOX:
	    FR->createBoxPointCmd(Vector($3),
	      localColor,localFont,localText,localProps);
	    break;
	  case DIAMOND:
	    FR->createDiamondPointCmd(Vector($3),
	      localColor,localFont,localText,localProps);
	    break;
	  case CROSS:
	    FR->createCrossPointCmd(Vector($3),
	      localColor,localFont,localText,localProps);
	    break;
	  case X:
	    FR->createExPointCmd(Vector($3),
	      localColor,localFont,localText,localProps);
	    break;
	  case ARROW:
	    FR->createArrowPointCmd(Vector($3),
	      localColor,localFont,localText,localProps);
	    break;
	  case BOXCIRCLE:
	    FR->createBoxCirclePointCmd(Vector($3),
	      localColor,localFont,localText,localProps);
	    break;
	  }
	}
	| CIRCLE_ POINT_ bp coord ep shapeComment
	  {FR->createCirclePointCmd(Vector($4),
	    localColor,localFont,localText,localProps);}
	| BOX_ POINT_ bp coord ep shapeComment
	  {FR->createBoxPointCmd(Vector($4),
	    localColor,localFont,localText,localProps);}
	| DIAMOND_ POINT_ bp coord ep shapeComment
	  {FR->createDiamondPointCmd(Vector($4),
	    localColor,localFont,localText,localProps);}
	| CROSS_ POINT_ bp coord ep shapeComment
	  {FR->createCrossPointCmd(Vector($4),
	    localColor,localFont,localText,localProps);}
	| X_ POINT_ bp coord ep shapeComment
	  {FR->createExPointCmd(Vector($4),
	    localColor,localFont,localText,localProps);}
	| ARROW_ POINT_ bp coord ep shapeComment
	  {FR->createArrowPointCmd(Vector($4),
	    localColor,localFont,localText,localProps);}
	| BOXCIRCLE_ POINT_ bp coord ep shapeComment
	  {FR->createBoxCirclePointCmd(Vector($4),
	    localColor,localFont,localText,localProps);}

	| ANNULUS_ bp coord sp value sp value ep shapeComment
	  {FR->createAnnulusCmd(Vector($3),$5,$7,1,
	    localColor,localFont,localText,localProps);}
	| ANNULUS_ bp coord sp value sp value sp {aNum=2;} aRads ep
	   shapeComment
	{
	  aAnnuli[0] = $5;
	  aAnnuli[1] = $7;
	  FR->createAnnulusCmd(Vector($3),aAnnuli,aNum,
	  localColor,localFont,localText,localProps);
	}
	| ANNULUS_ bp coord sp value sp value sp numberof ep shapeComment
	  {FR->createAnnulusCmd(Vector($3),$5,$7,$9,
	    localColor,localFont,localText,localProps);}

	| ELLIPSE_ ANNULUS_ bp coord sp angle sp
	    value sp value sp value sp value sp
	    {aNum=2;} vRads ep shapeComment
	{
	  aVector[0] = Vector($8,$10);
	  aVector[1] = Vector($12,$14);
	  FR->createEllipseAnnulusCmd(Vector($4),$6,aNum,
	  aVector,localColor,localFont,localText,localProps);
	}
	| ELLIPSE_ ANNULUS_ bp coord sp angle sp
	  value sp value sp value sp value sp
	  numberof ep shapeComment
	  {FR->createEllipseAnnulusCmd(Vector($4),
	    Vector($8,$10),Vector($12,$14),$16,$6,
	    localColor,localFont,localText,localProps);}
	| ELLIPSE_ bp coord sp value sp value sp optangle ep '&' '!' 
	  ELLIPSE_ bp coord sp value sp value sp optangle ep
	{	
	  aStatus = 2;
	  aVector[aNum++] = Vector($5,$7);
	}

	| BOX_ ANNULUS_ bp coord sp angle sp
	    value sp value sp value sp value sp
	    {aNum=2;} vRads ep shapeComment
	{
	  aVector[0] = Vector($8,$10);
	  aVector[1] = Vector($12,$14);
	  FR->createBoxAnnulusCmd(Vector($4),$6,aNum,
	  aVector,localColor,localFont,localText,localProps);
	}
	| BOX_ ANNULUS_ bp coord sp angle sp
	  value sp value sp value sp value sp
	  numberof ep shapeComment
	  {FR->createBoxAnnulusCmd(Vector($4),
	    Vector($8,$10),Vector($12,$14),$16,$6,
	    localColor,localFont,localText,localProps);}
	| BOX_ bp coord sp value sp value sp optangle ep '&' '!' 
	  BOX_ bp coord sp value sp value sp optangle ep
	{	
	  aStatus = 4;
	  aVector[aNum++] = Vector($5,$7);
	}

	| PIE_ bp coord sp angle sp angle ep shapeComment
	| PIE_ bp coord sp angle sp angle sp pieAngs ep shapeComment
	| PIE_ bp coord sp angle sp angle sp numberof ep shapeComment
	| FIELD_ bp ep shapeComment
	| PANDA_ bp coord sp angle sp angle sp INT sp 
	    numeric sp numeric sp INT ep shapeComment
	;

polyNodes : polyNodes sp polyNode
	| polyNode
	;

polyNode : coord {polylist->append(new Vertex($1));}
	;

aRads	: aRads sp aRad
	| aRad
	;

aRad	: value {aAnnuli[aNum++] = $1;}
	;

vRads	: vRads sp vRad
	| vRad
	;

vRad	: value sp value {aVector[aNum++] = Vector($1,$3);}
	;

pieAngs	: pieAngs sp pieAng
	| pieAng
	;

pieAng	: angle {}
	;

processAnnulus : /* empty */
	{
	  switch (aStatus) {
	  case 0: // do nothing
	    break;
	  case 1: // we found just an ellipse, do nothing
	    break;
	  case 2: // ok we have an ellipse annulus
	    FR->markerDeleteLastCmd(); // delete the previous ellipse
	    FR->createEllipseAnnulusCmd(aCenter,aAngle,aNum,aVector,
	      aColor,aFont,aText,aProps);
	    break;
	  case 3: // we found just a box, do nothing
	    break;
	  case 4: // ok, we have a box annulus
	    FR->markerDeleteLastCmd(); // delete the previous box
	    FR->createBoxAnnulusCmd(aCenter,aAngle,aNum,aVector,
	      aColor,aFont,aText,aProps);
	    break;
	  }
	  aStatus = 0;
	}

%%

static void setProps(unsigned short* props, unsigned short prop, int value)
{
  if (value)
    *props |= prop;
  else
    *props &= ~prop;
}

static Frame::CoordSystem checkWCSSystem()
{
  // if we have not defined a specific WCS, set to Native
  switch (globalSystem) {
    case (Frame::IMAGE):
    case (Frame::PHYSICAL):
    case (Frame::DETECTOR):
    case (Frame::AMPLIFIER):
      return (Frame::NATIVEWCS);
    default:
      return globalSystem;
  }
}
