extern "C" {
#include "grf.h"
}

#include <iostream.h>
#include "grid.h"

extern Grid* gr;

/*
 *  Name:
 *     astGFlush
 *
 *  Purpose:
 *     Flush all pending graphics to the output device
 *
 *  Synopsis:
 *     #include "grf.h"
 *     int astGFlush( void )
 *
 *  Description:
 *     This function ensures that the display device is up-to-date,
 *     by flushing any pending graphics to the output device.
 *
 *  Parameters:
 *     None
 *
 *  Returned Value:
 *     A value of 0 is returned if an error occurrs, and 1 is returned
 *     otherwise
 */

int astGFlush(void)
{
  gr->gFlush();
  return 1;
}

/*
 *  Name:
 *     astGLine
 *
 *  Purpose:
 *     Draw a polyline (i.e. a set of connected lines)
 *
 *  Synopsis:
 *     #include "grf.h"
 *     int astGLine( int n, const float *x, const float *y )
 *
 *  Description:
 *     This function displays lines joining the given positions
 *
 *  Parameters:
 *     n
 *        The number of positions to be joined together
 *     x 
 *        A pointer to an array holding the "n" x values
 *     y 
 *        A pointer to an array holding the "n" y values
 *
 *  Returned Value:
 *     A value of 0 is returned if an error occurrs, and 1 is returned
 *     otherwise
 *
 *  Notes:
 *     -  Nothing is done if "n" is less than 2, or if a NULL pointer is
 *     given for either "x" or "y"
 */

int astGLine(int n, const float *x, const float *y)
{
  gr->gLine(n, x, y);
  return 1;
}

/*
 *  Name:
 *     astGQch
 *
 *  Purpose:
 *     Return the character height in world cooridnates
 *
 *  Synopsis:
 *     #include "grf.h"
 *     int astGQch( float *chv, float *chh )
 *
 *  Description:
 *     This function returns the heights of characters drawn vertically and
 *     horizontally in world coordinates
 *
 *  Parameters:
 *     chv
 *        A pointer to the double which is to receive the height of
 *        characters drawn vertically. This will be an increment in the X
 *        axis
 *     chh
 *        A pointer to the double which is to receive the height of
 *        characters drawn vertically. This will be an increment in the Y
 *        axis
 *
 *  Returned Value:
 *     A value of 0 is returned if an error occurrs, and 1 is returned
 *     otherwise
 */

int astGQch(float *chv, float *chh)
{
  gr->gQch(chv, chh);
  return 1;
}

/*
 *  Name:
 *     astGMark
 *
 *  Purpose:
 *     Draw a set of markers
 *
 *  Synopsis:
 *     #include "grf.h"
 *     int astGMark( int n, const float *x, const float *y, int type )
 *
 *  Description:
 *     This function displays markers at the given positions
 *
 *  Parameters:
 *     n
 *        The number of markers to draw
 *     x 
 *        A pointer to an array holding the "n" x values
 *     y 
 *        A pointer to an array holding the "n" y values
 *     type
 *        An integer which can be used to indicate the type of marker symbol
 *        required
 *
 *  Returned Value:
 *     A value of 0 is returned if an error occurrs, and 1 is returned
 *     otherwise
 *
 *  Notes:
 *     -  Nothing is done if "n" is less than 1, or if a NULL pointer is
 *     given for either "x" or "y"
 *
 */

int astGMark(int n, const float *x, const float *y, int type)
{
  gr->gMark(n, x, y, type);
  return 1;
}

/*
 *  Name:
 *     astGText
 *
 *  Purpose:
 *     Draw a character string
 *
 *  Synopsis:
 *     #include "grf.h"
 *     int astGText( const char *text, float x, float y, const char *just,
 *                   float upx, float upy )
 *
 *  Description:
 *     This function displays a character string at a given position
 *     using a specified justification and up-vector
 *
 *  Parameters:
 *     text 
 *        Pointer to a null-terminated character string to be displayed
 *     x 
 *        The reference x coordinate
 *     y 
 *        The reference y coordinate
 *     just
 *        A character string which specifies the location within the
 *        text string which is to be placed at the reference position
 *        given by x and y. The first character may be 'T' for "top",
 *        'C' for "centre", or 'B' for "bottom", and specifies the
 *        vertical location of the reference position. Note, "bottom"
 *        corresponds to the base-line of normal text. Some characters 
 *        (eg "y", "g", "p", etc) descend below the base-line. The second 
 *        character may be 'L' for "left", 'C' for "centre", or 'R' 
 *        for "right", and specifies the horizontal location of the 
 *        reference position. If the string has less than 2 characters
 *        then 'C' is used for the missing characters
 *     upx
 *        The x component of the up-vector for the text, in graphics world
 *        coordinates. If necessary the supplied value should be negated
 *        to ensure that positive values always refer to displacements from 
 *        left to right on the screen
 *     upy
 *        The y component of the up-vector for the text, in graphics world
 *        coordinates. If necessary the supplied value should be negated
 *        to ensure that positive values always refer to displacements from 
 *        bottom to top on the screen
 *
 *  Returned Value:
 *     A value of 0 is returned if an error occurrs, and 1 is returned
 *     otherwise
 *
 *  Notes:
 *     -  Any graphics within the rotated box enclosing the text are erased
 *     -  A NULL value for "just" causes a value of "CC" to be used
 *     -  Both "upx" and "upy" being zero causes an error
 *     -  Any unrecognised character in "just" causes an error
 */

int astGText(const char *text, float x, float y, const char *just,
	     float upx, float upy)
{
  gr->gText(text, x ,y, just, upx, upy);
  return 1;
}               

/*
*  Name:
*     astGTxExt
*
*  Purpose:
*     Get the extent of a character string
*
*  Synopsis:
*     #include "grf.h"
*     int astGTxExt( const char *text, float x, float y, const char *just,
*                   float upx, float upy, float *xb, float *yb )
*
*  Description:
*     This function returns the corners of a box which would enclose the 
*     supplied character string if it were displayed using astGText
*
*     The returned box INCLUDES any leading or trailing spaces
*
*  Parameters:
*     text 
*        Pointer to a null-terminated character string to be displayed
*     x 
*        The reference x coordinate
*     y 
*        The reference y coordinate
*     just
*        A character string which specifies the location within the
*        text string which is to be placed at the reference position
*        given by x and y. The first character may be 'T' for "top",
*        'C' for "centre", or 'B' for "bottom", and specifies the
*        vertical location of the reference position. Note, "bottom"
*        corresponds to the base-line of normal text. Some characters 
*        (eg "y", "g", "p", etc) descend below the base-line. The second 
*        character may be 'L' for "left", 'C' for "centre", or 'R' 
*        for "right", and specifies the horizontal location of the 
*        reference position. If the string has less than 2 characters
*        then 'C' is used for the missing characters 
*     upx
*        The x component of the up-vector for the text, in graphics world
*        coordinates. If necessary the supplied value should be negated
*        to ensure that positive values always refer to displacements from 
*        left to right on the screen
*     upy
*        The y component of the up-vector for the text, in graphics world
*        coordinates. If necessary the supplied value should be negated
*        to ensure that positive values always refer to displacements from 
*        bottom to top on the screen
*     xb
*        An array of 4 elements in which to return the x coordinate of
*        each corner of the bounding box
*     yb
*        An array of 4 elements in which to return the y coordinate of
*        each corner of the bounding box
*
*  Returned Value:
*     A value of 0 is returned if an error occurrs, and 1 is returned
*     otherwise
*
*  Notes:
*     -  The order of the corners is anti-clockwise (in world coordinates)
*        starting at the bottom left
*     -  A NULL value for "just" causes a value of "CC" to be used
*     -  Both "upx" and "upy" being zero causes an error
*     -  Any unrecognised character in "just" causes an error
*     -  Zero is returned for all bounds of the box if an error occurs
*/

int astGTxExt(const char *text, float x, float y, const char *just,
	      float upx, float upy, float *xb, float *yb)
{
  gr->gTxExt(text, x, y, just, upx, upy, xb, yb);
  return 1;
}               

/*
 *  Name:
 *     astGAttr
 *
 *  Purpose:
 *     Enquire or set a graphics attribute value
 *
 *  Synopsis:
 *     #include "grf.h"
 *     int int astGAttr( int attr, double value, double *old_value, int prim )
 *
 *  Description:
 *     This function returns the current value of a specified graphics
 *     attribute, and optionally establishes a new value. The supplied
 *     value is converted to an integer value if necessary before use
 *
 *  Parameters:
 *     attr
 *        An integer value identifying the required attribute. The
 *        following symbolic values are defined in grf.h: 
 *
 *           GRF__STYLE  - Line style
 *           GRF__WIDTH  - Line width
 *           GRF__SIZE   - Character and marker size scale factor
 *           GRF__FONT   - Character font
 *           GRF__COLOUR - Colour index
 *     value 
 *        A new value to store for the attribute. If this is AST__BAD
 *        no value is stored
 *     old_value 
 *        A pointer to a double in which to return the attribute value
 *        If this is NULL, no value is returned
 *     prim
 *        The sort of graphics primative to be drawn with the new attribute
 *        Identified by the following values defined in grf.h:
 *           GRF__LINE
 *           GRF__MARK
 *           GRF__TEXT
 *
 *  Returned Value:
 *     A value of 0 is returned if an error occurrs, and 1 is returned
 *     otherwise
 *
 *  Notes:
 */

int astGAttr(int attr, double value, double *old, int prim)
{
  gr->gAttr(attr, value, old, prim);
  return 1;
}
