/*
 *      Copyrights:
 *
 *    	  Copyright (c) 1996, 1998 Smithsonian Astrophysical Observatory
 *
 *        Permission to use, copy, modify, distribute, and  sell  this
 *        software  and  its  documentation  for any purpose is hereby
 *        granted without  fee,  provided  that  the  above  copyright
 *        notice  appear  in  all  copies and that both that copyright
 *        notice and this permission notice appear in supporting docu-
 *        mentation,  and  that  the  name  of the  Smithsonian Astro-
 *	  physical Observatory not be used in advertising or publicity
 *  	  pertaining to distribution of the software without specific,
 *	  written  prior  permission.   The Smithsonian  Astrophysical 
 *        Observatory makes no representations about  the  suitability
 *	  of  this  software for any purpose.  It is provided  "as is"
 *	  without express or implied warranty.
 *        THE  SMITHSONIAN  ASTROPHYSICAL  OBSERVATORY  DISCLAIMS  ALL
 *	  WARRANTIES  WITH  REGARD  TO  THIS  SOFTWARE,  INCLUDING ALL
 *	  IMPLIED  WARRANTIES  OF  MERCHANTABILITY AND FITNESS, IN  NO
 *	  EVENT SHALL THE  SMITHSONIAN  ASTROPHYSICAL  OBSERVATORY  BE 
 *	  LIABLE FOR  ANY SPECIAL, INDIRECT  OR  CONSEQUENTIAL DAMAGES
 *	  OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS  OF USE,  DATA
 *	  OR  PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 *	  OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
 *	  THE  USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <xpap.h>

/*
 *----------------------------------------------------------------------------
 *
 *
 * 			Private Routines and Data
 *
 *
 *----------------------------------------------------------------------------
 */

/* this is the head of the global list of client xpas */
static XPA xpaclienthead=NULL;

static char errbuf[SZ_LINE];		/* holds current error message */
static int id=0;			/* id of current command */

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientNewInput
 *
 * Purpose:	allocate a new input struct for reading data from stdin
 *
 * Return:	input struct, or NULL on error
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static XPAInput
XPAClientNewInput(XPA xpa)
#else
static XPAInput XPAClientNewInput(xpa)
    XPA xpa;
#endif
{
  XPAInput xnew, inp;

  /* allocate a new record */
  if( (xnew=(XPAInput)calloc(1, sizeof(XPAInputRec))) == NULL ){
    return(NULL);
  }
  /* allocate the data buffer */
  xnew->buf = (char *)malloc(XPA_BIOSIZE);
  /* this buffer starts (and currently ends) at the current byte count */
  xnew->start = xpa->inpbytes;
  xnew->end = xpa->inpbytes;
  xnew->bytes = 0;

  /* add this input to end of list of input's */
  if( xpa->inphead == NULL ){
    xpa->inphead = xnew;
  }
  else{
    for(inp=xpa->inphead; inp->next!=NULL; inp=inp->next)
      ;
    inp->next = xnew;
  }

  /* return the record struct */
  return(xnew);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientFreeInput
 *
 * Purpose:	free a input buffer once its been sent to all targets
 *
 * Return:	0 on success, -1 on failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void
XPAClientFreeInput (XPA xpa, XPAInput inp)
#else
static void XPAClientFreeInput(xpa, inp)
     XPA xpa;
     XPAInput inp;
#endif
{
  XPAInput cur;

  if( !xpa || !inp )
    return;

  if( inp == xpa->inphead ){
    xpa->inphead = inp->next;
  }
  else{
    for(cur=xpa->inphead; cur!=NULL; cur=cur->next){
      if( cur->next == inp ){
	cur->next = inp->next;
	break;
      }
    }
  }

  /* free current record */
  if( inp != NULL ){ 
    if( inp->buf != NULL )
      free(inp->buf );
    free(inp);
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientFreeAllInputs
 *
 * Purpose:	free remaining input buffers
 *
 * Return:	0 on success, -1 on failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void
XPAClientFreeAllInputs (XPA xpa)
#else
static void XPAClientFreeAllInputs(xpa)
     XPA xpa;
#endif
{
  XPAInput cur, tmp;

  if( !xpa )
    return;

  for(cur=xpa->inphead; cur!=NULL; ){
    tmp = cur->next;
    XPAClientFreeInput(xpa, cur);
    cur = tmp;
  }
  xpa->inpbytes = 0;
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientProcessInput
 *
 * Purpose:	read input from stdin and store in an input struct
 *
 * Return:	bytes read
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int
XPAClientProcessInput(XPA xpa)
#else
static int XPAClientProcessInput(xpa)
    XPA xpa;
#endif
{
  static XPAInput cur=NULL;
  int get, got;

  /* set up next buffer, if necessary */
  for(cur=xpa->inphead; cur!=NULL; cur=cur->next){
    if( cur->bytes < XPA_BIOSIZE )
      break;
  }
  if( cur == NULL ){
    cur = XPAClientNewInput(xpa);
  }

  /* read data from stdin */
  get = MIN(XPA_IOSIZE, XPA_BIOSIZE - cur->bytes);
  if( !isatty(xpa->ifd) )
    got = read(xpa->ifd, &(cur->buf[cur->bytes]), get);
  else
    got = XPARdl(xpa->ifd, &(cur->buf[cur->bytes]), get, -1);
  switch(got){
  case -1:
    if( XPAVerbosity() ){
      perror("XPA client read");
    }
    return(0);
  case 0:
    xpa->ifd = -1;
    break;
  default:
    break;
  }
  cur->bytes += got;
  cur->end += got;
  xpa->inpbytes +=  got;

  /* verify to stdout, if necessary */
  if( xpa->client_mode & XPA_CLIENT_VERIFY ){
    fwrite(&(cur->buf[cur->bytes-got]), sizeof(char), got, stdout);
  }

  /* return the number of bytes just read */
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientFree
 *
 * Purpose:	free a client record and remove from list
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void 
XPAClientFree (XPA xpa, XPAClient client)
#else
static void XPAClientFree(xpa, client)
     XPA xpa;
     XPAClient client;
#endif
{
  XPAClient cur;

  /* remove from list of xpa's */
  if( xpa->clienthead ){
    if( xpa->clienthead == client ){
      xpa->clienthead = client->next;
    }
    else{
      for(cur=xpa->clienthead; cur!=NULL; cur=cur->next){
	if( cur->next == client ){
	  cur->next = client->next;
	  break;
	}
      }
    }
  }

  /* close communication channel */
  if( client->cmdfd >=0 ){
    fclose(client->cmdin);
    fclose(client->cmdout);
    close(client->cmdfd);
  }
  if( client->datafd >= 0 )
    close(client->datafd);
  if( client->method )
    free(client->method);
  if( client->xtemplate )
    free(client->xtemplate);
  if( client->xclass )
    free(client->xclass);
  if( client->name )
    free(client->name);
  if( client->id )
    free(client->id);
  /* xpaget's fd mode has an alloc'ed bufptr and lenptr */
  if( (client->type == 'g') && (client->mode & XPA_CLIENT_FD) ){
    if( client->bufptr && *(client->bufptr) )
      free(*(client->bufptr));
    if( client->bufptr )
      free(client->bufptr);
    if( client->lenptr )
      free(client->lenptr);
  }
  free(client);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientEnd
 *
 * Purpose:	finish up with this client
 *
 * Returns:	error message or null
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static char *
XPAClientEnd (XPA xpa, XPAClient client)
#else
static char *XPAClientEnd(xpa, client)
     XPA xpa;
     XPAClient client;
#endif
{
  char *error=NULL;
  char *eptr;

  /* close the data channel, which should trigger a result from the server */
  if( client->datafd >= 0 ){
    close(client->datafd);
    client->datafd = -1;
  }

  /* this is not really active any longer */
  client->active = 1;

  /* read the status line */
  if( client->mode & XPA_CLIENT_ACK ){
retry:
    if( XPAGets(client->cmdin, errbuf, SZ_LINE, XPALongTimeout()) >0 ){
      eptr = errbuf;
      /* this should never happen */
      if( *eptr == '?' ){
	sprintf(errbuf, "XPA$WARNING: Protocol mismatch - missing id\n%s",
		eptr);
	error = NULL;
      }
      else{
	/* make sure we are dealing with a proper message */
	if( strncmp(eptr, client->id, strlen(client->id)) ){
	  if( XPAVerbosity() > 1 ){
	    fprintf(stderr,
		    "XPA$WARNING: ignoring out of sync server message:\n");
	    fprintf(stderr, "%s", errbuf);
	  }
	  goto retry;
	}
	/* go past id */
	eptr += strlen(client->id);
	while( isspace(*eptr) ) eptr++;
	if( !strncmp(eptr, "XPA$OK", 6) ){
	  error = NULL;
	}
	else{
	  error = eptr;
	}
      }
    }
    else{
      if( XPAVerbosity() > 1 ){
	fprintf(stderr,
		"XPA$WARNING: no reply from server callback (assuming OK)\n");
      }
      error = NULL;
    }
  }
  else
    error = NULL;

  /* store the error return */
  if( client->errptr )
    *(client->errptr) = NewString(error);

  /* remove this client if we are not meant to persist */
  if( !xpa->persist ){
    XPAClientFree(xpa, client);
  }
  /* otherwise mark as inactive */
  else{
    client->active = 0;
    client->bytes = 0;
  }

  /* return error status */
  return(error);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientNew
 *
 * Purpose:	allocate a new xpa client
 *
 * Returns:	xpa client struct
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static XPAClient 
XPAClientNew (XPA xpa, char *xtemplate, int type, char *xclass, char *name,
	      char *method)
#else
static XPAClient XPAClientNew(xpa, xtemplate, type, xclass, name, method)
     XPA xpa;
     char *xtemplate;
     int type;
     char *xclass;
     char *name;
     char *method;
#endif
{
  XPAClient xnew, client;
  struct sockaddr_in sock_in;
  struct sockaddr_un sock_un;
  unsigned short port;
  unsigned int ip;
  int fd;
  int tries;
  int keep_alive=1;

  /* allocate new send record */
  if( (xnew=(XPAClient)calloc(1, sizeof(XPAClientRec))) == NULL )
    return(NULL);

  /* mark sockets with impossible values (over-ridden later on) */
  xnew->cmdfd = -1;
  xnew->datafd = -1;
  
  switch(XPAMethod(method)){
  case XPA_INET:
    XPAParseIpPort(method, &ip, &port);
    /* connect to the server before we go further */
    if( (fd = socket(AF_INET, SOCK_STREAM, 0)) < 0 ){
      free(xnew);
      return(NULL);
    }
    setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE,
	       (char *)&keep_alive, sizeof(keep_alive));
    memset((char *)&sock_in, 0, sizeof(sock_in));
    sock_in.sin_family = AF_INET;
    sock_in.sin_addr.s_addr = htonl(ip);
    sock_in.sin_port = htons(port);
    /* make the connection with the server */
    if( connect(fd, (struct sockaddr *)&sock_in, sizeof(sock_in)) <0 ){
      close(fd);
      free(xnew);
      return(NULL);
    }
    /* make sure we close on exec */
    fcntl(fd, F_SETFD, FD_CLOEXEC);
    break;
  case XPA_UNIX:
    tries = 0;
again:
    /* open a socket and fill in socket information */
    if( (fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 ){
      free(xnew);
      return(NULL);
    }
    setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE,
	       (char *)&keep_alive, sizeof(keep_alive));
    memset((char *)&sock_un, 0, sizeof(sock_un));
    sock_un.sun_family = AF_UNIX;
    strcpy(sock_un.sun_path, method);
    /* make the connection with the server */
    if( connect(fd, (struct sockaddr *)&sock_un, sizeof(sock_un)) <0 ){
      /* Unix sockets get ECONNREFUSED when the listen queue is full,
	 so we try a few times to give the server a chance to recover */
      close(fd);
      if( (errno == ECONNREFUSED) && (tries < XPA_RETRIES) ){
	tries++;
	XPASleep(10);
	goto again;
      }
      /* give up */
      else{
	free(xnew);
	return(NULL);
      }
    }
    /* make sure we close on exec */
    fcntl(fd, F_SETFD, FD_CLOEXEC);
    break;
  default:
    return(NULL);
  }

  /* fill in the blanks */
  xnew->xtemplate = NewString(xtemplate);
  xnew->type = type;
  xnew->cmdfd = fd;
  xnew->cmdin = fdopen(fd, "r");
  xnew->cmdout = fdopen(fd, "w");
  xnew->xclass = NewString(xclass);
  xnew->name = NewString(name);
  xnew->method = NewString(method);
  xnew->active = 1;

  /* now that we have a valid client, add to list */
  if( xpa->clienthead == NULL ){
    xpa->clienthead = xnew;
  }
  else{
    for(client=xpa->clienthead; client->next!=NULL; client=client->next)
      ;
    client->next = xnew;
  }

#ifdef XPA_DEBUG
  fprintf(stderr, "XPAClientNew %d\n", xnew->cmdfd);
#endif

  /* return the good news */
  return(xnew);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientAddSelect
 *
 * Purpose:	add one or more xpa client sockets to the select flags
 *
 * Return:	number of clients that were added to the select flags
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int 
XPAClientAddSelect (XPA xpa, fd_set *readfdsptr, fd_set *writefdsptr)
#else
static int XPAClientAddSelect(xpa, readfdsptr, writefdsptr)
     XPA xpa;
     fd_set *readfdsptr;
     fd_set *writefdsptr;
#endif
{
  XPAClient client;
  int got=0;

  /* better have some place to set the flags */
  if( readfdsptr == NULL )
    return(0);

  /* add stdin, if there is one */
  if( xpa->ifd >= 0 )
    FD_SET(xpa->ifd, readfdsptr);

  /* set select flags for all clients */
  for(client=xpa->clienthead; client!=NULL; client=client->next){
    /* if this client is active */
    if( (client->active>1)  && (client->datafd >=0) ){
      if( client->type == 'g' ){
	FD_SET(client->datafd, readfdsptr);
	got++;
      }
      else if( client->type == 's' ){
	FD_SET(client->datafd, writefdsptr);
	got++;
      }
    }
  }
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientGet
 *
 * Purpose:	process an xpaget request for a given client
 *
 * Return:	0 on success, -1 on failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int 
XPAClientGet (XPA xpa, XPAClient client)
#else
static int XPAClientGet(xpa, client)
     XPA xpa;
     XPAClient client;
#endif
{
  char *s;
  int status;
  char tbuf[SZ_LINE];

  /* now retrieve the data from the server */
  status = XPAGetBuf(client->datafd, client->bufptr, client->lenptr,
		     XPAShortTimeout());
  /* status < 0 means error */
  switch(status){
  /* error */
  case -1:
    /* clean up after error */
    if( *(client->bufptr) ){
      free(*(client->bufptr));
      *(client->bufptr) = NULL;
    }
    *(client->lenptr) = 0;
    s = XPAClientEnd(xpa, client);
    /* we need to flag some sort of error, if nothing came across */
    if( *(client->errptr) == NULL ){
      *(client->errptr) = NewString("XPA$ERROR: incomplete transmission from server\n");
    }
    break;
  /* eof */
  case 0:
    /* if we have multiple clients, we now need to write this one */
    if( client->mode & XPA_CLIENT_FD ){
      if( xpa->nclient > 1 ){
	sprintf(tbuf, "XPA$BEGIN %s:%s %s\n",
		client->xclass, client->name, client->method);
	write(client->fd, tbuf, strlen(tbuf));
      }
      write(client->fd, *(client->bufptr), *(client->lenptr));
      if( xpa->nclient > 1 ){
	sprintf(tbuf, "XPA$END %s:%s %s\n",
		client->xclass, client->name, client->method);
	write(client->fd, tbuf, strlen(tbuf));
      }
      /* we can free buf, since its not being passed back */
      if( *(client->bufptr) ){
	free(*(client->bufptr));
	*(client->bufptr) = NULL;
      }
    }
    /* for all clients, we need to clean up */
    s = XPAClientEnd(xpa, client);
    break;
  /* status > 0: bytes read, then wouldblock */
  case 1:
    /* for single client fd mode, we write immediately */
    /* this deals with the important case of one client with a large
       amount of data */
    if( (client->mode & XPA_CLIENT_FD) && (xpa->nclient == 1) ){
      write(client->fd, *(client->bufptr), *(client->lenptr));
      /* reset buf for next read */
      if( *(client->bufptr) )
	free(*(client->bufptr));
      *(client->bufptr) = NULL;
      *(client->lenptr) = 0;
    }
  break;
  }
  return(status);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientSet
 *
 * Purpose:	process an xpaset request for a given client
 *
 * Return:	0 on success, -1 on failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int 
XPAClientSet (XPA xpa, XPAClient client)
#else
static int XPAClientSet(xpa, client)
     XPA xpa;
     XPAClient client;
#endif
{
  int len;
  int got;
  int status;
  int total;
  int left;
  char *s;
  XPAInput inp;

  if( client->mode & XPA_CLIENT_BUF ){
    while( 1 ){
      len = MIN(XPA_IOSIZE, client->len - client->bytes);
      /* see if we have written it all */
      if( len == 0 ){
	status = 1;
	goto done;
      }
      /* write the next chunk */
      got=write(client->datafd, &(client->buf[client->bytes]), len);
      if( got > 0 ){
	client->bytes += got;
      }
      /* this should not happen -- treat as an error */
      else if( got == 0 ){
	status = -1;
	goto done;
      }
      else{
	/* check for error */
	if( errno != EWOULDBLOCK ){
	  status = -1;
	  goto done;
	}
	/* write would block, so we return and wait the server */
	else{
	  return(0);
	}
      }
    }
  }
  /* reading from stdin and writing to servers */
  else{
    /* find the input buf that contains the data we need */
    for(inp=xpa->inphead; inp!=NULL; inp=inp->next){
      if( (client->bytes >= inp->start) && (client->bytes < inp->end) ){
	break;
      }
    }
    /* if we can't find a buffer ... */
    if( !inp ){
      /*  ... and we have all the input, we are done */
      if( xpa->ifd < 0 ){
	status = 1;
	goto done;
      }
      /* ... but there is more input to come, return */
      else{
	return(0);
      }
    }
    /* optimization: don't write a buffer until its full (or until eof) */
    if( (xpa->ifd >=0) && (inp->bytes < XPA_BIOSIZE) ){
      return(0);
    }

    /* write bytes until we would block or until end of this buffer, etc */
    while( 1 ){
      len = MIN(XPA_IOSIZE, inp->end - client->bytes);
      /* if we are done with this buffer, just return */
      if( (xpa->status == XPA_CLIENT_START) && (len <=0) ){
	/* see if everyone else is done with this buffer as well,
	   in which case we can free it */
	left = 0;
	for(client=xpa->clienthead; client!=NULL; client=client->next){
	  if( (client->type != 's') || !(client->mode & XPA_CLIENT_FD) )
	    continue;
	  /* in order to be totally written out, the following must be true:
	   * 1. send->bytes must be past the end of this buffer
	   *  and
	   * 2. this buffer must be filled or else we hit eof
	   */
	  if( (client->bytes >= inp->end) &&
	      ((inp->bytes >= XPA_BIOSIZE) || (xpa->ifd < 0)) ){
	    /* buffer complete written */
	    ;
	  }
	  else{
	    /* buffer not complete written */
	    left++;
	    break;
	  }
	}
	/* if nothing is left, we can free this input struct */
	if( !left ){
	  XPAClientFreeInput(xpa, inp);
	}
	return(0);
      }
      /* write to the server */
      got = write(client->datafd, &(inp->buf[client->bytes-inp->start]), len);

      /* check for success */
      if( got > 0 ){
	/* update the number of bytes we wrote */
	client->bytes += got;
	/* go back for more */
	continue;
      }
      /* check for error */
      else{
	/* anything but a "would block" error is bad */
	if( errno != EWOULDBLOCK ){
	  status = -1;
	  goto done;
	}
	/* write would block, so we return and wait the server */
	else
	  return(0);
      }
    }
  }

done:
  /* nothing more to send */
  s = XPAClientEnd(xpa, client);
  return(status);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientProcessSelect
 *
 * Purpose:	process xpas that have pending reads or writes
 *
 * Return:	number of xpas processed
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int 
XPAClientProcessSelect (XPA xpa,
			fd_set *readfdsptr, fd_set *writefdsptr, int maxreq)
#else
static int XPAClientProcessSelect(xpa, readfdsptr, writefdsptr, maxreq)
     XPA xpa;
     fd_set *readfdsptr;
     fd_set *writefdsptr;
     int maxreq;
#endif
{
  int got=0;
  XPAClient client, tclient;

  /* <= 0 means do all of them */
  if( maxreq < 0 ){
    maxreq = 0;
  }
  /* first process any new input before we write output */
  if( (xpa->ifd >= 0) && FD_ISSET(xpa->ifd, readfdsptr) ){
    FD_CLR(xpa->ifd, readfdsptr);
    XPAClientProcessInput(xpa);
  }
  /* look at all clients */
  for(client=xpa->clienthead; client!=NULL; ){
    tclient = client->next;
    if( (client->active>1)  && (client->datafd >=0) ){
      /* then handle new requests */
      if((client->type == 'g') && FD_ISSET(client->datafd, readfdsptr)){
	FD_CLR(client->datafd, readfdsptr);
	/* the get always is completed in one go */
	XPAClientGet(xpa, client);
	got++;
	if( maxreq && (got >= maxreq) ){
	  return(got);
	}
      }
      else if((client->type == 's') && FD_ISSET(client->datafd, writefdsptr)){
	FD_CLR(client->datafd, writefdsptr);
	/* if the return is > 0, we completed the send */
	if( XPAClientSet(xpa, client) > 0 )
	  got++;
	if( maxreq && (got >= maxreq) ){
	  return(got);
	}
      }
    }
    client = tclient;
  }
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientPoll
 *
 * Purpose:	non-blocking handling of XPA access points
 *		timeout in millisecs, but if negative, no timeout is used
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int
XPAClientPoll (XPA xpa, int msec, int maxreq)
#else
static int XPAClientPoll(xpa, msec, maxreq)
     XPA xpa;
     int msec;
     int maxreq;
#endif
{
  int got=0;
  int sgot;
  fd_set readfds;
  fd_set writefds;
  struct rlimit rlp;
  struct timeval tv;
  struct timeval *tvp;
  static int width=0;

  if( width == 0 ){
    getrlimit(RLIMIT_NOFILE, &rlp);
    width = rlp.rlim_cur;
  }
  if( msec >= 0 ){
    tv.tv_sec = msec / 1000;
    tv.tv_usec = (msec % 1000) * 1000;
    tvp = &tv;
  }
  /* negative value means just block and wait */
  else{
    tvp = NULL;
  }
  FD_ZERO(&readfds);
  FD_ZERO(&writefds);
  if( XPAClientAddSelect(xpa, &readfds, &writefds) ||
      XPAAddSelect(NULL, &readfds) ){
    sgot = select(width, &readfds, &writefds, NULL, tvp);
    /* error -- what should we do? */
    if( sgot < 0 ){
      if( XPAVerbosity() ){
	perror("XPAClientPoll() select");
      }
      exit(1);
    }
    /* timeout -- just return */
    else if( sgot == 0 )
      ;
    /* finally ... something to do */
    else{
      /* process the servers first, because the client processing itself
	 might process an outstanding server request, which would cause
	 big problems */
      got += XPAProcessSelect(&readfds, maxreq-got);
      if( maxreq && (got >= maxreq) ){
	return(got);
      }
      got += XPAClientProcessSelect(xpa, &readfds, &writefds, maxreq-got);
    }
  }
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientLoop
 *
 * Purpose:	non-X programs event loop for handling XPA client events
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int
XPAClientLoop (XPA xpa)
#else
static int XPAClientLoop(xpa)
     XPA xpa;
#endif
{
  int got=0;
  int sgot;
  int ltimeout;
  fd_set readfds;
  fd_set writefds;
  struct rlimit rlp;
  static int width=0;
  struct timeval tv;
  struct timeval *tvp;

  if( width == 0 ){
    getrlimit(RLIMIT_NOFILE, &rlp);
    width = rlp.rlim_cur;
  }
  ltimeout = XPALongTimeout();
  FD_ZERO(&readfds);
  FD_ZERO(&writefds);
  while( XPAClientAddSelect(xpa, &readfds, &writefds) ){
    /* add other XPA's and process them as we process the client */
    XPAAddSelect(NULL, &readfds);
    /* hopefully, a server will respond in a finite amount of time */
    if( ltimeout > 0 ){
      tv.tv_sec = ltimeout;
      tv.tv_usec = 0;
      tvp = &tv;
    }
    /* wait forever, if necessary */
    else{
      tvp = NULL;
    }
    /* wait for a server to respond */
    sgot = select(width, &readfds, &writefds, NULL, tvp);
    /* error -- what should we do? */
    if( sgot < 0 ){
      if( XPAVerbosity() ){
	perror("XPAClientLoop() select");
      }
      exit(1);
    }
    /* timed out -- no one responded */
    else if( sgot == 0 ){
      break;
    }
    else{
      /* process the servers first, because the client processing itself
	 might process an outstanding server request, which would cause
	 big problems */
      got += XPAProcessSelect(&readfds, 0);
      got += XPAClientProcessSelect(xpa, &readfds, &writefds, 0);
    }
    FD_ZERO(&readfds);
    FD_ZERO(&writefds);
  }
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientConnect
 *
 * Purpose:	go to name service and get new names, merge with old,
 *		and connect to servers
 *
 * Returns:	number of connections
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int 
XPAClientConnect (XPA xpa, char *xtemplate, int type)
#else
static int XPAClientConnect(xpa, xtemplate, type)
     XPA xpa;
     char *xtemplate;
     int type;
#endif
{
  int i;
  int n;
  int got;
  int total=0;
  char *s;
  char **xclasses;
  char **names;
  char **methods;
  char xtype[2];
  XPAClient client;

  /* do some initialization */
  XPAInitEnv();

  /* make a string out of the type for the lookup */
  xtype[0] = type;
  xtype[1] = '\0';

  /* reset the number of clients we are processing */
  xpa->nclient = 0;

  /* this xpa is active */
  xpa->status = XPA_CLIENT_ACTIVE;

  /* get ips and ports for this name and type */
  n = XPANSLookup(xtemplate, xtype, &xclasses, &names, &methods);
  /* mark existing clients who do not match this template */
  for(got=0, client=xpa->clienthead; client !=NULL; client=client->next){
    for(i=0; i<n; i++){
      if( (client->type  == type)			&&
	  (!strcmp(client->xclass, xclasses[i]))	&&
	  (!strcmp(client->name, names[i]))		&&
	  (!strcmp(client->method, methods[i]))		){
	got++;
      }
    }
    /* don't unmark if its a different type -- someone else might be active */
    if( !got && (client->type == type) ){
      client->active = 0;
    }
  }
  /* add new clients for this type */
  for(i=0; i<n; i++){
    for(got=0, client=xpa->clienthead; client !=NULL; client=client->next){
      if( (client->type  == type)			&&
	  (!strcmp(client->xclass, xclasses[i]))	&&
	  (!strcmp(client->name, names[i]))		&&
	  (!strcmp(client->method, methods[i]))		){
	/* might have to change the template */
	if( strcmp(client->xtemplate, xtemplate) ){
	  free(client->xtemplate);
	  client->xtemplate = NewString(xtemplate);
	}
	client->active = 1;
	got++;
	total++;
	break;
      }
    }
    if( !got ){
      if( XPAClientNew(xpa, xtemplate, type,
		       xclasses[i], names[i], methods[i]) )
	total++;
    }
    /* done with these strings */
    free(xclasses[i]);
    free(names[i]);
    free(methods[i]);
  }
  /* free up arrays alloc'ed by names server */
  if( n > 0 ){
    free(xclasses);
    free(names);
    free(methods);
  }
  return(total);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClientStart
 *
 * Purpose:	send init string to server and perform other authentication
 *		tasks
 *
 * Returns:	0 if success, -1 otherwise
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int 
XPAClientStart (XPA xpa, XPAClient client, char *paramlist, char *mode)
#else
static int XPAClientStart(xpa, client, paramlist, mode)
     XPA xpa;
     XPAClient client;
     char *paramlist;
     char *mode;
#endif
{
  XPAClient tclient;
  unsigned short port;
  unsigned int ip;
  int fd;
  int tries;
  int flags;
  int keep_alive=1;
  char method[SZ_LINE];
  char tbuf[SZ_LINE];
  char tbuf2[SZ_LINE];
  char tbuf3[SZ_LINE];
  char *cmd;
  char *tptr;
  struct sockaddr_in sock_in;
  struct sockaddr_un sock_un;

  switch(client->type){
  case 'g':
    cmd = "xpaget";
    break;
  case 'i':
    cmd = "xpainfo";
    break;
  case 's':
    cmd = "xpaset";
    break;
  }

  /* get mode flags */
  XPAMode(mode, &(client->mode), "ack", XPA_CLIENT_ACK, 1);
  if( client->type == 's' )
    XPAMode(mode, &(xpa->client_mode), "verify", XPA_CLIENT_VERIFY, 0);

  /* package up and send the initialization message */
  strcpy(tbuf, cmd);
  /* set and save the id value */
  sprintf(tbuf3, "%c%d", client->type, id++);
  sprintf(tbuf2, " -i %s", tbuf3);
  strcat(tbuf, tbuf2);
  if( client->id )
    free(client->id);
  client->id = NewString(tbuf3);
  /* set the value of the client big-endian-ness */
  sprintf(tbuf2, " -e %s", XPAEndian() ? "big" : "little");
  strcat(tbuf, tbuf2);
  if( !(client->mode & XPA_CLIENT_ACK) )
    strcat(tbuf, " -n");
  sprintf(tbuf2, " %s:%s", client->xclass, client->name);
  strcat(tbuf, tbuf2);
  if( paramlist && *paramlist ){
    strcat(tbuf, " ");
    strncat(tbuf, paramlist, MAX(0,(int)(SZ_LINE-strlen(tbuf)-2)));
  }
  strcat(tbuf, "\n");
#ifdef XPA_DEBUG
  fprintf(stderr, "XPAClientStart %d sends: %s", client->cmdfd, tbuf);
#endif
  if( XPAPuts(client->cmdout, tbuf, XPAShortTimeout()) <= 0 ){
    goto error;
  }

  /* authentication */
retry:
  if( XPAGets(client->cmdin, tbuf2, SZ_LINE, XPAShortTimeout()) >0 ){
#ifdef XPA_DEBUG
    fprintf(stderr, "XPAClientStart %d received: %s", client->cmdfd, tbuf2);
#endif
    tptr = tbuf2;
    /* this should never happen */
    if( *tptr == '?' ){
      sprintf(errbuf, "XPA$WARNING: Protocol mismatch - missing id\n%s", tptr);
      goto error;
    }
    /* make sure we are dealing with a proper message */
    if( strncmp(tptr, client->id, strlen(client->id)) ){
      if( XPAVerbosity() > 1 ){
	fprintf(stderr, "XPA$WARNING: ignoring out of sync server message:\n");
	fprintf(stderr, "%s", tbuf2);
      }
      goto retry;
    }
    /* go past id */
    tptr += strlen(client->id);
    while( isspace(*tptr) ) tptr++;
    if( !strncmp(tptr, "XPA$NOBUF", 9) ){
      xpa->nclient += 1;
      goto started;
    }
    else if( !strncmp(tptr, "XPA$BUF", 7) ){
      if( sscanf(&(tptr[7]), "%s", method) != 1 ){
	sprintf(errbuf, "XPA$ERROR: bad authentication (%s:%s)\n",
		client->xclass, client->name);
	goto error;
      }
      switch(XPAMethod(method)){
      case XPA_INET:
	XPAParseIpPort(method, &ip, &port);
	/* connect to the server before we go further */
	if( (fd = socket(AF_INET, SOCK_STREAM, 0)) < 0 ){
	  sprintf(errbuf, "XPA$ERROR: bad socket for xpa dchan (%s:%s)\n",
		  client->xclass, client->name);
	  if( XPAVerbosity() ){
	    perror("XPA client socket");
	  }
	  goto error;
	}
	setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE,
		   (char *)&keep_alive, sizeof(keep_alive));
	memset((char *)&sock_in, 0, sizeof(sock_in));
	sock_in.sin_family = AF_INET;
	sock_in.sin_addr.s_addr = htonl(ip);
	sock_in.sin_port = htons(port);
#ifdef XPA_DEBUG
	fprintf(stderr, "attempting dchan connect: %s\n", method);
#endif
	if( connect(fd, (struct sockaddr *)&sock_in, sizeof(sock_in)) < 0 ){
#ifdef XPA_DEBUG
	  perror("dchan connect");
#endif
	  sprintf(errbuf,
		  "XPA$ERROR: can't connect to xpa server dchan (%s:%s)\n",
		  client->xclass, client->name);
	  if( XPAVerbosity() ){
	    perror("XPA client connect");
	  }
	  close(fd);
	  goto error;
	}
	break;
      case XPA_UNIX:
	tries = 0;
again:
	/* open a socket and fill in socket information */
	if( (fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 ){
	  sprintf(errbuf,
		  "XPA$ERROR: bad socket for xpa dchan (%s:%s)\n",
		  client->xclass, client->name);
	  if( XPAVerbosity() ){
	    perror("XPA client socket");
	  }
	  goto error;
	}
	setsockopt(xpa->fd, SOL_SOCKET, SO_KEEPALIVE,
		   (char *)&keep_alive, sizeof(keep_alive));
	memset((char *)&sock_un, 0, sizeof(sock_un));
	sock_un.sun_family = AF_UNIX;
	strcpy(sock_un.sun_path, method);
#ifdef XPA_DEBUG
	fprintf(stderr, "attempting dchan connect: %s\n", method);
#endif
	if( connect(fd, (struct sockaddr *)&sock_un, sizeof(sock_un)) < 0 ){
#ifdef XPA_DEBUG
	  perror("dchan connect");
#endif
	  close(fd);
	  if( (errno == ECONNREFUSED) && (tries < XPA_RETRIES) ){
	    tries++;
	    close(fd);
	    XPASleep(10);
	    goto again;
	  }
	  /* give up */
	  else{
	    sprintf(errbuf,
		    "XPA$ERROR: can't connect to xpa server dchan (%s:%s)\n",
		    client->xclass, client->name);
	    if( XPAVerbosity() ){
	      perror("XPA client connect");
	    }
	    goto error;
	  }
	}
	break;
      default:
	sprintf(errbuf,
		"XPA$ERROR: unknown XPA connect method (%s:%s)\n",
		client->xclass, client->name);
	goto error;
      }
      client->datafd = fd;
      /* make sure we close on exec */
      fcntl(client->datafd, F_SETFD, FD_CLOEXEC);
      /* for senders, set to no block mode */
      if( client->type == 's' ){
	flags = fcntl(client->datafd, F_GETFL);
	fcntl(client->datafd, F_SETFL, flags|NONBIO);
      }
      xpa->nclient += 1;
      goto started;
    }
    /* handle error message */
    else if( !strncmp(tptr, "XPA$ERROR", 9) ){
      sprintf(errbuf, "%s", tptr);
      goto error;
    }
    /* everything else is an error */
    else{
      sprintf(errbuf, "%s", tptr);
      goto error;
    }
  }
  else{
    sprintf(errbuf,
	    "XPA$ERROR: timeout waiting for server authentication (%s:%s)\n",
	    client->xclass, client->name);
    goto error;
  }

error:
#ifdef DEBUF
  fprintf(stderr, "Error in XPAClientStart: %s", errbuf);
#endif
  XPAClientFree(xpa, client);
  return(-1);

started:
  /* it is necessary to add this this extra hack/ack between the
     authentication ack and the error return (both from the server)
     to avoid getting Nagle buffered */
#ifdef XPA_DEBUG
  fprintf(stderr, "XPAClientStart %d sending nagle\n", client->cmdfd);
#endif
  XPAPuts(client->cmdout, "XPA$OK\n", XPAShortTimeout());

  /* mark as active and started */
  client->active = 2;
  return(0);
}

/*
 *----------------------------------------------------------------------------
 *
 *
 *			Public Routines
 *
 *
 *----------------------------------------------------------------------------
 */

/*
 *---------------------------------------------------------------------------
 *
 * Routine: 	XPAClientValid
 *
 * Purpose: 	see if the xpa client struct is valid
 *
 * Results:	1 on success, 0 for failure
 *
 *---------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAClientValid (XPA xpa)
#else
int XPAClientValid(xpa)
     XPA xpa;
#endif
{
  return(_XPAValid(xpaclienthead, xpa, "c"));
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAOpen
 *
 * Purpose:	open a persistent XPA client connection
 *
 * Returns:	XPA struct on success
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
XPA 
XPAOpen (char *mode)
#else
XPA XPAOpen(mode)
     char *mode;
#endif
{
  XPA xpa, cur;

  /* allocate xpa struct */
  if( (xpa = (XPA)calloc(1, sizeof(XPARec))) == NULL )
    return(NULL);
  /* add version */
  xpa->version = NewString(XPA_VERSION);
  /* mark this as a client struct */
  xpa->type = NewString("c");
  /* mark as persistent so we don't destroy at the end of the transfer */
  xpa->persist = 1;

  /* add this xpa to end of list of client xpas */
  XPAListAdd(&xpaclienthead, xpa);

  return(xpa);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAClose
 *
 * Purpose:	close a persistent XPA client connection
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
void 
XPAClose (XPA xpa)
#else
void XPAClose(xpa)
     XPA xpa;
#endif
{
  XPA cur;
  XPAClient client, tclient;

  /* ignore struct if its not a client struct */
  if( !XPAClientValid(xpa) )
      return;

  /* remove from list of client xpas */
  XPAListDel(&xpaclienthead, xpa);

  /* free each remaining client */
  for(client=xpa->clienthead; client!=NULL; ){
    tclient = client->next;
    XPAClientFree(xpa, client);
    client = tclient;
  }
  /* free string space */
  if( xpa->version )
    free(xpa->version);
  if( xpa->type )
    free(xpa->type);
  if( xpa )
    free(xpa);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAGet
 *
 * Purpose:	get XPA values
 *
 * Returns:	0 for success, -1 for failure
 *		len bytes of data returned in buf
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAGet (XPA xpa, char *xtemplate, char *paramlist, char *mode,
	char **bufs, int *lens, char **names, char **messages, int n)
#else
int XPAGet(xpa, xtemplate, paramlist, mode, bufs, lens, names, messages, n)
     XPA  xpa; 
     char *xtemplate;
     char *paramlist;
     char *mode;
     char **bufs;
     int  *lens;
     char **names;
     char **messages;
     int n;
#endif
{
  int get;
  int oldmode;
  int got=0;
  int type='g';
  char tbuf[SZ_LINE];
  XPAClient client, tclient;

  /* if not persistent, we need a temp xpa struct;
     (also ignore passed struct if its not a client struct) */
  if( (xpa == NULL) || strcmp(xpa->type, "c")  ){
    if( (xpa = XPAOpen(NULL)) == NULL )
      return(-1);
    /* mark this as not persistent */
    xpa->persist = 0;
  }
  /* save xpa mode -- this call might override */
  else{
    /* make sure we have a valid client handle */
    if( !XPAClientValid(xpa) ){
      if( XPAVerbosity() ){
	fprintf(stderr, "XPA$ERROR: invalid xpa client handle\n");
      }
      return(-1);
    }
    oldmode = xpa->client_mode;
  }

  /* these arrays are required */
  if( (bufs == NULL) || (lens == NULL) ){
    got = -1;
    goto done;
  }

  /* flag that we don't read from stdin */
  xpa->ifd = -1;

  /* zero out the return buffers */
  memset((char *)bufs,  0, ABS(n)*sizeof(char *));
  memset((char *)lens,  0, ABS(n)*sizeof(int));
  if( names != NULL )
    memset((char *)names, 0, ABS(n)*sizeof(char *));
  if( messages != NULL )
    memset((char *)messages,  0, ABS(n)*sizeof(char *));

  /* connect to clients and grab data */
  if( (get=XPAClientConnect(xpa, xtemplate, type)) >0 ){
    /* retrieve data from n active clients */
    for(client=xpa->clienthead; client!=NULL; ){
      tclient = client->next;
      if( (client->type == type) && (client->active > 0) && (got<n) ){
	if( names != NULL ){
	  sprintf(tbuf, "%s:%s %s",
		  client->xclass, client->name, client->method);
	  names[got] = NewString(tbuf);
	}
	if( XPAClientStart(xpa, client, paramlist, mode) >=0 ){
	  /* we fill buffers */
	  client->mode |= XPA_CLIENT_BUF;
	  client->bufptr = &(bufs[got]);
	  client->lenptr = &(lens[got]);
	  if( names != NULL )
	    client->nameptr = &(names[got]);
	  if( messages != NULL )
	    client->errptr = &(messages[got]);
	}
	else{
	  if( messages != NULL )
	    messages[got] = NewString(errbuf);
	}
	got++;
      }
      client = tclient;
    }
    /* we now have started all clients */
    xpa->status = XPA_CLIENT_START;
    /* if we have active clients */
    if( got ){
      /* enter the main loop and process */
      XPAClientLoop(xpa);
    }
  }

done:
  /* we are no longer active */
  xpa->status = XPA_CLIENT_IDLE;
  /* remove this xpa if we are not meant to persist */
  if( xpa && !xpa->persist )
    XPAClose(xpa);
  /* restore xpa mode -- this call might override */
  else
    xpa->client_mode = oldmode;

  /* return number of clients processes (including errors) */
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAGetFd
 *
 * Purpose:	get XPA values
 *
 * Returns:	0 for success, -1 for failure
 *		len bytes of data returned in buf
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAGetFd (XPA xpa, char *xtemplate, char *paramlist, char *mode,
	      int *fds, char **names, char **messages, int n)
#else
int XPAGetFd(xpa, xtemplate, paramlist, mode, fds, names, messages, n)
     XPA  xpa;
     char *xtemplate;
     char *paramlist;
     int  *fds;
     char *mode;
     char **names;
     char **messages;
     int n;
#endif
{
  int len;
  int get;
  int oldmode;
  int got=0;
  int type='g';
  char *s;
  char *buf;
  char tbuf[SZ_LINE];
  XPAClient client, tclient;

  /* if not persistent, we need a temp xpa struct;
     (also ignore passed struct if its not a client struct) */
  if( (xpa == NULL) || strcmp(xpa->type, "c")  ){
    if( (xpa = XPAOpen(NULL)) == NULL )
      return(-1);
    /* mark this as not persistent */
    xpa->persist = 0;
  }
  /* save xpa mode -- this call might override */
  else{
    /* make sure we have a valid client handle */
    if( !XPAClientValid(xpa) ){
      if( XPAVerbosity() ){
	fprintf(stderr, "XPA$ERROR: invalid xpa client handle\n");
      }
      return(-1);
    }
    oldmode = xpa->client_mode;
  }

  /* flag that we don't read from stdin */
  xpa->ifd = -1;

  /* zero out the return buffers */
  if( names != NULL )
    memset((char *)names, 0, ABS(n)*sizeof(char *));
  if( messages != NULL )
    memset((char *)messages,  0, ABS(n)*sizeof(char *));

  /* connect to clients and grab data */
  if( (get=XPAClientConnect(xpa, xtemplate, type)) > 0 ){
    /* retrieve data from n active clients */
    for(client=xpa->clienthead; client!=NULL; ){
      tclient = client->next;
      if( (client->type == type) && (client->active > 0) && (got<ABS(n)) ){
	if( names != NULL ){
	  sprintf(tbuf, "%s:%s %s",
		  client->xclass, client->name, client->method);
	  names[got] = NewString(tbuf);
	}
	if( XPAClientStart(xpa, client, paramlist, mode) >= 0 ){
	  /* we write to an fd */
	  client->mode |= XPA_CLIENT_FD;
	  /* negative value => one channel for all clients */
	  if( n < 0 )
	    client->fd = fds[0];
	  else
	    client->fd = fds[got];
	  client->bufptr = (char **)calloc(1, sizeof(char *));
	  client->lenptr = (int *)calloc(1, sizeof(int));
	  if( names != NULL )
	    client->nameptr = &(names[got]);
	  if( messages != NULL )
	    client->errptr = &(messages[got]);
	}
	else{
	  if( messages != NULL )
	    messages[got] = NewString(errbuf);
	}
	got++;
      }
      client = tclient;
    }
    /* we now have started all clients */
    xpa->status = XPA_CLIENT_START;
    /* if we have active clients */
    if( got ){
      /* enter the main loop and process */
      XPAClientLoop(xpa);
    }
  }

done:
  /* we are no longer active */
  xpa->status = XPA_CLIENT_IDLE;
  /* remove this xpa if we are not meant to persist */
  if( xpa && !xpa->persist )
    XPAClose(xpa);
  /* restore xpa mode -- this call might override */
  else
    xpa->client_mode = oldmode;

  /* return number of clients processes (including errors) */
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPASet
 *
 * Purpose:	set XPA values
 *
 * Returns:	0 for success, -1 for failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPASet (XPA xpa, char *xtemplate, char *paramlist, char *mode,
	char *buf, int len, char **names, char **messages, int n)
#else
int XPASet(xpa, xtemplate, paramlist, mode, buf, len, names, messages, n)
     XPA  xpa;
     char *xtemplate;
     char *paramlist;
     char *mode;
     char *buf;
     int len;
     char **names;
     char **messages;
     int n;
#endif
{
  int get;
  int oldmode;
  int got=0;
  int type='s';
  char *s;
  char tbuf[SZ_LINE];
  XPAClient client, tclient;

  /* if not persistent, we need a temp xpa struct;
     (also ignore passed struct if its not a client struct) */
  if( (xpa == NULL) || strcmp(xpa->type, "c")  ){
    if( (xpa = XPAOpen(NULL)) == NULL )
      return(-1);
    /* mark this as not persistent */
    xpa->persist = 0;
  }
  /* save xpa mode -- this call might override */
  else{
    /* make sure we have a valid client handle */
    if( !XPAClientValid(xpa) ){
      if( XPAVerbosity() ){
	fprintf(stderr, "XPA$ERROR: invalid xpa client handle\n");
      }
      return(-1);
    }
    oldmode = xpa->client_mode;
  }

  /* flag that we don't read from stdin */
  xpa->ifd = -1;

  /* zero out the return buffers */
  if( names != NULL )
    memset((char *)names, 0, ABS(n)*sizeof(char *));
  if( messages != NULL )
    memset((char *)messages,  0, ABS(n)*sizeof(char *));

  /* connect to clients and grab data */
  if( (get=XPAClientConnect(xpa, xtemplate, type)) >0 ){
    /* retrieve data from n active clients */
    for(client=xpa->clienthead; client!=NULL; ){
      tclient = client->next;
      if( (client->type == type) && (client->active > 0) ){
	if( names != NULL ){
	  sprintf(tbuf, "%s:%s %s",
		  client->xclass, client->name, client->method);
	  names[got] = NewString(tbuf);
	}
	if( XPAClientStart(xpa, client, paramlist, mode) >= 0 ){
	  /* we fill buffers */
	  client->mode |= XPA_CLIENT_BUF;
	  client->buf = buf;
	  client->len = len;
	  if( names != NULL )
	    client->nameptr = &(names[got]);
	  if( messages != NULL )
	    client->errptr = &(messages[got]);
	}
	else{
	  if( messages != NULL )
	    messages[got] = NewString(errbuf);
	}
	got++;
      }
      client = tclient;
    }
    /* we now have started all clients */
    xpa->status = XPA_CLIENT_START;
    /* if we have active clients */
    if( got ){
      /* enter the main loop and process */
      XPAClientLoop(xpa);
    }
  }

done:
  /* we are no longer active */
  xpa->status = XPA_CLIENT_IDLE;
  /* remove this xpa if we are not meant to persist */
  if( xpa && !xpa->persist )
    XPAClose(xpa);
  /* restore xpa mode -- this call might override */
  else
    xpa->client_mode = oldmode;

  /* return number of clients processes (including errors) */
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPASetFd
 *
 * Purpose:	set XPA values
 *
 * Returns:	0 for success, -1 for failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPASetFd (XPA xpa, char *xtemplate, char *paramlist, char *mode,
	      int fd, char **names, char **messages, int n)
#else
int XPASetFd(xpa, xtemplate, paramlist, mode, fd, names, messages, n)
     XPA  xpa;
     char *xtemplate;
     char *paramlist;
     char *mode;
     int fd;
     char **names;
     char **messages;
     int n;
#endif
{
  int len;
  int get;
  int oldmode;
  int got=0;
  int got2=0;
  int type='s';
  int flags;
  char *s;
  char buf[XPA_BIOSIZE];
  char tbuf[SZ_LINE];
  XPAClient client, tclient;

  /* if not persistent, we need a temp xpa struct;
     (also ignore passed struct if its not a client struct) */
  if( (xpa == NULL) || strcmp(xpa->type, "c")  ){
    if( (xpa = XPAOpen(NULL)) == NULL )
      return(-1);
    /* mark this as not persistent */
    xpa->persist = 0;
  }
  /* save xpa mode -- this call might override */
  else{
    /* make sure we have a valid client handle */
    if( !XPAClientValid(xpa) ){
      if( XPAVerbosity() ){
	fprintf(stderr, "XPA$ERROR: invalid xpa client handle\n");
      }
      return(-1);
    }
    oldmode = xpa->client_mode;
  }

  /* Set non-blocking mode for the input fd, if its not a tty */
  xpa->ifd = fd;
  flags = fcntl(xpa->ifd, F_GETFL);
  if( !isatty(xpa->ifd) )
    fcntl(xpa->ifd, F_SETFL, flags|NONBIO);

  /* zero out the return buffers */
  if( names != NULL )
    memset((char *)names, 0, ABS(n)*sizeof(char *));
  if( messages != NULL )
    memset((char *)messages,  0, ABS(n)*sizeof(char *));

  /* connect to clients and grab data */
  if( (get=XPAClientConnect(xpa, xtemplate, type)) >0 ){
    /* open clients all at once */
    for(client=xpa->clienthead; client!=NULL; ){
      tclient = client->next;
      if( (client->type == type) && (client->active > 0) ){
	if( names != NULL ){
	  sprintf(tbuf, "%s:%s %s",
		  client->xclass, client->name, client->method);
	  names[got] = NewString(tbuf);
	}
	if( XPAClientStart(xpa, client, paramlist, mode) >= 0 ){
	  /* we fill buffers */
	  client->mode |= XPA_CLIENT_FD;
	  if( names != NULL )
	    client->nameptr = &(names[got]);
	  if( messages != NULL )
	    client->errptr = &(messages[got]);
	}
	else{
	  if( messages != NULL )
	    messages[got] = NewString(errbuf);
	}
	got++;
      }
      client = tclient;
    }
    /* if we have active clients */
    if( got ){
      /* if fd is null, user did not want to send data, just the paramlist */
      if( fd < 0 ){
	for(client=xpa->clienthead; client!=NULL; ){
	  tclient = client->next;
	  if( (client->type == type) && (client->active > 0) ){
	    s = XPAClientEnd(xpa, client);
	    if( messages != NULL )
	      messages[got2] = NewString(s);
	    got2++;
	  }
	  client = tclient;
	}
      }
      else{
	/* we now have started all clients */
	xpa->status = XPA_CLIENT_START;
	/* enter the main loop and process */
	XPAClientLoop(xpa);
      }
    }
  }

done:
  /* we are no longer active */
  xpa->status = XPA_CLIENT_IDLE;
  /* reset flags, if necessary */
  if( xpa->ifd >=0 )
    fcntl(xpa->ifd, F_SETFL, flags);
  /* free all input structs */
  XPAClientFreeAllInputs(xpa);
  /* remove this xpa if we are not meant to persist */
  if( xpa && !xpa->persist )
    XPAClose(xpa);
  /* restore xpa mode -- this call might override */
  else{
    xpa->client_mode = oldmode;
  }

  /* return number of clients processes (including errors) */
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAInfo
 *
 * Purpose:	send XPA info
 *
 * Returns:	0 for success, -1 for failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAInfo (XPA xpa, char *xtemplate, char *paramlist, char *mode,
	 char **names, char **messages, int n)
#else
int XPAInfo(xpa, xtemplate, paramlist, mode, names, messages, n)
     XPA  xpa;
     char *xtemplate;
     char *paramlist;
     char *mode;
     char **names;
     char **messages;
     int n;
#endif
{
  int get;
  int oldmode;
  int got=0;
  char type='i';
  char *s;
  char tbuf[SZ_LINE];
  XPAClient client, tclient;

  /* if not persistent, we need a temp xpa struct;
     (also ignore passed struct if its not a client struct) */
  if( (xpa == NULL) || strcmp(xpa->type, "c")  ){
    if( (xpa = XPAOpen(NULL)) == NULL )
      return(-1);
    /* mark this as not persistent */
    xpa->persist = 0;
  }
  /* save xpa mode -- this call might override */
  else{
    /* make sure we have a valid client handle */
    if( !XPAClientValid(xpa) ){
      if( XPAVerbosity() ){
	fprintf(stderr, "XPA$ERROR: invalid xpa client handle\n");
      }
      return(-1);
    }
    oldmode = xpa->client_mode;
  }

  /* flag that we don't read from stdin */
  xpa->ifd = -1;

  /* zero out the return buffers */
  if( names != NULL )
    memset((char *)names, 0, ABS(n)*sizeof(char *));
  if( messages != NULL )
    memset((char *)messages,  0, ABS(n)*sizeof(char *));

  /* connect to clients and grab data */
  if( (get=XPAClientConnect(xpa, xtemplate, type)) >0 ){
    /* retrieve data from n active clients */
    for(client=xpa->clienthead; client!=NULL; ){
      tclient = client->next;
      if( (client->type == type) && (client->active > 0) ){
	if( names != NULL ){
	  sprintf(tbuf, "%s:%s %s",
		  client->xclass, client->name, client->method);
	  names[got] = NewString(tbuf);
	}
	if( XPAClientStart(xpa, client, paramlist, mode) >= 0 ){
	  s = XPAClientEnd(xpa, client);
	  if( messages != NULL )
	    messages[got] = NewString(s);
	}
	else{
	  if( messages != NULL )
	    messages[got] = NewString(errbuf);
	}
	got++;
      }
      client = tclient;
    }
  }

done:
  /* we are no longer active */
  xpa->status = XPA_CLIENT_IDLE;
  /* remove this xpa if we are not meant to persist */
  if( xpa && !xpa->persist )
    XPAClose(xpa);
  /* restore xpa mode -- this call might override */
  else
    xpa->client_mode = oldmode;

  /* return number of clients processes (including errors) */
  return(got);
}
