/*
 *      Copyrights:
 *
 *    	  Copyright (c) 1996, 1998 Smithsonian Astrophysical Observatory
 *
 *        Permission to use, copy, modify, distribute, and  sell  this
 *        software  and  its  documentation  for any purpose is hereby
 *        granted without  fee,  provided  that  the  above  copyright
 *        notice  appear  in  all  copies and that both that copyright
 *        notice and this permission notice appear in supporting docu-
 *        mentation,  and  that  the  name  of the  Smithsonian Astro-
 *	  physical Observatory not be used in advertising or publicity
 *  	  pertaining to distribution of the software without specific,
 *	  written  prior  permission.   The Smithsonian  Astrophysical 
 *        Observatory makes no representations about  the  suitability
 *	  of  this  software for any purpose.  It is provided  "as is"
 *	  without express or implied warranty.
 *        THE  SMITHSONIAN  ASTROPHYSICAL  OBSERVATORY  DISCLAIMS  ALL
 *	  WARRANTIES  WITH  REGARD  TO  THIS  SOFTWARE,  INCLUDING ALL
 *	  IMPLIED  WARRANTIES  OF  MERCHANTABILITY AND FITNESS, IN  NO
 *	  EVENT SHALL THE  SMITHSONIAN  ASTROPHYSICAL  OBSERVATORY  BE 
 *	  LIABLE FOR  ANY SPECIAL, INDIRECT  OR  CONSEQUENTIAL DAMAGES
 *	  OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS  OF USE,  DATA
 *	  OR  PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 *	  OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
 *	  THE  USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <sys/stat.h>
#include <xpap.h>

/*
 *----------------------------------------------------------------------------
 *
 *
 * 			Private Routines and Data
 *
 *
 *----------------------------------------------------------------------------
 */

/* this is the head of the global list -- too lazy to do anything more */
static XPA xpahead=NULL;
static XPA txpahead=NULL;

/* globals for the xpa environment, controlled by environment variables */
static int stimeout=XPA_SHORT_TIMEOUT; 	/* select, gets timeout in secs */
static int ltimeout=XPA_LONG_TIMEOUT; 	/* fillbuf timeout in secs */
static int verbosity=XPA_VERBOSITY;	/* 0=quiet, 1=normal, 2=full */
static int guseacl=1;		/* 0=don't use acls, 1=enable acls */
static int sigusr1=1;		/* 0=don't use sigusr1, 1=enable sigusr1 */
static char *tmpdir=NULL;  	/* temporary dir for logs. etc. */

/* name service variables -- used by all XPAs in this process */
static unsigned short nsport=0;
static unsigned int nsip=0;
static int nsfd=-1;
static FILE *nsin=NULL;
static FILE *nsout=NULL;
static char *activefds=NULL;
static char nsmethod[SZ_LINE];
static char nsusers[SZ_LINE];
static char nscmd[SZ_LINE];

/* global method type: unix, inet, server */
static int mtype=0;

/* width of select channel flag (this is overridden at init time) */
static int swidth=64;

/* erro code strings -- must match the error codes in xpap.h */
char *xpaMessbuf[] = {
  "oh, what a mess we've made!",
  "authentication failed",
  "xpa connection refused",
  "can't resolve host name for xpa",
  "can't read initialization info for xpa",
  "invalid xpa command in initialization string",
  "no receive function defined for this xpa",
  "no send function defined for this xpa",
  "no info function defined for this xpa",
  "undefined command for this xpa",
  "missing command for this xpa",
  "name does not match target template",
  "can't create data channel socket",
  "could not read data buf (possible timeout)"
};

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAEnableOneInput
 *
 * Purpose:	Enable 1 XPA entry to the XPA event loop
 *
 * Results:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void 
XPAEnableOneInput(void *client_data)
#else
static void XPAEnableOneInput(client_data)
     void *client_data;
#endif
{
  int fd = (int)client_data;
  if( (fd < 0) || (activefds == NULL) )
    return;
  activefds[fd] = 1;
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPADisableOneInput
 *
 * Purpose:	Disable 1 XPA entry from the XPA event loop
 *
 * Results:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void 
XPADisableOneInput (void *client_data)
#else
static void XPADisableOneInput(client_data)
     void *client_data;
#endif
{
  int fd = (int)client_data;
  if( (fd < 0) || (activefds == NULL) )
    return;
  activefds[fd] = 0;
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAAddOneInput
 *
 * Purpose:	Add 1 XPA entry to the XPA event loop
 *
 * Results:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void *
XPAAddOneInput (void *client_data, int fd)
#else
static void *XPAAddOneInput(client_data, fd)
     void *client_data;
     int fd;
#endif
{
  if( (fd < 0) || (activefds == NULL) )
    return(NULL);
  activefds[fd] = 1;
  return((void *)fd);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPADelOneInput
 *
 * Purpose:	Delete 1 XPA entry from the XPA event loop (called by XPAFree)
 *
 * Results:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void 
XPADelOneInput (void *client_data)
#else
static void XPADelOneInput(client_data)
     void *client_data;
#endif
{
  int fd = (int)client_data;
  if( (fd < 0) || (activefds == NULL) )
    return;
  activefds[fd] = 0;
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPANSOpen
 *
 * Purpose:	open a connection to the name service
 *
 * Returns:	connection fd on success, -1 on failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int 
XPANSOpen (XPA xpa, char *host, int force)
#else
static int XPANSOpen(xpa, host, force)
     XPA xpa;
     char *host;
     int force;
#endif
{
  int status;
  int pid;
  int i;
  int keep_alive=1;
  static int errinit=0;
  char *s;
  char *path;
  char *fullpath;
  char *method;
  char *argv[10];
  struct sockaddr_in sock_in;
  struct sockaddr_un sock_un;
  char *findname=NULL;

  /* if the name service already is open, just return fd */
  if( nsfd >=0 )
    return(nsfd);

  /* get name server method */
  method = XPANSMethod(0);

  /* get users for this user */
  if( (s=(char *)getenv("XPA_NSUSERS")) != NULL ){
    strcpy(nsusers, s);
  }
  /* default is just this use's userrname, from the environment */
  else if( (s=(char *)getenv("LOGNAME")) != NULL ){
    strcpy(nsusers, s);
  }
  /* this is a last resort */
  else{
    cuserid(nsusers);
  }
  /* if nothing good has happened, make it "all" */
  if( *nsusers == '\0' ){
    strcpy(nsusers, "*");
  }

  /* set up communications socket for this method */
  switch(mtype){
  case XPA_INET:
    XPAParseIpPort(method, &nsip, &nsport);
    if( (nsfd = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
      goto nons;
    memset((char *)&sock_in, 0, sizeof(sock_in));
    sock_in.sin_family = AF_INET;
    sock_in.sin_addr.s_addr = htonl(nsip);
    sock_in.sin_port = htons(nsport);
    /* try connecting to the name server */
    status=connect(nsfd, (struct sockaddr *)&sock_in, sizeof(sock_in));
    if( status == 0 ){
#ifdef XPA_DEBUG
      fprintf(stderr, "XPANSOpen: connect to xpans\n");
#endif
      goto okns;
    }
    else{
      close(nsfd);
    }
    break;
  case XPA_UNIX:
    /* open a socket and fill in socket information */
    if( (nsfd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 )
      goto nons;
    memset((char *)&sock_un, 0, sizeof(sock_un));
    sock_un.sun_family = AF_UNIX;
    strcpy(sock_un.sun_path, method);
    nsip = 0;
    /* try connecting to the name server */
    status=connect(nsfd, (struct sockaddr *)&sock_un, sizeof(sock_un));
    if( status == 0 ){
#ifdef XPA_DEBUG
      fprintf(stderr, "XPANSOpen: connect to xpans\n");
#endif
      goto okns;
    }
    else{
      close(nsfd);
    }
    break;
  }

  /* if force is set, we try to start up the name server */
  if(!force )
    goto noforce;

  /* make up the xpans command we will use */
  if( *nscmd == '\0' ){
    if( (nsip == gethostip(NULL)) || (mtype == XPA_UNIX) ){
      if( (path = (char *)getenv("path")) == NULL )
	path = (char *)getenv("PATH");
      fullpath = (char *)malloc(strlen(path)+132);
      strcpy(fullpath, path);
      if( *fullpath )
	strcat(fullpath, " ");
      /* add our bin directory */
      strcat(fullpath, "$SAORD_BIN");
      findname = (char *)Find("xpans", "x", NULL, fullpath);
      free(fullpath);
    }
    if( findname != NULL ){
      switch(mtype){
      case XPA_INET:
	if( (nsfd = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
	  goto nons;
	sprintf(nscmd, "%s -e -p %d -l %s/xpans_%d.log &\n",
		findname, nsport, tmpdir, nsport);
	break;
      case XPA_UNIX:
	if( (nsfd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 )
	  goto nons;
	sprintf(nscmd, "%s -e -f %s -l %s.log &\n",
		findname, method, method);
	break;
      }
    }
    else{
      *nscmd = '\0';
    }
  }

  /* did not find the name server -- start it up if we can, i.e.,
     if its on the same machine as we are on and we found it in the path */
  if( ((nsip == gethostip(NULL)) || (mtype == XPA_UNIX)) && (*nscmd != '\0') ){
    system(nscmd);
    /* enter loop looking to connect */
    for(i=0; i<XPA_RETRIES; i++){
      switch(mtype){
      case XPA_INET:
	if( (nsfd = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
	  goto nons;
	if(connect(nsfd, (struct sockaddr *)&sock_in, sizeof(sock_in)) ==0)
	  goto okns;
	break;
      case XPA_UNIX:
	if( (nsfd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 )
	  goto nons;
	if(connect(nsfd, (struct sockaddr *)&sock_un, sizeof(sock_un)) ==0)
	  goto okns;
	break;
      }
      close(nsfd);
      XPASleep(100);
    }
  }
  /* name server is on a remote machine, so there is no hope */
  else{
    goto nons;
  }

okns:
  /* make sure we close on exec */
  fcntl(nsfd, F_SETFD, FD_CLOEXEC);
  setsockopt(nsfd, SOL_SOCKET, SO_KEEPALIVE,
	     (char *)&keep_alive, sizeof(keep_alive));
  /* get file pointers for this socket */
  nsin = fdopen(nsfd, "r");
  nsout = fdopen(nsfd, "w");
  if( findname != NULL )
    free(findname);
  return(nsfd);

nons:
  switch(mtype){
  case XPA_INET:
    if( !errinit && verbosity ){
      if( nsip == gethostip(NULL) ){
	fprintf(stderr,
		"XPA$WARNING: xpans needs to be running on this machine.\n");
      }
      else{
	fprintf(stderr,
		"XPA$WARNING: xpans needs to be running on machine: ");
	fprintf(stderr, "%s\n", getiphost(nsip));
      }
    }
    break;
  case XPA_UNIX:
    if( !errinit && verbosity ){
      fprintf(stderr,
	      "XPA$WARNING: xpans needs to be running on this machine.\n");
    }
    break;
  default:
    break;
  }
  if( (xpa !=NULL) && verbosity ){
    if( !errinit ){
      fprintf(stderr, "Please start it using the command:\n\t%s", nscmd);
      fprintf(stderr, "Once xpans is running, please register all xpas in this process using:\n");
    }
    else{
      fprintf(stderr, "All xpa's in this process also can be registered with xpans using:\n");
    }
    fprintf(stderr, "\txpaset -p %s -nsconnect\n", xpa->method);
    fprintf(stderr, "(Until then, xpa names will not be available, but you can use:\n");
    fprintf(stderr, "\txpaset %s ...\nor\n\txpaget %s ...\nto communicate with this xpa.)\n",
	    xpa->method, xpa->method);
  }
  errinit++;
noforce:
  /* this will cause an init next time through */
  XPANSClose();
  if( findname != NULL )
    free(findname);
  return(-1);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPANSAdd
 *
 * Purpose:	add this XPA to the name service
 *
 * Returns:	0 on success, -1 on failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int 
XPANSAdd (XPA xpa)
#else
static int XPANSAdd(xpa)
     XPA xpa;
#endif
{
  char tbuf[SZ_LINE*2];
  char username[SZ_LINE];
  char *pathname;
  char *path;
  char *s;

  /* handle special case of the name server itself -- its a known entry */
  if( !strcmp(xpa->name, "xpans") ){
    return(0);
  }

  /* open a connection to the name service */
  if( XPANSOpen(xpa, NULL, 1) >= 0 ){
    /* get the user name, from the environment */
    if( (s=(char *)getenv("LOGNAME")) != NULL ){
      strcpy(username, s);
    }
    /* this is a last resort */
    else{
      cuserid(username);
    }
    /* if nothing good has happened, make it "unknown" */
    if( *username == '\0' ){
      strcpy(username, "unknown");
    }
    /* write the command to add this xpa */
    sprintf(tbuf, "add %s %s:%s %s %s\n",
	    xpa->method, xpa->xclass, xpa->name, xpa->type, username);
    XPAPuts(nsout, tbuf, stimeout);
    /* get result */
    if( XPAGets(nsin, tbuf, SZ_LINE, stimeout) >0 ){
      if( !strncmp(tbuf, "XPA$OK", 6) )
	return(0);
      else
	return(-1);
    }
    else
      return(-1);
  }
  else{
    return(-1);
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPANSDel
 *
 * Purpose:	remove public knowledge of this XPA
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int 
XPANSDel (XPA xpa)
#else
static int XPANSDel(xpa)
     XPA xpa;
#endif
{
  char tbuf[SZ_LINE];

  /* handle special case of the name server itself -- its a known entry */
  if( xpa->name && !strcmp(xpa->name, "xpans") ){
    return(0);
  }

  /* if there is no method, just return */
  if( (xpa->method == NULL) || (*xpa->method == '\0') ){
    return(0);
  }

  /* open a connection to the name service */
  if( XPANSOpen(xpa, NULL, 0) >= 0 ){
    /* write the command to delete this xpa */
    sprintf(tbuf, "del %s\n", xpa->method);
    XPAPuts(nsout, tbuf, stimeout);
    /* get result */
    if( XPAGets(nsin, tbuf, SZ_LINE, stimeout) >0 ){
      if( !strncmp(tbuf, "XPA$OK", 6) )
	return(0);
      else
	return(-1);
    }
    else
      return(-1);
  }
  else
    return(-1);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	CommNew
 *
 * Purpose:	allocate a new comm record and add to end of list
 *
 * Returns:	allocated comm record
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static XPAComm 
CommNew (XPA xpa)
#else
static XPAComm CommNew(xpa)
     XPA xpa;
#endif
{
  XPAComm comm, cur;
  static int id=0;
  int slen;
  int i;
  struct sockaddr_in sock_in;
  struct sockaddr_un sock_un;

  /* create the comm struct */
  if( (comm = (XPAComm)calloc(1, sizeof(XPACommRec))) == NULL )
    return(NULL);
  
  /* accept the connection */
  switch(mtype){
  case XPA_INET:
    slen = sizeof(struct sockaddr_in);
    if((comm->cmdfd=accept(xpa->fd, (struct sockaddr *)&sock_in, &slen)) >=0){
      /* make sure we close on exec */
      fcntl(comm->cmdfd, F_SETFD, FD_CLOEXEC);
      comm->cmdip = ntohl(sock_in.sin_addr.s_addr);
      comm->cmdport = ntohs(sock_in.sin_port);
      /* get file descriptors for this socket */
      comm->cmdin = fdopen(comm->cmdfd, "r");
      comm->cmdout = fdopen(comm->cmdfd, "w");
    }
    else{
      free(comm);
      return(NULL);
    }
    break;
  case XPA_UNIX:
    slen = sizeof(struct sockaddr_un);
    if((comm->cmdfd=accept(xpa->fd, (struct sockaddr *)&sock_un, &slen)) >=0){
      /* make sure we close on exec */
      fcntl(comm->cmdfd, F_SETFD, FD_CLOEXEC);
      comm->cmdname = NewString(sock_un.sun_path);
      /* get file descriptors for this socket */
      comm->cmdin = fdopen(comm->cmdfd, "r");
      comm->cmdout = fdopen(comm->cmdfd, "w");
    }
    else{
      free(comm);
      return(NULL);
    }
    break;
  default:
    free(comm);
    return(NULL);
  }

  /* mark data socket with impossible value */
  comm->datafd = -1;
  /* default is to ack */
  comm->ack = 1;
  /* set default cendian flag */
  comm->cendian = "?";
  /* clear the acl flags */
  for(i=0; i<4; i++){
    comm->acl[i] = -1;
  }

  /* add this comm to end of list of comms */
  if( xpa->commhead == NULL ){
    xpa->commhead = comm;
  }
  else{
    for(cur=xpa->commhead; cur->next!=NULL; cur=cur->next){
      ;
    }
    cur->next = comm;
  }

  /* we might have to add this fd specially to a non-select event loop */
  if( xpa->seladd )
    comm->selcptr = (xpa->seladd)(xpa, comm->cmdfd);
  
#ifdef XPA_DEBUG
  fprintf(stderr, "CommNew: ip=%x port=%d fd=%d\n",
	  comm->cmdip, comm->cmdport, comm->cmdfd);
#endif

  /* return the good news */
  return(comm);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	CommFree
 *
 * Purpose:	free a comm record and remove from list
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void 
CommFree (XPA xpa, XPAComm comm)
#else
static void CommFree(xpa, comm)
     XPA xpa;
     XPAComm comm;
#endif
{
  XPAComm cur;

#ifdef XPA_DEBUG
  fprintf(stderr, "CommFree: ip=%x port=%d fd=%d dfd=%d\n",
	  comm->cmdip, comm->cmdport, comm->cmdfd, comm->datafd);
#endif

  /* remove from list of this xpa's comms */
  if( xpa->commhead ){
    if( xpa->commhead == comm ){
      xpa->commhead = comm->next;
    }
    else{
      for(cur=xpa->commhead; cur!=NULL; cur=cur->next){
	if( cur->next == comm ){
	  cur->next = comm->next;
	  break;
	}
      }
    }
  }

  /* close socket connections */
  if( comm->cmdfd >= 0 ){
    /* delete the comm cmd fd specially from a non-select event loop */
    if( xpa->seldel && comm->selcptr ){
      (xpa->seldel)(comm->selcptr);
      comm->selcptr = NULL;
    }
    fclose(comm->cmdin);
    fclose(comm->cmdout);
    close(comm->cmdfd);
  }
  /* close data channel */
  XPACloseData(xpa, comm);
  /* if we have a file name (unix sockets), free it */
  if( comm->cmdname != NULL ){
    unlink(comm->cmdname);
    free(comm->cmdname);
  }
  if( comm->dataname != NULL ){
    unlink(comm->dataname);
    free(comm->dataname);
  }

  /* free up the space */
  if( comm->id != NULL )        free(comm->id);
  if( comm->target != NULL )    free(comm->target);
  if( comm->paramlist != NULL ) free(comm->paramlist);
  free(comm);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	CommActive
 *
 * Purpose:	make the cmd and data fds active or inactive for select
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void 
CommActive (XPA xpa, XPAComm comm, int flag)
#else
static void CommActive(xpa, comm, flag)
     XPA xpa;
     XPAComm comm;
     int flag;
#endif
{
  if( !xpa || !comm )
    return;

  /* remove this comm from the active list */
  if( flag == 0 ){
    if( xpa->seloff ){
      if( (comm->cmdfd >=0) && comm->selcptr ){
	(xpa->seloff)(comm->selcptr);
      }
      if((comm->datafd >=0) && comm->seldptr && (comm->datafd != comm->cmdfd)){
	(xpa->seloff)(comm->seldptr);
      }
    }
  }
  /* add this comm to the active list */
  else{
    if( xpa->selon ){
      if( (comm->cmdfd >=0) && comm->selcptr ){
	(xpa->selon)(comm->selcptr);
      }
      if((comm->datafd >=0) && comm->seldptr && (comm->datafd != comm->cmdfd)){
	(xpa->selon)(comm->seldptr);
      }
    }
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAOK
 *
 * Purpose:	send an XPA OK message to the client
 *
 * Returns:	0 on success, -1 on error
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int
XPAOK (XPA xpa)
#else
static int XPAOK(xpa)
     XPA xpa;
#endif
{
  int len;
  int status=0;
  char tbuf[SZ_LINE];

  /* make sure we have a valid struct */
  if( xpa == NULL )
    return(-1);

  /* send message, if necessary */
  if( !(xpa->status & XPA_STATUS_ACTIVE) || (xpa->ack == 1) ){
    if( xpa->id )
      sprintf(tbuf, "%s %sOK (%s:%s:%s)\n",
	      xpa->id, XPA_PREFIX, xpa->method, xpa->xclass, xpa->name);
    else
      sprintf(tbuf, "? %sOK (%s:%s:%s)\n",
	      XPA_PREFIX, xpa->method, xpa->xclass, xpa->name);
    len = strlen(tbuf);
    if( XPAPuts(xpa->cmdout, tbuf, stimeout) != len ){
      status = -1;
    }
  }

  /* return status */
  return(status);
}

/*
 *---------------------------------------------------------------------------
 *
 * Routine: 	_XPAFree
 *
 * Purpose: 	free up alloc'ed memory in the XPA record structure
 *		(internal version)
 *
 * Results:	1 on success, 0 for failure
 *
 *---------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int 
_XPAFree (XPA xpa)
#else
static int _XPAFree(xpa)
     XPA xpa;
#endif
{
  XPA cur;
  XPACmd cmd;
  XPAComm comm, tcomm;

  /* make sure we have something to do */
  if( xpa == NULL )
    return(-1);

  /* remove this xpa from public availability */
  if( xpa->type != NULL )
    XPANSDel(xpa);

  /* free all sub-commands */
  for(cmd=xpa->commands; cmd!=NULL; cmd=cmd->next){
    XPACmdDel(xpa, cmd);
  }

  /* remove from list of xpas */
  XPAListDel(&xpahead, xpa);

  /* free up space */
  if( xpa->version )	free(xpa->version);
  if( xpa->type )	free(xpa->type);
  if( xpa->method )	free(xpa->method);
  if( xpa->xclass ) 	free(xpa->xclass);
  if( xpa->name )	free(xpa->name);
  if( xpa->help )	free(xpa->help);

  /* if we have a file name (unix sockets), free it */
  if( xpa->filename )		free(xpa->filename);

  /* close down listening socket */
  if( xpa->fd >= 0 )
    close(xpa->fd);

  /* call the select free routine for the listening socket and loop type.
     we use an indirect routine to avoid having to link Xt, Tcl, etc. */
  if( xpa->seldel && xpa->selptr ){
    (xpa->seldel)(xpa->selptr);
    xpa->selptr = NULL;
  }

  /* close communication channels */
  for(comm=xpa->commhead; comm!=NULL; ){
    tcomm = comm->next;
    CommFree(xpa, comm);
    comm = tcomm;
  }

  /* free up record struct */
  free((char *)xpa);
  return(0);
}

/*
 *----------------------------------------------------------------------------
 *
 *
 *			Semi-Public Routines
 *			(mainly used by xpaset and xpaget)
 *
 *
 *----------------------------------------------------------------------------
 */

/*
 *---------------------------------------------------------------------------
 *
 * Routine: 	_XPAValid
 *
 * Purpose: 	see if the xpa struct is valid
 *
 * Results:	1 on success, 0 for failure
 *
 *---------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
_XPAValid (XPA head, XPA xpa, char *type)
#else
int _XPAValid(head, xpa, type)
     XPA head;
     XPA xpa;
     char *type;
#endif
{
  XPA cur;

  if( xpa == NULL )
    return(0);
  for(cur=head; cur!=NULL; cur=cur->next){
    if( (cur == xpa) && !strcspn(cur->type, type) ){
      return(1);
    }
  }
  return(0);
}

/*
 *---------------------------------------------------------------------------
 *
 * Routine: 	XPAValid
 *
 * Purpose: 	see if the xpa struct is valid
 *
 * Results:	1 on success, 0 for failure
 *
 *---------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAValid (XPA xpa)
#else
int XPAValid(xpa)
     XPA xpa;
#endif
{
  int got;

  if( _XPAValid(xpahead, xpa, XPA_ACLS) )
    return(1);
  else if( _XPAValid(txpahead, xpa, XPA_ACLS) )
    return(1);
  else
    return(0);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPABigendian
 *
 * Purpose:	semi-public routine to return the endian-ness of this
 *		machine
 *
 * Results:	0 if little endian, 1 if bigendian
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAEndian(void)
#else
int XPAEndian()
#endif
{
  union
  {
    long l;
    char c[sizeof (long)];
  } u;
  u.l = 1;
  return(u.c[sizeof (long) - 1] == 1);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAListHead
 *
 * Purpose:	semi-public routine to return the head of the xpa list
 *
 * Results:	XPA list pointer on success
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
XPA 
XPAListHead (void)
#else
XPA XPAListHead()
#endif
{
  return(xpahead);
}

/*
 *---------------------------------------------------------------------------
 *
 * Routine: 	XPAListAdd
 *
 * Purpose: 	add a member of an xpa list
 *
 * Results:	1 on success, 0 for failure
 *
 *---------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
void
XPAListAdd (XPA *head, XPA xpa)
#else
void XPAListAdd(head, xpa)
     XPA *head;
     XPA xpa;
#endif
{
  XPA cur;

  if( *head == NULL ){
    *head = xpa;
  }
  else{
    for(cur=*head; cur->next!=NULL; cur=cur->next)
      ;
    cur->next = xpa;
  }
}

/*
 *---------------------------------------------------------------------------
 *
 * Routine: 	XPAListDel
 *
 * Purpose: 	remove a member of an xpa list
 *
 * Results:	1 on success, 0 for failure
 *
 *---------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
void
XPAListDel (XPA *head, XPA xpa)
#else
void XPAListDel(head, xpa)
     XPA *head;
     XPA xpa;
#endif
{
  XPA cur;

  /* remove from list of xpas */
  if( *head ){
    if( *head == xpa ){
      *head = xpa->next;
    }
    else{
      for(cur=*head; cur!=NULL; cur=cur->next){
	if( cur->next == xpa ){
	  cur->next = xpa->next;
	  break;
	}
      }
    }
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAAddSelect
 *
 * Purpose:	add one or more xpa sockets to the select flags
 *
 * Return:	number of xpas that were added to the select flags
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAAddSelect (XPA xpa, fd_set *readfdsptr)
#else
int XPAAddSelect(xpa, readfdsptr)
     XPA xpa;
     fd_set *readfdsptr;
#endif
{
  XPA cur;
  XPAComm comm;
  int got=0;

  /* better have some place to set the flags */
  if( readfdsptr == NULL )
    return(0);

  /* if a specific xpa was specified, just set its select flags */
  if( xpa != NULL ){
    if( (xpa->fd >= 0) && (activefds[xpa->fd] != 0) ){
      FD_SET(xpa->fd, readfdsptr);
      got = 1;
    }
    for(comm=xpa->commhead; comm!=NULL; comm=comm->next){
      if( (comm->cmdfd >= 0) && (activefds[comm->cmdfd] != 0) ){
	FD_SET(comm->cmdfd, readfdsptr);
	if( (comm->datafd >= 0) && (activefds[comm->datafd] != 0) && 
	    (comm->datafd != comm->cmdfd) )
	  FD_SET(comm->datafd, readfdsptr);
	got++;
      }
    }
  }
  /* otherwise set select flags for all xpas */
  else{
    for(cur=xpahead; cur!=NULL; cur=cur->next){
      if( (cur->fd >= 0) && (activefds[cur->fd] != 0) ){
	FD_SET(cur->fd, readfdsptr);
	got++;
      }
      for(comm=cur->commhead; comm!=NULL; comm=comm->next){
	if( (comm->cmdfd >= 0) && (activefds[comm->cmdfd] != 0) ){
	  FD_SET(comm->cmdfd, readfdsptr);
	  if( (comm->datafd >= 0) && (activefds[comm->datafd] != 0) &&
	      (comm->datafd != comm->cmdfd) )
	    FD_SET(comm->datafd, readfdsptr);
	  got++;
	}
      }
    }
  }
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAProcessSelect
 *
 * Purpose:	process xpas that have pending reads or writes
 *
 * Return:	number of xpas processed
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAProcessSelect (fd_set *readfdsptr, int maxreq)
#else
int XPAProcessSelect(readfdsptr, maxreq)
     fd_set *readfdsptr;
     int maxreq;
#endif
{
  int got=0;
  int status;
  XPA cur;
  XPAComm comm, tcomm;

  /* <= 0 means do all of them */
  if( maxreq < 0 ){
    maxreq = 0;
  }
  for(cur=xpahead; cur!=NULL; cur=cur->next){
    /* handle command requests */
    for(comm=cur->commhead; comm!=NULL; ){
      tcomm = comm->next;
      if( (comm->cmdfd >=0) && FD_ISSET(comm->cmdfd, readfdsptr) ){
	FD_CLR(comm->cmdfd, readfdsptr);
	if( (status=XPAHandler(cur, comm->cmdfd)) != XPA_RTN_OK ){
	  break;
	}
	got++;
	if( maxreq && (got >= maxreq) ){
	  return(got);
	}
      }
      comm = tcomm;
    }
    /* handle data requests */
    for(comm=cur->commhead; comm!=NULL; ){
      tcomm = comm->next;
      if( (comm->datafd >=0) && FD_ISSET(comm->datafd, readfdsptr) ){
	FD_CLR(comm->datafd, readfdsptr);
	if( (status=XPAHandler(cur, comm->datafd)) != XPA_RTN_OK ){
	  break;
	}
	got++;
	if( maxreq && (got >= maxreq) ){
	  return(got);
	}
      }
      comm = tcomm;
    }
    /* handle new requests */
    if( (cur->fd >= 0) && FD_ISSET(cur->fd, readfdsptr) ){
      FD_CLR(cur->fd, readfdsptr);
      if( (status=XPAHandler(cur, cur->fd)) != XPA_RTN_OK ){
	break;
      }
      got++;
      if( maxreq && (got >= maxreq) ){
	return(got);
      }
    }
  }
#ifdef XPA_DEBUG
  fprintf(stderr, "XPAProcessSelect returns: %d\n", got);
#endif
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPACloseData
 *
 * Purpose:	close data fd if its not the cmd fd
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
void 
XPACloseData (XPA xpa, XPAComm comm)
#else
void XPACloseData(xpa, comm)
     XPA xpa;
     XPAComm comm;
#endif
{
  /* we close the data channel if its not the command channel */
  if( xpa && comm && (comm->datafd >=0)  ){
    if( comm->cmdfd != comm->datafd ){ 
      close(comm->datafd);
      /* delete the comm data fd specially from a non-select event loop */
      if( xpa->seldel && comm->seldptr ){
	(xpa->seldel)(comm->seldptr);
	comm->seldptr = NULL;
      }
    }
    /* reset data channel to impossible value */
    comm->datafd = -1;
    xpa->datafd = -1;
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAHandler
 *
 * Purpose:	handle one request for an xpaset or xpaget
 *
 * Return:	0 on success, xpa error code on failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAHandler (XPA xpa, int fd)
#else
int XPAHandler(xpa, fd)
     XPA xpa;
     int fd;
#endif
{
  char tbuf[SZ_LINE];
  char tbuf2[SZ_LINE];
  char tfile[SZ_LINE];
  char lbuf[SZ_LINE];
  char cmd[SZ_LINE];
  char target[SZ_LINE];
  char ctmpl[SZ_LINE];
  char ntmpl[SZ_LINE];
  char *paramlist=NULL;
  char *tfptr;
  int slen;
  int sock;
  int got=0;
  int lp=0;
  int keep_alive=1;
  struct sockaddr_in sock_in;
  struct sockaddr_un sock_un;
  struct timeval tv;
  struct timeval *tvp;
  XPARec tempxpa;
  XPA txpa;
  XPAComm comm=NULL;
  fd_set readfds;

#ifdef XPA_DEBUG
  fprintf(stderr, "XPAHandler entering with fd: %d\n", fd);
#endif

  /*  this is a defensive measure: we have to ensure that there really
   *  is a request read.  It is possible that a user-defined select loop
   *  might call us to handle a request that we had already handled
   *  (i.e. we handle a request but can't reset someone else's select flags)
   */
  FD_ZERO(&readfds);
  FD_SET(fd, &readfds);
  tv.tv_sec = 0;
  tv.tv_usec = 0;
  if( select(fd+1, &readfds, NULL, NULL, &tv) <=0){
    return(XPA_RTN_NOCMD);
  }

  /* if this is a first connection, we create a new comm channel and exit */
  if( fd == xpa->fd ){
    if( (comm = CommNew(xpa)) == NULL )
      return(XPA_RTN_NOCONN);
    /* Return to prevent xpa from finishing before other xpa's are started.
       Otherwise, each xpa's associated with a target template will be
       processed serially, which will defeat all the hard work done on
       the client side to send data to servers when they ask for it.
    */
    else
      return(XPA_RTN_OK);
  }

  /* look for a currently active comm channel: cmd or data */
  for(comm=xpa->commhead; comm!= NULL; comm=comm->next){
    if( (fd == comm->cmdfd) || (fd == comm->datafd) )
      break;
  }
  /* make sure we have something */
  if( comm == NULL ){
    return(XPA_RTN_NOCONN);
  }
  /* but don't recurse */
  if( comm->level ){
    return(-1);
  }

  /* Make temp xpa record that we will manipulate from now on. We do this
     mainly so we can populate the struct with comm information and pass
     it to the user-defined callback routine. But we also want to call
     XPAError in this routine. Thus we use the temp struct at this level,
     except when we call CommFree, which modified the comm list in the struct
     and thus requires the original. */
  txpa = &tempxpa;
  memcpy(txpa, xpa, sizeof(XPARec));
  txpa->cmdfd = comm->cmdfd;
  txpa->cmdin = comm->cmdin;
  txpa->cmdout = comm->cmdout;
  txpa->id = "?";

  /* data channel: go right to data handling section */
  if( fd == comm->datafd ){
    goto cb;
  }

  /* cmd channel: we are processing a new command */
retry:
  lp = 0;
  /* read next command */
  if( XPAGets(comm->cmdin, lbuf, SZ_LINE, stimeout) <=0 ){
#ifdef XPA_DEBUG
    if( *lbuf )
      fprintf(stderr, "XPAHandler %d command: %s", comm->cmdfd, lbuf);
    else
      fprintf(stderr, "XPAHandler %d EOF\n", comm->cmdfd);
#endif
    /* if this is the first connection, EOF is unexpected */
    if( comm->n == 1 ){
      got = XPA_RTN_NOBUF;
    }
    else{
      got = XPA_RTN_OK;
    }
    /* we must use original xpa struct, since this modifies its comm list */
    CommFree(xpa, comm);
    comm = NULL;
    /* and exit */
    goto done;
  }
  /* new-lines imply we entered telnet mode */
  else if( (*lbuf == '\n') || (*lbuf == '\r') || !strcmp(lbuf, "telnet") ){
    if( comm->telnet == 0 )
      XPAPuts(comm->cmdout, "Entering telnet mode ...\n", stimeout);
    comm->telnet = 1;
    comm->ack = 0;
    stimeout = -1;
    goto retry;
  }

  /* validate the command */
  if( !word(lbuf, cmd, &lp) ){
    got = XPA_RTN_NOCMD;
    goto done;
  }

  /* send help message -- also go into telnet mode */
  if( !strcmp(cmd, "help") ){
    XPAPuts(comm->cmdout,
      "xpaset|xpaget|xpainfo  [-e <b|l>] [-i id] [-n] [-t] class:name [paramlist]\n",
	    stimeout);
    XPAPuts(comm->cmdout,
	    "switches:\n",
	    stimeout);
    XPAPuts(comm->cmdout,
	    "\t-e <b|l>\tclient endian-ness: big(b) or little(l)\n",
	    stimeout);
    XPAPuts(comm->cmdout,
	    "\t-i id\tclient id string\n",
	    stimeout);
    XPAPuts(comm->cmdout,
	    "\t-n\t\tdon't ack back to client\n",
	    stimeout);
    XPAPuts(comm->cmdout,
	   "\t-t\t\ttelnet mode: use cmdfd for datafd for xpaget, else -1\n",
	   stimeout);
    comm->telnet = 1;
    comm->ack = 0;
    goto retry;
  }

  /* free previous id string and set to unknown */
  if( comm->id != NULL ){
    free(comm->id);
    comm->id = NULL;
  }

  /* process switches */
  while( word(lbuf, tbuf, &lp) ){
    if( *tbuf != '-' )
      break;
    if( !strcmp(tbuf, "-e") ){
      if( word(lbuf, tbuf, &lp) ){
	if( *tbuf == 'b' )
	  comm->cendian = "big";
	else if( *tbuf == 'l' )
	  comm->cendian = "little";
      }
    }
    else if( !strcmp(tbuf, "-i") ){
      if( word(lbuf, tbuf, &lp) ){
	comm->id = NewString(tbuf);
      }
    }
    else if( !strcmp(tbuf, "-n") )
      comm->ack = 0;
    else if( !strcmp(tbuf, "-t") )
      comm->telnet = 1;
    else
      break;
  }

  /* make sure we have some sort of id */
  if( comm->id == NULL ){
    comm->id = NewString("?");
  }
  txpa->id = comm->id;

  /* the first non-switch word we found was the target */
  if( *tbuf == '\0' ){
    XPAError(txpa, xpaMessbuf[XPA_RTN_NOCMD]);
    got = XPA_RTN_NOCMD;
    goto done;
  }
  else
    strcpy(target, tbuf);

  /* the rest of the input string is paramlist */
  paramlist = &(lbuf[lp]);
  nowhite(paramlist, paramlist);
  comm->paramlist = NewString(paramlist);

  /* set up command-specific info */
  if( !strcmp(cmd, "xpaset") ){
    if( !xpa->receive_callback ){
      XPAError(txpa, xpaMessbuf[XPA_RTN_NOREC]);
      got = XPA_RTN_NOREC;
      goto done;
    }
    comm->cmd = XPA_SET;
    comm->usebuf = (xpa->receive_mode & XPA_MODE_BUF);
    comm->useacl = guseacl && 
      (xpa->receive_mode & XPA_MODE_ACL) && (mtype != XPA_UNIX);
  }
  else if( !strcmp(cmd, "xpaget") ){
    if( !xpa->send_callback ){
      XPAError(txpa, xpaMessbuf[XPA_RTN_NOSEND]);
      got = XPA_RTN_NOSEND;
      goto done;
    }
    comm->cmd = XPA_GET;
    comm->usebuf = (xpa->send_mode & XPA_MODE_BUF);
    comm->useacl = guseacl &&
      (xpa->send_mode & XPA_MODE_ACL) && (mtype != XPA_UNIX);
  }
  else if( !strcmp(cmd, "xpainfo") ){
    if( !xpa->info_callback ){
      XPAError(txpa, xpaMessbuf[XPA_RTN_NOINFO]);
      got = XPA_RTN_NOINFO;
      goto done;
    }
    comm->cmd = XPA_INFO;
    comm->usebuf = 0;
    comm->useacl = guseacl &&
      (xpa->info_mode & XPA_MODE_ACL) && (mtype != XPA_UNIX);
  }
  else{
    XPAError(txpa, xpaMessbuf[XPA_RTN_NOCMD]);
    got = XPA_RTN_NOCMD;
    goto done;
  }

  /* perform access authentication */
  if( comm->useacl ){
    /* determine acl for this ip, if necessary */
    if( comm->acl[comm->cmd] < 0 ){
      comm->acl[comm->cmd] = XPAAclCheck(txpa, comm->cmdip, cmd[3]);
    }
    /* check acl */
    if( comm->acl[comm->cmd] <=0 ){
      XPAError(txpa, xpaMessbuf[XPA_RTN_NOAUTH]);
      got = XPA_RTN_NOAUTH;
      goto done;
    }
  }

  /* validate the target */
  XPAParseName(target, ctmpl, ntmpl, SZ_LINE);
  if( (strcmp(ctmpl,"?") && !tmatch(xpa->xclass, ctmpl)) || 
      (strcmp(ntmpl,"?") && !tmatch(xpa->name, ntmpl)) ){
    XPAError(txpa, xpaMessbuf[XPA_RTN_NOTARG]);
    got = XPA_RTN_NOTARG;
    goto done;
  }
  else{
    comm->target = NewString(target);
  }

  /* handle request for data connection from client */
  if( comm->usebuf && !comm->telnet ){
    switch(mtype){
    case XPA_INET:
      /* open a socket for data connections and fill in socket information */
      if( (sock = socket(AF_INET, SOCK_STREAM, 0)) < 0 ){
#ifdef XPA_DEBUG
	perror("dchan socket");
#endif
	XPAError(txpa, xpaMessbuf[XPA_RTN_NODCHAN]);
	got = XPA_RTN_NODCHAN;
	goto done;
      }
      setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE,
		 (char *)&keep_alive, sizeof(keep_alive));
      memset((char *)&sock_in, 0, sizeof(sock_in));
      sock_in.sin_family = AF_INET;
      sock_in.sin_addr.s_addr = htonl(gethostip(NULL));
      sock_in.sin_port = htons(0);
      /* bind to a port */
      if( bind(sock, (struct sockaddr *)&sock_in, sizeof(sock_in)) < 0 ){
#ifdef XPA_DEBUG
	perror("dchan bind");
#endif
	XPAError(txpa, xpaMessbuf[XPA_RTN_NODCHAN]);
	got = XPA_RTN_NODCHAN;
	close(sock);
	goto done;
      }
      /* we now can determine which port the system assigned */
      slen = sizeof(struct sockaddr_in);
      if( getsockname(sock, (struct sockaddr *)&sock_in, &slen) < 0 ){
#ifdef XPA_DEBUG
	perror("dchan getsockname");
#endif
	XPAError(txpa, xpaMessbuf[XPA_RTN_NODCHAN]);
	close(sock);
	got = XPA_RTN_NODCHAN;
	goto done;
      }
      sprintf(tbuf2, "%x:%d", 
	      ntohl(sock_in.sin_addr.s_addr), ntohs(sock_in.sin_port));
      comm->dataip = ntohl(sock_in.sin_addr.s_addr);
      comm->dataport = ntohs(sock_in.sin_port);
      break;
    case XPA_UNIX:
      /* get filename part, composed of class and name and unique id */
      sprintf(tfile, "%s_%s.%d-%s", xpa->xclass,xpa->name, getpid(), comm->id);
      /* change "/" to "_" for filename */
      for(tfptr = tfile; *tfptr != '\0'; tfptr++){
	if( *tfptr == '/' )
	  *tfptr = '_';
      }
      /* create full pathname */
      sprintf(tbuf2, "%s/%s", tmpdir, tfile);
      /* delete old copy */
      unlink (tbuf2);
      /* open a socket and fill in socket information */
      if( (sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 ){
	/* XPAError(txpa, xpaMessbuf[XPA_RTN_NODCHAN]); */
	got = XPA_RTN_NODCHAN;
	goto done;
      }
      setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE,
		 (char *)&keep_alive, sizeof(keep_alive));
      memset((char *)&sock_un, 0, sizeof(sock_un));
      sock_un.sun_family = AF_UNIX;
      strcpy(sock_un.sun_path, tbuf2);
      /* bind to a port */
      if( bind(sock, (struct sockaddr *)&sock_un, sizeof(sock_un)) < 0 ){
	XPAError(txpa, xpaMessbuf[XPA_RTN_NODCHAN]);
	close(sock);
	got = XPA_RTN_NODCHAN;
	goto done;
      }
      comm->dataname = NewString(tbuf2);
      break;
    default:
      XPAError(txpa, xpaMessbuf[XPA_RTN_NODCHAN]);
      got = XPA_RTN_NODCHAN;
      goto done;
    }

    /* send port to client so they can connect */
    /* first listen for the connection */
    if( listen(sock, XPA_MAXLISTEN) < 0 ){
#ifdef XPA_DEBUG
      perror("dchan listen");
#endif
      XPAError(txpa, xpaMessbuf[XPA_RTN_NODCHAN]);
      close(sock);
      got = XPA_RTN_NODCHAN;
      goto done;
    }

    /* and tell the client that we are listening */
    sprintf(tbuf, "%s %sBUF %s (%s:%s:%s)\n",
	    txpa->id ? txpa->id : "?",
	    XPA_PREFIX, tbuf2, txpa->method, txpa->xclass, txpa->name);
#ifdef XPA_DEBUG
      fprintf(stderr, "XPAHandler %d requesting dchan: %s", comm->cmdfd, tbuf);
#endif
    if( XPAPuts(comm->cmdout, tbuf, stimeout) <= 0 ){
#ifdef XPA_DEBUG
      perror("dchan write");
#endif
      close(sock);
      got = XPA_RTN_NODCHAN;
      goto done;
    }

    /* set up select timeout */
    FD_ZERO(&readfds);
    FD_SET(sock, &readfds);
    /* select timeout is in seconds */
    if( stimeout > 0 ){
      tv.tv_sec = stimeout;
      tv.tv_usec = 0;
      tvp = &tv;
    }
    /* this has to be able to time out! */
    else{
      tv.tv_sec = 5;
      tv.tv_usec = 0;
      tvp = &tv;
    }
    /* wait for the connection */
    got = select(swidth, &readfds, NULL, NULL, tvp);
#ifdef XPA_DEBUG
    fprintf(stderr, "return from dchan select: %d\n", got);
    if( got < 0 )
      perror("dchan select");
#endif
    /* process results of select */
    if( got > 0 ){
      switch(mtype){
      case XPA_INET:
	slen = sizeof(struct sockaddr_in);
	if((comm->datafd=accept(sock, (struct sockaddr *)&sock_in, &slen))<0){
#ifdef XPA_DEBUG
	  perror("dchan acccept");
#endif
	  XPAError(txpa, xpaMessbuf[XPA_RTN_NODCHAN]);
	  got = XPA_RTN_NODCHAN;
	  close(sock);
	  goto done;
	}
	/* make sure we close on exec */
	fcntl(comm->datafd, F_SETFD, FD_CLOEXEC);
	break;
      case XPA_UNIX:
	slen = sizeof(struct sockaddr_un);
	if((comm->datafd=accept(sock, (struct sockaddr *)&sock_un, &slen))<0){
	  XPAError(txpa, xpaMessbuf[XPA_RTN_NODCHAN]);
	  got = XPA_RTN_NODCHAN;
	  close(sock);
	  goto done;
	}
	/* make sure we close on exec */
	fcntl(comm->datafd, F_SETFD, FD_CLOEXEC);
	break;
      default:
	XPAError(txpa, xpaMessbuf[XPA_RTN_NODCHAN]);
	got = XPA_RTN_NODCHAN;
	close(sock);
	goto done;
      }
      close(sock);
      /* we might have to add this fd specially to a non-select event loop */
      if( xpa->seladd )
	comm->seldptr = (xpa->seladd)(xpa, comm->datafd);
    }
    /* timeout */
    else if( got == 0 ){
#ifdef XPA_DEBUG
      fprintf(stderr, "XPAHandler dchan timeout for %d\n", comm->cmdfd);
#endif
      XPAError(txpa, xpaMessbuf[XPA_RTN_NODCHAN]);
      got = XPA_RTN_NODCHAN;
      close(sock);
      goto done;
    }
    /* error */
    else{
#ifdef XPA_DEBUG
      perror("dchan select");
#endif
      XPAError(txpa, xpaMessbuf[XPA_RTN_NODCHAN]);
      got = XPA_RTN_NODCHAN;
      close(sock);
      goto done;
    }
  }
  /* no data channel being set up */
  else{
    /* handle telnet specially */
    if( comm->telnet ){
      if( comm->cmd == XPA_GET )
	comm->datafd = comm->cmdfd;
      else
	comm->datafd = -1;
    }
    /* send nobuf acknowledgement to client */
    else{
      sprintf(tbuf, "%s %sNOBUF (%s:%s:%s)\n",
	    txpa->id ? txpa->id : "?",
	    XPA_PREFIX, txpa->method, txpa->xclass, txpa->name);
      if( XPAPuts(comm->cmdout, tbuf, stimeout) <= 0 ){
	got = -1;
	goto done;
      }
    }
  }

  /* it is necessary to add this this extra hack/ack between the
     authentication ack and the error return (both from the server)
     to avoid getting Nagle buffered */
  if( !comm->telnet ){
    if( XPAGets(comm->cmdin, tbuf, SZ_LINE, stimeout) <=0 ){
#ifdef XPA_DEBUG
      perror("dchan nagle");
#endif
      got = XPA_RTN_NODCHAN;
      close(sock);
      goto done;
    }
  }

  /* if we have a data channel, we return at this stage and wait
     for data to be sent by the client */
  if( comm->usebuf && (comm->cmd == XPA_SET) && (comm->datafd >=0) ){
#ifdef XPA_DEBUG
    fprintf(stderr, "XPAHandler %d %d exiting to await data\n",
	    comm->cmdfd, comm->datafd);
#endif
    return(XPA_RTN_OK);
  }

  /* data channel complete: ready to execute the user callback */
cb:  
  /* increment level, we don't want to recurse through here */
  comm->level = 1;
  /* fill in the data part of the temp xpa record */
  txpa->datafd = comm->datafd;
  txpa->cendian = comm->cendian;
  txpa->ack = comm->ack;
  txpa->id = comm->id;
  txpa->comm = comm;
  /* zero out buf and len, just to make sure */
  txpa->buf = NULL;
  txpa->len = 0;
  /* zero out next pointer */
  txpa->next = NULL;

  /* we now are active */
  txpa->status |= XPA_STATUS_ACTIVE;
  xpa->status |= XPA_STATUS_ACTIVE;

  /* remove the current cmdfd and datafd from the list of xpa fds,
     in case the server callback re-enters the event loop */
  CommActive(txpa, comm, 0);

  /* add this temp xpa to the list of temps */
  XPAListAdd(&txpahead, txpa);

  /* process command */
  switch(comm->cmd){
  case XPA_GET:
#ifdef XPA_DEBUG
  fprintf(stderr, "processing xpaget for %d %d\n", comm->cmdfd, comm->datafd);
#endif
    /* check for a reserved command */
    lp = 0;
    if( XPACmdLookupReserved(txpa, comm->paramlist, &lp) ){
      got = XPASendCommands(txpa->send_data, txpa, comm->paramlist,
			    &(txpa->buf), &(txpa->len));
    }
    else{
      got = (txpa->send_callback)(txpa->send_data, txpa, comm->paramlist,
				  &(txpa->buf), &(txpa->len));
    }
    if( (got == 0) && (txpa->buf != NULL) && (txpa->len > 0) ){
      if( XPAPutBuf(comm->datafd, txpa->buf, txpa->len, ltimeout) < 0 ){
#ifdef XPA_DEBUG
	perror("XPAHandler write buf");
#endif
	XPAError(txpa, "XPA could not write data to client");
	goto done;
      }
    }
    if( (txpa->send_mode & XPA_MODE_FREEBUF) && (txpa->buf != NULL) ){
      free(txpa->buf);
    }
    /* send client an error message, unless its already been done */
    if( got ){
      if( !txpa->message )
	XPAError(txpa, "error detected in send callback routine");
    }
    else
      XPAOK(txpa);
#ifdef XPA_DEBUG
  fprintf(stderr, "finished xpaget for %d %d\n", comm->cmdfd, comm->datafd);
#endif
    break;
  case XPA_SET:
#ifdef XPA_DEBUG
  fprintf(stderr, "processing xpaset for %d %d\n", comm->cmdfd, comm->datafd);
#endif
    /* check for a reserved command */
    lp = 0;
    if( XPACmdLookupReserved(txpa, comm->paramlist, &lp) ){
      got=XPAReceiveCommands(txpa->receive_data, txpa,
			     comm->paramlist, NULL, 0);
    }
    else{
      /* fill buf if necessary */
      if( (comm->datafd >= 0) &&
	  comm->usebuf && (txpa->receive_mode & XPA_MODE_FILLBUF) ){
	if( XPAGetBuf(comm->datafd, &(txpa->buf), &(txpa->len), ltimeout) <0 ){
	  XPAError(txpa, xpaMessbuf[XPA_RTN_NODATA]);
	  got = -1;
	  goto done;
	}
	else{
	  /* close the data fd now that we are done with it */
	  XPACloseData(txpa, comm);
	}
      }
      /* execute the receive callback */
      got = (txpa->receive_callback)(txpa->receive_data, txpa,
				     comm->paramlist, txpa->buf, txpa->len);
    }
    if( (txpa->receive_mode & XPA_MODE_FREEBUF) && (txpa->buf != NULL) ){
      free(txpa->buf);
    }
    /* send client an error message, unless its already been done */
    if( got ){
      if( !txpa->message )
	XPAError(txpa, "error detected in receive callback routine");
    }
    else
      XPAOK(txpa);
#ifdef XPA_DEBUG
  fprintf(stderr, "finished xpaset for %d %d\n", comm->cmdfd, comm->datafd);
#endif
    break;
  case XPA_INFO:
    /* send OK before callback because we do not want the client waiting */
    XPAOK(txpa);
    /* and don't ack in callback */
    txpa->ack = 0;
    /* execute the info callback -- don't bother checking for errors */
    (txpa->info_callback)(txpa->info_data, txpa, comm->paramlist);
    break;
  default:
    XPAError(txpa, xpaMessbuf[XPA_RTN_NOCMD]);
    break;
  }

done:
#ifdef XPA_DEBUG
  fprintf(stderr, "finished processing %d with status %d\n", fd, got);
#endif
    
  /* add xpa back to list of active ones */
  CommActive(txpa, comm, 1);

  /* remove this temp xpa from the list of temps */
  XPAListDel(&txpahead, txpa);

  /* xpa is no longer active */
  xpa->status &= ~XPA_STATUS_ACTIVE;

  /* process comm record */
  if( comm ){
    /* back to level 0 */
    comm->level = 0;
    XPACloseData(txpa, comm);
  }

  /* if a free was requested by the callback, do it now */
  if( txpa &&  (txpa->status & XPA_STATUS_FREE) ){
      XPAFree(xpa);
  }
  /* if callback tried to free the real xpa, do it now */
  else if( xpa->status & XPA_STATUS_FREE )
    XPAFree(xpa);

  /* return the status */
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAMode
 *
 * Purpose:	parse the mode string and set flags
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
void 
XPAMode (char *mode, int *flag, char *name, int mask, int def)
#else
void XPAMode(mode, flag, name, mask, def)
     char *mode;
     int *flag;
     char *name;
     int mask;
     int def;
#endif
{
  char tbuf[SZ_LINE];
  char s[SZ_LINE];

  /* keyword routine requires an input buffer that can be modified */
  if( mode && *mode )
    strcpy(s, mode);
  else
    goto error;
  /* look for the keyword=<value> string */
  if( keyword(s, name, tbuf, SZ_LINE) ){
    if( istrue(tbuf) )
      *flag |= mask;
    else
      *flag &= ~mask;
    return;
  }
  else
    goto error;

error:
  if( def )
    *flag |= mask;
  else
    *flag &= ~mask;
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAShortTimeout
 *
 * Purpose:	return short (select, gets) timeout value
 *
 * Return:	timeout in seconds
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAShortTimeout (void)
#else
int XPAShortTimeout()
#endif
{
  return(stimeout);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPALongTimeout
 *
 * Purpose:	return long (fillbuf) timeout value
 *
 * Return:	timeout in seconds
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPALongTimeout (void)
#else
int XPALongTimeout()
#endif
{
  return(ltimeout);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAReceiveLTimeout
 *
 * Purpose:	modify long timeout value for this process
 *
 * Returns:	0 for success, -1 for failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAReceiveLTimeout (void *client_data, void *call_data, char *paramlist,
	       char *buf, int len)
#else
int XPAReceiveLTimeout(client_data, call_data, paramlist, buf, len)
     void *client_data;
     void *call_data;
     char *paramlist;
     char *buf;
     int len;
#endif
{
  XPA xpa = (XPA)call_data;
  char tbuf[SZ_LINE];
  char *s;

  if( paramlist && *paramlist ){
    strcpy(tbuf, paramlist);
    nocr(tbuf);
    culc(tbuf);
    if( !strcmp(tbuf, "reset") ){
      ltimeout = XPA_LONG_TIMEOUT;
      if( (s=(char *)getenv("XPA_LONG_TIMEOUT")) != NULL )
	ltimeout = atoi(s);
    }
    else{
      ltimeout = atoi(tbuf);
    }
    return(0);
  }
  else{
    XPAError(xpa, "missing long timeout value");
    return(-1);
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPASendLTimeout
 *
 * Purpose:	return the long timeout for this process
 *
 * Returns:	0 for success, -1 for failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPASendLTimeout (void *client_data, void *call_data, char *paramlist,
	    char **buf, int *len)
#else
int XPASendLTimeout(client_data, call_data, paramlist, buf, len)
     void *client_data;
     void *call_data;
     char *paramlist;
     char **buf;
     int *len;
#endif
{
  XPA xpa = (XPA)call_data;
  char tbuf[SZ_LINE];

  sprintf(tbuf, "%d\n", ltimeout);
  write(xpa_datafd(xpa), tbuf, strlen(tbuf));
  return(0);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAReceiveSTimeout
 *
 * Purpose:	modify short timeout value for this process
 *
 * Returns:	0 for success, -1 for failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAReceiveSTimeout (void *client_data, void *call_data, char *paramlist,
	       char *buf, int len)
#else
int XPAReceiveSTimeout(client_data, call_data, paramlist, buf, len)
     void *client_data;
     void *call_data;
     char *paramlist;
     char *buf;
     int len;
#endif
{
  XPA xpa = (XPA)call_data;
  char tbuf[SZ_LINE];
  char *s;

  if( paramlist && *paramlist ){
    strcpy(tbuf, paramlist);
    nocr(tbuf);
    culc(tbuf);
    if( !strcmp(tbuf, "reset") ){
      stimeout = XPA_SHORT_TIMEOUT;
      if( (s=(char *)getenv("XPA_SHORT_TIMEOUT")) != NULL )
	stimeout = atoi(s);
    }
    else{
      stimeout = atoi(tbuf);
    }
    return(0);
  }
  else{
    XPAError(xpa, "missing short timeout value");
    return(-1);
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPASendSTimeout
 *
 * Purpose:	return the short timeout for this process
 *
 * Returns:	0 for success, -1 for failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPASendSTimeout (void *client_data, void *call_data, char *paramlist,
	    char **buf, int *len)
#else
int XPASendSTimeout(client_data, call_data, paramlist, buf, len)
     void *client_data;
     void *call_data;
     char *paramlist;
     char **buf;
     int *len;
#endif
{
  XPA xpa = (XPA)call_data;
  char tbuf[SZ_LINE];

  sprintf(tbuf, "%d\n", stimeout);
  write(xpa_datafd(xpa), tbuf, strlen(tbuf));
  return(0);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAVerbosity
 *
 * Purpose:	return verbosity value
 *
 * Return:	verbosity value (0, 1, 2)
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAVerbosity (void)
#else
int XPAVerbosity()
#endif
{
  return(verbosity);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPASigusr1
 *
 * Purpose:	return flag for using sigusr1
 *
 * Return:	0 or 1
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPASigusr1 (void)
#else
int XPASigusr1()
#endif
{
  return(sigusr1);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAInitEnv
 *
 * Purpose:	initialize the xpa environment
 *
 * Return:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
void 
XPAInitEnv (void)
#else
void XPAInitEnv()
#endif
{
  char *s;
  struct rlimit rlp;
  static int init=0;

  if( !init ){
    /* determine the communication method */
    mtype=XPAMethod(NULL);
    /* enable access controls */
    if( mtype != XPA_UNIX )
      XPAAclNew(NULL, 0);
    /* set short (select,gets) timeout, if necessary */
    if( (s=(char *)getenv("XPA_SHORT_TIMEOUT")) != NULL )
      stimeout = atoi(s);
    /* set long (fillbuf) timeout, if necessary */
    if( (s=(char *)getenv("XPA_LONG_TIMEOUT")) != NULL )
      ltimeout = atoi(s);
    /* set verbosity level, if necessary */
    if( (s=(char *)getenv("XPA_VERBOSITY")) != NULL ){
      if( istrue(s) )
	verbosity = XPA_VERBOSITY;
      else if( isfalse(s) )
	verbosity = 0;
      else
	verbosity = atoi(s);
      if( verbosity < 0 )
	verbosity = 0;
    }
    /* check for acl enable/disable */
    if( (s=(char *)getenv("XPA_ACL")) != NULL )
      guseacl = istrue(s);
    /* check for use of siguser1 */
    if( (s=(char *)getenv("XPA_SIGUSR1")) != NULL )
      sigusr1 = istrue(s);
    /* make sure we have a temp dir */
    if( tmpdir == NULL ){
      if( (s=(char *)getenv("XPA_TMPDIR")) != NULL )
	tmpdir = NewString(s);
      else
	tmpdir = NewString(XPA_TMPDIR);
    }
    /* create directory, if necessary */
    mkdir(tmpdir, 0777);
    chmod(tmpdir, 0777);
    getrlimit(RLIMIT_NOFILE, &rlp);
    swidth = rlp.rlim_cur;
    activefds = (char *)calloc(swidth+1, sizeof(char));
    init++;
  }

  /* set up default signal actions */
  XPADefaultSigaction();
}

/*
 *----------------------------------------------------------------------------
 *
 *
 *			Public Routines
 *
 *
 *----------------------------------------------------------------------------
 */

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAMethod
 *
 * Purpose:	return communication method type
 *
 * Returns:	XPA__INET, XPA_UNIX
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAMethod (char *method)
#else
int XPAMethod(method)
     char *method;
#endif
{
  char *s1;

  /* if no method is passed, we return the default method type */
  if( method == NULL ){
    if( mtype == 0 ){
      s1 = (char *)getenv("XPA_METHOD");
      if( s1 == NULL )
	mtype = XPA_INET;
      else if( !strcasecmp(s1, "inet") )
	mtype = XPA_INET;
      else if( !strcasecmp(s1, "unix") )
	mtype = XPA_UNIX;
      else if( !strcasecmp(s1, "local") )
	mtype = XPA_UNIX;
      else
	mtype = XPA_INET;
    }
    return(mtype);
  }
  /* otherwise, we analyze the input method to get the type */
  else{
    /* inet is ip:port, else unix filename */
    if( strchr(method, ':') != NULL )
      return(XPA_INET);
    else
      return(XPA_UNIX);
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPANSMethod
 *
 * Purpose:	return string containing name server method
 *
 * Returns:	name server method string
 *
 *----------------------------------------------------------------------------
 */

#ifdef ANSI_FUNC
char * 
XPANSMethod (int flag)
#else
char *XPANSMethod(flag)
     int flag;
#endif
{
  char *s, *t;
  int port;

  switch( XPAMethod(NULL) ){
  case XPA_INET:
    if( (s=(char *)getenv("XPA_NSINET")) != NULL )
      strcpy(nsmethod, s);
    else
      strcpy(nsmethod, XPA_NSINET);
    /* if flag is set, we are getting the XPA access ip:port, not the
       socket port and we have to increment by one */
    if( flag ){
      s = strrchr(nsmethod, ':');
      if( s != NULL ){
	t = s+1;
	if( (*t == '\0') || !strcmp(t,"*") )
	  port = XPA_NSIPORT+1;
	else
	  port = atoi(t)+1;
	*t = '\0';
	sprintf(t, "%d", port);
      }
    }
    break;
  case XPA_UNIX:
    if( (s=(char *)getenv("XPA_NSUNIX")) != NULL )
      strcpy(nsmethod, s);
    else{
      sprintf(nsmethod, "%s/%s", tmpdir, XPA_NSUNIX);
    }
    /* if flag is set, we are getting the XPA access file, not the
       socket file and we have to change the name slightly */
    if( flag ){
      /* replace the ending, if possible */
      s = strrchr(nsmethod, '.');
      t = strrchr(nsmethod, '/');
      if( (s != NULL) && (s > t) )
	*s = '\0';
      strcat(nsmethod, ".xpa");
    }
    break;
  default:
    if( (s=(char *)getenv("XPA_NSINET")) != NULL )
      strcpy(nsmethod, s);
    else
      strcpy(nsmethod, XPA_NSINET);
    break;
  }

  /* return the static method string */
  return(nsmethod);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAParseName
 *
 * Purpose:	split the xpaname into a class and name string
 *
 * Results:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
void 
XPAParseName (char *xpaname, char *xclass, char *name, int len)
#else
void XPAParseName(xpaname, xclass, name, len)
     char *xpaname;
     char *xclass;
     char *name;
     int len;
#endif
{
  char *s;
  char *t;
  char *cptr=NULL;
  char *nptr=NULL;

  /* if nothing is passed to us, allow everything */
  if( (xpaname == NULL) || (*xpaname == '\0') ){
    strncpy(xclass, "*", len-1);
    strncpy(name, "*", len-1);
    return;
  }
    
  /* split the xpaname into class and name */
  s = NewString(xpaname);
  if( (t=(char *)strchr(s, ':')) != NULL ){
    cptr = s;
    *t = '\0';
    nptr = t+1;
  }
  else{
    nptr = s;
    cptr = "*";
  }
  if( *cptr == '\0' )
    cptr = "*";
  if( *nptr == '\0' )
    nptr = "*";

  strncpy(xclass, cptr, len-1);
  strncpy(name, nptr, len-1);
  free(s);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAParseIpPort
 *
 * Purpose:	split the host into ip and port
 *
 * Results:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAParseIpPort (char *host, unsigned int *ip, unsigned short *port)
#else
int XPAParseIpPort(host, ip, port)
     char *host;
     unsigned int *ip;
     unsigned short *port;
#endif
{
  char *s1, *s2, *s3, *p1, *p2;
  int got;

  /* make sure we have something to work with */
  if( (host == NULL) || (*host == '\0') )
    return(0);

  /* parse up the string and look for a port number (which is essential) */
  s1 = NewString(host);
  /* if we have a ":", we will null it out (so that what comes before is
     the host name) and bump past it to point to the port */
  if( (p1 = (char *)strchr(s1, ':')) == NULL ){
    /* there is no ":", so the whole string is the port */
    p1 = s1;
    s2 = NULL;
  } else {
    /* null out ':' and bump port pointer */
    *p1 = '\0';
    p1++;
    s2 = s1;
  }

  /* get port */
  p2 = NULL;
  if( p1 && !strcmp(p1,"*") )
    *port = XPA_NSIPORT;
  else
    *port = (unsigned short)strtol(p1, &p2, 0);
  /* check for bad port number -- we lose */
  if( *port <=0 || (p1 == p2) ){
    *ip = 0;
    *port = 0;
    got = 0;
    goto done;
  }

  /* get ip */
  /* see if this already is a hex address in network byte order */
  if( s2 && *s2 && strcmp(s2, "*") ){
    *ip = strtoul16(s2, &s3);
    if( *s3 == '\0' ){
      got = 1;
      goto done;
    }
  }
  /* not hex -- convert ip string (might be null) to an ip address */
  if( s2 && !strcmp(s2, "*") )
    s2 = NULL;
  if( (*ip = gethostip(s2)) == 0 ){
    *port = 0;
    got = 0;
  }
  else{
    *port = *port;
    got = 1;
  }

done:
  free(s1);
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAParseUnixSocket
 *
 * Purpose:	see if host is actually a unix socket file
 *
 * Results:	1 if its a socket, 0 otherwise
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAParseUnixSocket (char *host)
#else
int XPAParseUnixSocket(host)
     char *host;
#endif
{
  struct stat buf;
  char dir[SZ_LINE];
  char xclass[SZ_LINE];
  char name[SZ_LINE];
  int uid;

  /* see if its a file in the right directory */
  if( !strncmp(host, tmpdir, strlen(tmpdir)) &&
      !stat(host, &buf) ){
    return(1);
  }
  else{
    return(0);
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAError
 *
 * Purpose:	send an XPA error message to the client
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int
XPAError (XPA xpa, char *s)
#else
int XPAError(xpa, s)
     XPA xpa;
     char *s;
#endif
{
  int status=0;
  char *t;
  char *u;

  /* make sure we have a valid struct */
  if( (xpa == NULL) || (xpa->cmdout == NULL) )
    return(-1);

  /* send message, if necessary */
  if( !(xpa->status & XPA_STATUS_ACTIVE) || (xpa->ack == 1) ){
    t = NewString(s);
    /* get rid of CR in message -- we add one at the end */
    nowhite(t, t);
    u = (char *)calloc(strlen(t)+SZ_LINE, sizeof(char));
    /* package up and write the message */
    sprintf(u, "%s %sERROR %s (%s:%s %s)\n",
	    xpa->id ? xpa->id : "?",
	    XPA_PREFIX, t, xpa->xclass, xpa->name, xpa->method);
    if( XPAPuts(xpa->cmdout, u, stimeout) != strlen(u) ){
      status = -1;
    }
    if( t )
      free(t);
    if( u )
      free(u);
  }

  /* flag that there was a message sent for this xpa */
  xpa->message = 1;
  /* return status */
  return(status);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAMessage
 *
 * Purpose:	send an XPA message to the client
 *
 * Returns:	0 on success, -1 on error
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAMessage (XPA xpa, char *s)
#else
int XPAMessage(xpa, s)
     XPA xpa;
     char *s;
#endif
{
  int status=0;
  char *t;
  char *u;

  /* make sure we have a valid struct */
  if( (xpa == NULL) || (xpa->cmdout == NULL) )
    return(-1);

  /* send message, if necessary */
  if( !(xpa->status & XPA_STATUS_ACTIVE) || (xpa->ack == 1) ){
    t = NewString(s);
    /* get rid of CR in message -- we add one at the end */
    nowhite(t, t);
    u = (char *)calloc(strlen(t)+SZ_LINE, sizeof(char));
    /* package up and write the message */
    sprintf(u, "%s %sMESSAGE %s (%s:%s %s)\n",
	    xpa->id ? xpa->id : "?",
	    XPA_PREFIX, t, xpa->xclass, xpa->name, xpa->method);
    if( XPAPuts(xpa->cmdout, u, stimeout) != strlen(u) ){
      status = -1;
    }
    if( t )
      free(t);
    if( u )
      free(u);
  }

  /* flag that there was a message sent for this xpa */
  xpa->message = 1;
  /* return status */
  return(status);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAArgvParamlist
 *
 * Purpose:	generate a paramlist string from an argv
 *
 * Results:	allocated paramlist string
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
char *
XPAArgvParamlist (int argc, char **argv, int start)
#else
char *XPAArgvParamlist(argc, argv, start)
     int argc;
     char **argv;
     int start;
#endif
{
  int plen;
  int i;
  char *paramlist;

  /* get length of paramlist */
  for(plen=0, i=start; i<argc; i++){
    plen += (strlen(argv[i])+1);
  }

  /* allocate enough space for it */
  if( (paramlist = (char *)calloc(sizeof(char), plen+1)) == NULL ){
    return(NULL);
  }

  /* gather up the paramlist */
  for(i=start; i<argc; i++){
    strcat(paramlist, argv[i]);
    strcat(paramlist, " ");
  }

  /* remove whitespace from beginning and end */
  nowhite(paramlist, paramlist);

  /* return paramlist */
  return(paramlist);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAAccess
 *
 * Purpose:	get number of name server matches
 *
 * Returns:	number of matches
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAAccess (char *tname, char *ttype)
#else
int XPAAccess(tname, ttype)
     char *tname;
     char *ttype;
#endif
{
  char **xclasses;
  char **names;
  char **methods;
  int i, n;

  
  /* get all matches */
  if( (tname == NULL) || (*tname == '\0') )
    return(0);

  n = XPANSLookup(tname, ttype, &xclasses, &names, &methods);

  /* free up the space */
  for(i=0; i<n; i++){
    /* done with these strings */
    free(xclasses[i]);
    free(names[i]);
    free(methods[i]);
  }
  /* free up arrays alloc'ed by name server */
  if( n > 0 ){
    free(xclasses);
    free(names);
    free(methods);
  }
#ifdef XPA_DEBUG
  fprintf(stderr, "XPAAccess finds %d entries\n", n);
#endif
  return(n);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPANSLookup
 *
 * Purpose:	get name server matches
 *
 * Returns:	number of matches
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPANSLookup (char *tname, char *ttype,
	     char ***xclasses, char ***names, char ***methods)
#else
int XPANSLookup(tname, ttype, xclasses, names, methods)
     char *tname;
     char *ttype;
     char ***xclasses;
     char ***names;
     char ***methods;
#endif
{
  unsigned short port;
  unsigned int ip;
  int fd;
  int own;
  int lp=0;
  int got=0;
  int nalloc=100;
  char *xtype;
  char xclass[SZ_LINE];
  char name[SZ_LINE];
  char method[SZ_LINE];
  char user[SZ_LINE];
  char type[SZ_LINE];
  char tbuf[SZ_LINE];
  char lbuf[SZ_LINE];
  XPA xpa;

  /* do some initialization */
  XPAInitEnv();

  /* use "*" as default if no type was specified */
  if( (ttype == NULL) || (*ttype == '\0') )
    xtype = "*";
  else
    xtype = ttype;

  /* handle special case of name server */
  if( !strcmp(tname, "xpans") ){
    *xclasses   = (char **)malloc(sizeof(char *));
    *names     = (char **)malloc(sizeof(char *));
    *methods     = (char **)malloc(sizeof(char *));
    (*xclasses)[0] = NewString("*");
    (*names)[0]   = NewString("xpans");
    (*methods)[0]   = NewString(XPANSMethod(1));
    return(1);
  }

  /* handle special case of a string containing ip:port */
  if( XPAParseIpPort(tname, &ip, &port) ){
    *xclasses   = (char **)malloc(sizeof(char *));
    *names      = (char **)malloc(sizeof(char *));
    *methods    = (char **)malloc(sizeof(char *));
    (*xclasses)[0] = NewString("?");
    (*names)[0]   = NewString("?");
    (*methods)[0] = NewString(tname);
    return(1);
  }

  /* handle special case of a string containing unix socket */
  if( XPAParseUnixSocket(tname) ){
    *xclasses   = (char **)malloc(sizeof(char *));
    *names      = (char **)malloc(sizeof(char *));
    *methods    = (char **)malloc(sizeof(char *));
    (*xclasses)[0] = NewString("?");
    (*names)[0]   = NewString("?");
    (*methods)[0] = NewString(tname);
    return(1);
  }

  /* pre-allocate the various arrays to an absurd number */
  *xclasses = (char **)malloc(nalloc * sizeof(char *));
  *names =   (char **)malloc(nalloc * sizeof(char *));
  *methods =   (char **)malloc(nalloc * sizeof(char *));

  /* open a connection to the name service */
  if( XPANSOpen(NULL, NULL, 0) >= 0 ){
    while( word(tname, lbuf, &lp) ){
      XPAParseName(lbuf, xclass, name, SZ_LINE);
      /* write the command to add this xpa */
      sprintf(tbuf, "lookup %s:%s %s %s\n", xclass, name, xtype, nsusers);
#ifdef XPA_DEBUG
      fprintf(stderr, "XPANSLookup sending: %s", tbuf);
#endif
      XPAPuts(nsout, tbuf, stimeout);
      /* read matches from the name server */
      while( 1 ){
	if( XPAGets(nsin, tbuf, SZ_LINE, stimeout) <=0 ){
#ifdef XPA_DEBUG
	  fprintf(stderr, "XPANSLookup got unexpected EOF from xpans\n");
#endif
	  /* reset the name server and try to recover */
	  XPANSReset();
	  break;
	}
#ifdef XPA_DEBUG
	fprintf(stderr, "XPANSLookup receiving: %s", tbuf);
#endif
	/* XPA$<result> signals end of input */
	if( !strncmp(tbuf, "XPA$", 4) )
	  break;
	/* otherwise scan next line */
	if( sscanf(tbuf, "%s %s %s %s %s\n", 
		   xclass, name, type, method, user) != EOF ){
	  /* make sure this entry is not in the current process
	     (i.e., we can't ever xpa ourselves!) */
	  for(own=0, xpa=xpahead; xpa!=NULL; xpa=xpa->next){
	    if( !strcmp(xpa->xclass, xclass) && !strcmp(xpa->name, name) &&
		!strcmp(xpa->method, method)				 ){
	      own = 1;
	      break;
	    }
	  }
	  /* if this xpa is in the current process, skip it */
	  if( own )
	    continue;
	  /* make sure we have enough space */
	  if( got >= nalloc ){
	    nalloc *= 2;
	    *xclasses = (char **)realloc(*xclasses, nalloc * sizeof(char *));
	    *names =   (char **)realloc(*names, nalloc * sizeof(char *));
	    *methods =   (char **)realloc(*methods, nalloc * sizeof(char *));
	  }
	  /* add this entry to the list */
	  (*xclasses)[got] = NewString(xclass);
	  (*names)[got] = NewString(name);
	  (*methods)[got] = NewString(method);
	  got++;
	}
      }
    }
  }
  /* reallocate the exact number of buffers we have */
  if( got > 0 ){
    *xclasses = (char **)realloc(*xclasses, got * sizeof(char *));
    *names    = (char **)realloc(*names, got * sizeof(char *));
    *methods  = (char **)realloc(*methods, got * sizeof(char *));
  }
  else{
    free(*xclasses);
    free(*names);
    free(*methods);
  }
#ifdef XPA_DEBUG
  fprintf(stderr, "XPANSLookup finds %d entries\n", got);
#endif
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPANSClose
 *
 * Purpose:	close connection to the name service
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int
XPANSClose (void)
#else
int XPANSClose()
#endif
{
  if( nsin != NULL )
    fclose(nsin);
  nsin = NULL;
  if( nsout != NULL )
    fclose(nsout);
  nsout = NULL;
  close(nsfd);
  nsfd = -1;
  nsport = 0;
  nsip = 0;
  *nsusers = '\0';
  *nscmd = '\0';
  return(0);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPANSReset
 *
 * Purpose:	close and reset connection to the name service,
 *		adding back all XPA's for this process
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int
XPANSReset (void)
#else
int XPANSReset()
#endif
{
  XPA cur;
  int got=0;

  /* close current connection */
  XPANSClose();

  /* open connection and add back all XPA's */
  for(cur=xpahead; cur!=NULL; cur=cur->next){
    if( (got=XPANSAdd(cur)) != 0 )
      break;
  }
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPANew
 *
 * Purpose:	add a new xpa name to a process to allow external
 * 		process to read/write data associated with this name.
 *
 * If the send callback is defined, it can send to an external process.
 * If the receive callback is defined, it can receive from an external process.
 *
 * Returns: xpa handle associated with this class.name or NULL
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
XPA 
XPANew (char *xclass, char *name, char *help,
	SendCb send_callback, void *send_data, char *send_mode,
	ReceiveCb rec_callback, void *rec_data, char *rec_mode)
#else
XPA XPANew(xclass, name, help,
	   send_callback, send_data, send_mode,
	   rec_callback, rec_data, rec_mode)
     char *xclass;
     char *name;
     char *help;
     SendCb send_callback;
     void *send_data;
     char *send_mode;
     ReceiveCb rec_callback;
     void *rec_data;
     char *rec_mode;
#endif
{
  unsigned short port;
  unsigned int ip;
  int slen = sizeof(struct sockaddr_in);
  int reuse_addr=1;
  int keep_alive=1;
  char tbuf[SZ_LINE];
  char tfile[SZ_LINE];
  char *tfptr;
  XPA xpa;
  struct sockaddr_in sock_in;
  struct sockaddr_un sock_un;

  /* do some initialization */
  XPAInitEnv();
  /* init the list of reserved commands */
  XPAInitReserved();

  /* we need a name, but no ":" allowed in the name */
  if( (name == NULL) || (*name == '\0') || strchr(name, ':') )
    return(NULL);
  
  /* we need either a send or a receive or both */
  if( (send_callback == NULL) && (rec_callback == NULL ) ){
    if( verbosity ){
      fprintf(stderr, "XPA$ERROR: requires send and/or receive callback\n");
    }
    return(NULL);
  }

  /* allocate xpa struct */
  if( (xpa = (XPA)calloc(1, sizeof(XPARec))) == NULL )
    return(NULL);

  /* fill in the blanks */
  xpa->version = NewString(XPA_VERSION);
  xpa->type  = (char *)calloc(10, sizeof(char));
  if( xclass && *xclass )
    xpa->xclass = NewString(xclass);
  else
    xpa->xclass = NewString("*");
  xpa->name  = NewString(name);
  xpa->help  = NewString(help);
  /* set the value of the server big-endian-ness */
  xpa->sendian = XPAEndian() ? "big" : "little";
  /* these can be changed by a call to XPAXtAddInput, XPATclAddInput, etc. */
  xpa->seladd = XPAAddOneInput;
  xpa->seldel = XPADelOneInput;
  xpa->selon =  XPAEnableOneInput;
  xpa->seloff = XPADisableOneInput;

  /* fill in send information */
  if( send_callback != NULL ){
    xpa->send_callback = send_callback;
    xpa->send_data = send_data;
    strcat(xpa->type, "g");
    /* process the mode string */
    xpa->send_mode = XPA_DEF_MODE_SEND;
    XPAMode(send_mode, &(xpa->send_mode), "freebuf", XPA_MODE_FREEBUF,1);
    XPAMode(send_mode, &(xpa->send_mode), "acl", XPA_MODE_ACL, 1);
  }

  /* fill in receive information */
  if( rec_callback != NULL ){
    xpa->receive_callback = rec_callback;
    xpa->receive_data = rec_data;
    strcat(xpa->type, "s");
    /* process the mode string */
    xpa->receive_mode = XPA_DEF_MODE_REC;
    XPAMode(rec_mode, &(xpa->receive_mode), "buf", XPA_MODE_BUF, 1);
    XPAMode(rec_mode, &(xpa->receive_mode), "fillbuf", XPA_MODE_FILLBUF, 1);
    XPAMode(rec_mode, &(xpa->receive_mode), "freebuf", XPA_MODE_FREEBUF, 1);
    XPAMode(rec_mode, &(xpa->receive_mode), "acl", XPA_MODE_ACL, 1);
  }

  /* set up communication method */
  switch(mtype){
  case XPA_INET:
    /* open a socket and fill in socket information */
    if( (xpa->fd = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
      goto error;
    setsockopt(xpa->fd, SOL_SOCKET, SO_KEEPALIVE,
	       (char *)&keep_alive, sizeof(keep_alive));
    memset((char *)&sock_in, 0, sizeof(sock_in));
    sock_in.sin_family = AF_INET;
    /* handle special case of xpans */
    if( !strcmp(xpa->name, "xpans") ){
      XPAParseIpPort(XPANSMethod(1), &ip, &port);
      sock_in.sin_addr.s_addr = htonl(ip);
      sock_in.sin_port = htons(port);
      setsockopt(xpa->fd, SOL_SOCKET, SO_REUSEADDR,
	       (char *)&reuse_addr, sizeof(reuse_addr));
    }
    else{
      sock_in.sin_addr.s_addr = htonl(gethostip(NULL));
      sock_in.sin_port = htons(0);
    }
    /* bind to a port */
    if( bind(xpa->fd, (struct sockaddr *)&sock_in, sizeof(sock_in)) < 0 )
      goto error;
    /* we now can determine which port the system assigned */
    if( getsockname(xpa->fd, (struct sockaddr *)&sock_in, &slen) < 0 )
      goto error;
    else{
      /* ip:port is the method */
      sprintf(tbuf, "%x:%d",
	      ntohl(sock_in.sin_addr.s_addr), ntohs(sock_in.sin_port));
      xpa->method = NewString(tbuf);
    }
    break;
  case XPA_UNIX:
    /* open a socket and fill in socket information */
    if( (xpa->fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 )
      goto error;
    setsockopt(xpa->fd, SOL_SOCKET, SO_KEEPALIVE,
	       (char *)&keep_alive, sizeof(keep_alive));
    memset((char *)&sock_in, 0, sizeof(sock_in));
    sock_un.sun_family = AF_UNIX;
    /* handle special case of xpans */
    if( !strcmp(xpa->name, "xpans") ){
      strcpy(tbuf, XPANSMethod(1));
      setsockopt(xpa->fd, SOL_SOCKET, SO_REUSEADDR,
		 (char *)&reuse_addr, sizeof(reuse_addr));
    }
    else{
      /* get filename part, composed of class and name and unique id */
      sprintf(tfile, "%s_%s.%d", xpa->xclass, xpa->name, getpid());
      /* change "/" to "_" for filename */
      for(tfptr = tfile; *tfptr != '\0'; tfptr++){
	if( *tfptr == '/' )
	  *tfptr = '_';
      }
      /* create full pathname */
      sprintf(tbuf, "%s/%s", tmpdir, tfile);
    }
    /* delete old copy */
    unlink (tbuf);
    strcpy(sock_un.sun_path, tbuf);
    /* bind to a port */
    if( bind(xpa->fd, (struct sockaddr *)&sock_un, sizeof(sock_un)) < 0 )
      goto error;
    /* path is the method */
    xpa->method = NewString(tbuf);
    /* save file name so we can delete it later */
    xpa->filename = NewString(tbuf);
    break;
  default:
    goto error;
  }

  /* listen for connections */
  if( listen(xpa->fd, XPA_MAXLISTEN) < 0 )
    goto error;

  /* make sure we close on exec */
  fcntl(xpa->fd, F_SETFD, FD_CLOEXEC);

  /* we have a valid fd -- add it to the active list */
  xpa->selptr = XPAAddOneInput((void *)xpa, xpa->fd);

  /* add this xpa to end of list of xpas */
  XPAListAdd(&xpahead, xpa);

  /* publish this entry to the world */
  XPANSAdd(xpa);

  /* return good news */
  return(xpa);

error:
  if( verbosity ){
    perror("XPANew");
  }
  _XPAFree(xpa);
  return(NULL);
}

/*
 *---------------------------------------------------------------------------
 *
 * Routine: 	XPAFree
 *
 * Purpose: 	free up alloc'ed memory in the XPA record structure
 *
 * Results:	1 on success, 0 for failure
 *
 *---------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAFree (XPA xpa)
#else
int XPAFree(xpa)
     XPA xpa;
#endif
{
  XPA cur;
  XPACmd cmd;
  XPAComm comm, tcomm;

  /* if the xpa is temp and the status is active, just reset and exit */
  if( _XPAValid(txpahead, xpa, XPA_ACLS) ){
    if( xpa->status & XPA_STATUS_ACTIVE ){
      xpa->status |= XPA_STATUS_FREE;
    }
    return(0);
  }

  /* make sure we have something to do */
  if( !_XPAValid(xpahead, xpa, XPA_ACLS) )
    return(-1);

  /* call the primitive routine */
  return(_XPAFree(xpa));

}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAInfoNew
 *
 * Purpose:	add a new xpa name to a process to allow external
 * 		process to send info messages
 *

 * Returns: xpa handle associated with this class.name or NULL
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
XPA 
XPAInfoNew (char *xclass, char *name,
	    InfoCb info_callback, void *info_data, char *info_mode)
#else
XPA XPAInfoNew(xclass, name, info_callback, info_data, info_mode)
     char *xclass;
     char *name;
     InfoCb info_callback;
     void *info_data;
     char *info_mode;
#endif
{
  int slen = sizeof(struct sockaddr_in);
  int keep_alive=1;
  char tbuf[SZ_LINE];
  char tfile[SZ_LINE];
  char *tfptr;
  XPA xpa;
  struct sockaddr_in sock_in;
  struct sockaddr_un sock_un;

  /* do some initialization */
  XPAInitEnv();
  /* init the list of reserved commands */
  XPAInitReserved();

  /* we need a name, but no ":" allowed in the name */
  if( (name == NULL) || (*name == '\0') || strchr(name, ':') )
    return(NULL);
  
  /* we need an info callback */
  if( info_callback == NULL ){
    if( verbosity ){
      fprintf(stderr, "XPA$ERROR: requires info callback\n");
    }
    return(NULL);
  }

  /* allocate xpa struct */
  if( (xpa = (XPA)calloc(1, sizeof(XPARec))) == NULL )
    return(NULL);

  xpa->version = NewString(XPA_VERSION);
  xpa->type  = (char *)calloc(10, sizeof(char));
  /* fill in the blanks */
  if( xclass != NULL )
    xpa->xclass = NewString(xclass);
  else
    xpa->xclass = NewString("*");
  xpa->name  = NewString(name);
  /* these can be changed by a call to XPAXtAddInput, XPATclAddInput, etc. */
  xpa->seladd = XPAAddOneInput;
  xpa->seldel = XPADelOneInput;
  xpa->selon =  XPAEnableOneInput;
  xpa->seloff = XPADisableOneInput;

  /* fill in send information */
  xpa->info_callback = info_callback;
  xpa->info_data = info_data;
  strcat(xpa->type, "i");
  /* process the mode string */
  xpa->info_mode = XPA_DEF_MODE_INFO;
  XPAMode(info_mode, &(xpa->info_mode), "acl", XPA_MODE_ACL, 1);

  /* set up communication method */
  switch(mtype){
  case XPA_INET:
    /* open a socket and fill in socket information */
    if( (xpa->fd = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
      goto error;
    setsockopt(xpa->fd, SOL_SOCKET, SO_KEEPALIVE,
	       (char *)&keep_alive, sizeof(keep_alive));
    memset((char *)&sock_in, 0, sizeof(sock_in));
    sock_in.sin_family = AF_INET;
    sock_in.sin_addr.s_addr = htonl(gethostip(NULL));
    sock_in.sin_port = htons(0);
    /* bind to a port */
    if( bind(xpa->fd, (struct sockaddr *)&sock_in, sizeof(sock_in)) < 0 )
      goto error;
    /* we now can determine which port the system assigned */
    if( getsockname(xpa->fd, (struct sockaddr *)&sock_in, &slen) < 0 )
      goto error;
    else{
      /* ip:port is the method */
      sprintf(tbuf, "%x:%d",
	      ntohl(sock_in.sin_addr.s_addr), ntohs(sock_in.sin_port));
      xpa->method = NewString(tbuf);
    }
    break;
  case XPA_UNIX:
    /* get filename part, composed of class and name and unique id */
    sprintf(tfile, "%s_%s.%d", xpa->xclass, xpa->name, getpid());
    /* change "/" to "_" for filename */
    for(tfptr = tfile; *tfptr != '\0'; tfptr++){
      if( *tfptr == '/' )
	*tfptr = '_';
    }
    /* create full pathname */
    sprintf(tbuf, "%s/%s", tmpdir, tfile);
    /* delete old copy */
    unlink (tbuf);
    /* open a socket and fill in socket information */
    if( (xpa->fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 )
      goto error;
    setsockopt(xpa->fd, SOL_SOCKET, SO_KEEPALIVE,
	       (char *)&keep_alive, sizeof(keep_alive));
    memset((char *)&sock_in, 0, sizeof(sock_in));
    sock_un.sun_family = AF_UNIX;
    strcpy(sock_un.sun_path, tbuf);
    /* bind to a port */
    if( bind(xpa->fd, (struct sockaddr *)&sock_un, sizeof(sock_un)) < 0 )
      goto error;
    /* path is the method */
    xpa->method = NewString(tbuf);
    break;
  default:
    goto error;
  }

  /* listen for connections */
  if( listen(xpa->fd, XPA_MAXLISTEN) < 0 )
    goto error;

  /* make sure we close on exec */
  fcntl(xpa->fd, F_SETFD, FD_CLOEXEC);

  /* we have a valid fd -- add it to the active list */
  xpa->selptr = XPAAddOneInput((void *)xpa, xpa->fd);

  /* publish this entry to the world */
  XPANSAdd(xpa);

  /* add this xpa to end of list of xpas */
  XPAListAdd(&xpahead, xpa);

  /* return good news */
  return(xpa);

error:
  XPAFree(xpa);
  return(NULL);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAPoll
 *
 * Purpose:	non-blocking handling of XPA access points
 *		timeout in millisecs, but if negative, no timeout is used
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int
XPAPoll (int msec, int maxreq)
#else
int XPAPoll(msec, maxreq)
     int msec;
     int maxreq;
#endif
{
  int sgot;
  int got=0;
  fd_set readfds;
  struct timeval tv;
  struct timeval *tvp;

  if( msec >= 0 ){
    tv.tv_sec = msec / 1000;
    tv.tv_usec = (msec % 1000) * 1000;
    tvp = &tv;
  }
  /* negative value means just block and wait */
  else{
    tvp = NULL;
  }
  FD_ZERO(&readfds);
  if( XPAAddSelect(NULL, &readfds) ){
    sgot = select(swidth, &readfds, NULL, NULL, tvp);
    /* error -- what should we do? */
    if( sgot < 0 ){
      if( verbosity ){
	perror("XPAPoll() select");
      }
      exit(1);
    }
    /* timeout -- just return */
    else if( sgot == 0 )
      ;
    /* finally ... something to do */
    else
      got = XPAProcessSelect(&readfds, maxreq);
  }
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAMainLoop
 *
 * Purpose:	non-X programs event loop for handling XPA
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int
XPAMainLoop (void)
#else
int XPAMainLoop()
#endif
{
  int sgot;
  int got=0;
  fd_set readfds;


  FD_ZERO(&readfds);
  while( XPAAddSelect(NULL, &readfds) ){
#ifdef XPA_DEBUG
    fprintf(stderr, "XPA waiting for select() ...\n");
#endif
    sgot = select(swidth, &readfds, NULL, NULL, NULL);
#ifdef XPA_DEBUG
    fprintf(stderr, "XPA select() returned: %d\n", sgot);
#endif
    /* error -- what should we do? */
    if( sgot < 0 ){
      if( verbosity ){
	perror("XPAMainLoop() select");
      }
      exit(1);
    }
    /* can't happen, since we have no timeout */
    else if( sgot == 0 )
      ;
    /* finally ... something to do */
    else
      got += XPAProcessSelect(&readfds, 0);
    FD_ZERO(&readfds);
  }
  return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPASleep
 *
 * Purpose:	sleep for specified milliseconds
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
void 
XPASleep (int msec)
#else
void XPASleep(msec)
     int msec;
#endif
{
  struct timeval tv;

  if( msec > 0 ){
    tv.tv_sec = msec / 1000;
    tv.tv_usec = (msec % 1000) * 1000;
    select(1, NULL, NULL, NULL, &tv);
  }
}

